Package: sumer
Type: Package
Title: Sumerian Cuneiform Text Analysis
Version: 1.6.0
Authors@R: c(
    person(
			given = "Robin",
           family = "Wellmann",
           role = c("aut", "cre"),
           email = "ro.wellmann@gmail.com"
		)
	)
Description: Provides functions for converting transliterated Sumerian texts to sign names and cuneiform characters,
    creating and querying dictionaries, analyzing the structure of
    Sumerian words, and creating translations. Includes a built-in dictionary and supports both
    forward lookup (Sumerian to English) and reverse lookup (English to
    Sumerian).
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.0.0)
Imports: stringr, officer, xml2, cli, rlang, ggplot2, ragg, shiny
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: yes
Maintainer: Robin Wellmann <ro.wellmann@gmail.com>
Packaged: 2026-05-05 07:56:57 UTC; rowel
Author: Robin Wellmann [aut, cre]
Repository: CRAN
Date/Publication: 2026-05-05 08:20:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-05 08:36:49 UTC; unix
Archs: sumer.so.dSYM

Package: sumer
Type: Package
Title: Sumerian Cuneiform Text Analysis
Version: 1.6.0
Authors@R: c(
    person(
			given = "Robin",
           family = "Wellmann",
           role = c("aut", "cre"),
           email = "ro.wellmann@gmail.com"
		)
	)
Description: Provides functions for converting transliterated Sumerian texts to sign names and cuneiform characters,
    creating and querying dictionaries, analyzing the structure of
    Sumerian words, and creating translations. Includes a built-in dictionary and supports both
    forward lookup (Sumerian to English) and reverse lookup (English to
    Sumerian).
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.0.0)
Imports: stringr, officer, xml2, cli, rlang, ggplot2, ragg, shiny
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: yes
Maintainer: Robin Wellmann <ro.wellmann@gmail.com>
Packaged: 2026-05-05 07:56:57 UTC; rowel
Author: Robin Wellmann [aut, cre]
Repository: CRAN
Date/Publication: 2026-05-05 08:20:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-05 10:49:34 UTC; unix
Archs: sumer.so.dSYM

# Compiled code
useDynLib(sumer, .registration = TRUE)

# Exports
export(as.cuneiform)
export(as.sign_name)
export(convert_to_dictionary)
export(info)
export(look_up)
export(make_dictionary)
export(read_dictionary)
export(read_translated_text)
export(save_dictionary)
export(split_sumerian)
export(ngram_frequencies)
export(sign_grammar)
export(plot_sign_grammar)
export(prior_probs)
export(grammar_probs)
export(skeleton)
export(mark_ngrams)
export(guess_substr_info)
export(translate)
export(apply_translation_rules)
export(grammatical_structure)
export(fill_substr_info)
export(merge_dictionaries)
export(translation_context)
export(translate_line)

# S3 methods
S3method(as.cuneiform, character)
S3method(as.cuneiform, default)
S3method(as.sign_name, character)
S3method(as.sign_name, default)
S3method(print, info)
S3method(print, cuneiform)
S3method(print, sign_name)
S3method(print, look_up)
S3method(print, skeleton)
S3method(print, grammatical_structure)

# Imports from stringr
importFrom(stringr, str_c)
importFrom(stringr, str_count)
importFrom(stringr, str_detect)
importFrom(stringr, str_ends)
importFrom(stringr, str_escape)
importFrom(stringr, str_remove)
importFrom(stringr, str_remove_all)
importFrom(stringr, str_replace)
importFrom(stringr, str_replace_all)
importFrom(stringr, str_locate_all)
importFrom(stringr, str_length)
importFrom(stringr, str_split)
importFrom(stringr, str_split_1)
importFrom(stringr, str_split_fixed)
importFrom(stringr, str_squish)
importFrom(stringr, str_starts)
importFrom(stringr, str_sub)
importFrom(stringr, str_trim)
importFrom(stringr, regex)
importFrom(stringr, fixed)

# Imports from utils
importFrom(utils, read.csv2)
importFrom(utils, read.table)
importFrom(utils, write.table)

# Imports from stats
importFrom(stats, aggregate)
importFrom(stats, setNames)

# Import .data pronoun for ggplot2 aes()
importFrom(rlang, .data)

# Imports from officer
importFrom(officer, read_docx)
importFrom(officer, docx_summary)

# Imports from xml2
importFrom(xml2, xml_find_all)
importFrom(xml2, xml_remove)

# Imports from cli
importFrom(cli, col_blue)
importFrom(cli, col_green)
importFrom(cli, col_red)
importFrom(cli, make_ansi_style)

# Imports from grDevices / tools (used in plot_sign_grammar)
importFrom(grDevices, dev.off)
importFrom(tools, file_ext)

# =============================================================================
# apply_translation_rules: recursively parse a bracketed structure and translate
# =============================================================================


#' Apply Translation Rules to a Bracketed Structure String
#'
#' Recursively parses a string with \code{\{\}} brackets and \code{#N} tags,
#' resolves all sub-groups via recursion, then delegates to
#' \code{\link{eval_operator}} for the actual evaluation.
#'
#' @param s Bracketed string from \code{\link{add_brackets}} (only \code{\{\}} brackets).
#' @param type Character vector of grammatical types. \code{type[k]} is the
#'   type string for tag \code{#k}.
#' @param translation Character vector of translations. \code{translation[k]}
#'   is the translation for tag \code{#k}.
#' @return \code{c(type, translation)} of length 2, or error message of length 1.
#' @export
apply_translation_rules <- function(s, type, translation) {
  chars <- strsplit(s, "")[[1]]
  n <- length(chars)
  elements <- list()   # list of c(type_str, translation_str)
  seps <- character(0)
  buf <- ""
  i <- 1L

  # --- Phase 1: Parse elements and separators ---
  while (i <= n) {
    ch <- chars[i]

    if (ch == "#") {
      if (length(elements) > 0) seps <- c(seps, buf)
      buf <- ""
      rt <- read_tag(chars, i)
      i <- rt$next_i
      elements[[length(elements) + 1L]] <- c(type[rt$tag_num], translation[rt$tag_num])

    } else if (ch == "{") {
      if (length(elements) > 0) seps <- c(seps, buf)
      buf <- ""
      depth <- 1L; inner <- ""; i <- i + 1L
      while (i <= n && depth > 0) {
        if (chars[i] == "{") depth <- depth + 1L
        if (chars[i] == "}") depth <- depth - 1L
        if (depth > 0) inner <- paste0(inner, chars[i])
        i <- i + 1L
      }
      result <- apply_translation_rules(inner, type, translation)
      if (length(result) == 1) return(result)
      elements[[length(elements) + 1L]] <- result

    } else {
      buf <- paste0(buf, ch)
      i <- i + 1L
    }
  }
  trailing_sep <- buf

  # --- Phase 2: Base cases ---
  if (length(elements) == 0)
    return(c(NA_character_, ""))
  if (length(elements) == 1) {
    return(eval_operator(elements[[1]][1], elements[[1]][2], trailing_sep = trailing_sep))
  }

  # --- Phase 3: Prepare parameters for eval_operator ---
  op_idx <- NULL
  op_parsed <- NULL
  for (j in seq_along(elements)) {
    p <- parse_type(elements[[j]][1])
    if (p$is_operator) { op_idx <- j; op_parsed <- p; break }
  }

  if (!is.null(op_idx)) {
    # Operator found: build named args list
    operator  <- elements[[op_idx]][1]
    rule_trans <- elements[[op_idx]][2]
    arg_el <- elements[-op_idx]

    args <- build_args(arg_el)

    return(eval_operator(operator, rule_trans, args, seps, trailing_sep))
  }

  # No operator: composition
  args <- build_args(elements)
  eval_operator(NULL, NULL, args, seps, trailing_sep)
}

skeleton <- function(x, mapping = NULL, fill = NULL, space = FALSE) {
  if ((!is.character(x)) | length(x) > 1) {
    stop("The parameter x must be a character vector of length 1.")
  }

  if (is.null(mapping)) {
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep = ";", na.strings = "")
  }

  "get_equation" <- function(x) {
    x   <- info(x, mapping = mapping)
    res <- paste0(c(paste0(x$reading, collapse = "-"),
                    paste0(x$name, collapse = "."),
                    paste0(x$sign, collapse = "")), collapse = "=")
    return(res)
  }

  # Convert to cuneiform and extract tokens
  x     <- as.cuneiform(x, mapping = mapping)
  token <- split_sumerian(x)$signs
  N     <- length(token)

  # Validate fill if provided
  if (!is.null(fill)) {
    fill_tokens <- split_sumerian(fill$expr[1])$signs
    if (!identical(fill_tokens, token) || nrow(fill) != N * (N + 1) / 2) {
      stop("The parameter fill does not match x.")
    }
  }

  # Normalize brackets and extract hierarchical structure
  x_marked <- mark_skeleton_entries(x)
  df       <- extract_skeleton_entries(x_marked)
  n_entries <- nrow(df)

  # Add column with indices into the fill data frame
  df$fill_idx <- substr_position(df$start, df$n_tokens, N)

  # Pre-compute type and translation columns from fill
  if (!is.null(fill)) {
    df$type        <- fill$type[df$fill_idx]
    df$translation <- fill$translation[df$fill_idx]
  } else {
    df$type        <- ""
    df$translation <- ""
  }

  df$type <- str_replace_all(df$type, "x","\u2612")
  df$type <- str_replace_all(df$type, "->","\u2192")

  # Replace operator "S☒→V" with "Vt"
  df$type[df$type == paste0("S\u2612\u2192V")] <- "Vt"

  # Build header line from root row (row 1, depth 0)
  if (N == 1) {
    header <- get_equation(x_marked)
  } else {
    header <- paste0(info(x_marked, mapping = mapping)$reading, collapse = "-")
  }
  result <- paste0("|", header, ": ", df$type[1], ": ", df$translation[1])

  # Build entry lines (rows 2..n, depth >= 1)
  if (n_entries >= 2) for (i in 2:n_entries) {
    # Insert blank line before depth-1 entries
    if (space && df$depth[i] == 1) {
      result <- c(result, "")
    }

    eq <- get_equation(df$expr[i])

    result <- c(result, paste0("|", strrep("\t", df$depth[i] - 1),
                                eq, ": ", df$type[i], ": ", df$translation[i]))
  }

  result <- c(paste0("Structure: ", str_trim(x)),"",result)
  class(result) <- c("skeleton", "character")
  result
}


print.skeleton <- function(x, ...) {
  cat(x, sep = "\n")
}

fill_substr_info <- function(skeleton) {
  # 1. Datei lesen falls nötig
  if (length(skeleton) == 1 && file.exists(skeleton))
    skeleton <- readLines(skeleton)

  # 2. x aus skeleton[1] extrahieren
  x <- str_split_1(skeleton[1], ":")[2]
  x <- str_trim(x)

  # 3. df erzeugen (Skeleton-Einträge mit Hierarchie)
  x_marked <- mark_skeleton_entries(x)
  df <- extract_skeleton_entries(x_marked)

  # 4. type und translation aus Skeleton-Zeilen in df übertragen
  entry_lines <- skeleton[startsWith(skeleton, "|")]
  df$type <- ""
  df$translation <- ""
  for (i in seq_len(nrow(df))) {
    line <- entry_lines[i]
    if (df$depth[i] == 0) {
      # Format: |header:type:translation (split at first two ":")
      stripped <- sub("^\\|", "", line)
      first  <- str_split_fixed(stripped, ":", 2)
      second <- str_split_fixed(first[,2], ":", 2)
    } else {
      # Format: |[tabs]eq: type: translation (split at first two ": ")
      stripped <- sub("^\\|\\t*", "", line)
      first  <- str_split_fixed(stripped, ": ", 2)
      second <- str_split_fixed(first[,2], ": ", 2)
    }
    df$type[i]        <- str_trim(second[,1])
    df$translation[i] <- str_trim(second[,2])
  }

  # 5. df_substr erzeugen (alle N*(N+1)/2 Substrings)
  token <- split_sumerian(x)$signs
  N <- length(token)
  df_substr <- init_substr_info(token)

  # 6. Einträge von df nach df_substr übertragen via substr_position
  positions <- substr_position(df$start, df$n_tokens, N)
  df_substr$type[positions]        <- df$type
  df_substr$translation[positions] <- df$translation

  df_substr
}

translate <- function(x, text = NULL, dic = NULL, mapping = NULL, fill = NULL,
                      min_freq = c(6, 4, 2), sentence_prob = 1.0,
                      viewer = shiny::paneViewer()) {


  # --- Helpers ---
  null_to_empty <- function(val) if (is.null(val)) "" else val

  # ===================================================================
  # Helper: detect sentence boundaries
  # Returns integer vector of cuneiform token indices after which a

  # sentence-ending dot occurs (e.g. c(7, 11, 13)).
  # ===================================================================
  detect_sentence_ends <- function(x_raw, mapping) {
    parts_raw <- split_sumerian(x_raw)
    # Find dots in separators: sep[i+1] contains "." means after raw token i
    sentence_ends_raw <- integer(0)
    for (i in seq_along(parts_raw$signs)) {
      if (grepl("\\.", parts_raw$separators[i + 1])) {
        sentence_ends_raw <- c(sentence_ends_raw, i)
      }
    }
    if (length(sentence_ends_raw) == 0) return(integer(0))

    # Map raw token indices to cuneiform token indices
    raw_to_cunei_end <- integer(length(parts_raw$signs))
    cunei_pos <- 0L
    for (i in seq_along(parts_raw$signs)) {
      cu <- as.cuneiform(parts_raw$signs[i], mapping = mapping)
      n_cunei <- length(split_sumerian(cu)$signs)
      cunei_pos <- cunei_pos + n_cunei
      raw_to_cunei_end[i] <- cunei_pos
    }
    raw_to_cunei_end[sentence_ends_raw]
  }

  # ===================================================================
  # 1. Resolve mapping
  # ===================================================================
  if (is.null(mapping)) {
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep = ";", na.strings = "")
  } else if (is.character(mapping) && length(mapping) == 1) {
    mapping <- read.csv2(mapping, sep = ";", na.strings = "")
  }

  # ===================================================================
  # 2. Resolve text
  # ===================================================================
  if (is.character(text) && length(text) == 1 && file.exists(text)) {
    text <- readLines(text, warn = FALSE)
  }

  # ===================================================================
  # 3. Resolve dic into a list
  # ===================================================================
  if (is.null(dic)) {
    dic_list <- list(read_dictionary())
  } else if (is.character(dic)) {
    dic_list <- lapply(dic, read_dictionary, verbose = FALSE)
  } else if (is.data.frame(dic)) {
    dic_list <- list(dic)
  } else {
    dic_list <- dic
  }

  if (is.character(dic)) {
    dic_names <- tools::file_path_sans_ext(basename(dic))
  } else {
    dic_names <- paste0("Dictionary ", seq_along(dic_list))
  }
  names(dic_list) <- dic_names

  # Remove translation rows with empty or "?" meanings
  dic_list <- lapply(dic_list, function(d) {
    drop <- d$row_type == "trans." &
            (is.na(d$meaning) | d$meaning == "" | d$meaning == "?")
    d[!drop, , drop = FALSE]
  })

  # ===================================================================
  # 4. Resolve x into cuneiform string and determine line_index
  # ===================================================================
  extract_line_number <- function(line) {
    m <- regmatches(line, regexpr("^\\s*\\d+", line))
    if (length(m) > 0) as.integer(m) else NA_integer_
  }

  if (is.numeric(x) && is.null(text))
    stop("When x is an integer line number, text must be provided.")

  line_index <- NA_integer_

  if (is.numeric(x)) {
    line_number <- as.integer(x)
    text_numbers <- vapply(text, extract_line_number, integer(1), USE.NAMES = FALSE)
    match_pos <- which(text_numbers == line_number)
    if (length(match_pos) > 0) {
      line_index <- match_pos[1]
    } else {
      line_index <- line_number
    }
    x_raw <- sub("^\\s*\\d+[.)\\s]\\s*", "", text[line_index])
  } else {
    x_raw <- x
    if (!is.null(text)) {
      x_cunei_test  <- as.cuneiform(x_raw, mapping = mapping)
      x_tokens_test <- split_sumerian(x_cunei_test)$signs
      for (li in seq_along(text)) {
        stripped <- sub("^\\s*\\d+[.)\\s]\\s*", "", text[li])
        line_cunei <- tryCatch(as.cuneiform(stripped, mapping = mapping),
                               error = function(e) NULL)
        if (is.null(line_cunei)) next
        line_tokens <- split_sumerian(line_cunei)$signs
        if (identical(line_tokens, x_tokens_test)) {
          line_index <- li
          break
        }
      }
    }
  }

  x_cunei <- as.cuneiform(x_raw, mapping = mapping)
  token   <- split_sumerian(x_cunei)$signs
  N       <- length(token)

  # ===================================================================
  # 5. Resolve fill
  # ===================================================================
  if (is.null(fill)) {
    fill <- guess_substr_info(x_cunei, dic_list, mapping = mapping)
  }
  fill_tokens <- split_sumerian(fill$expr[1])$signs
  if (!identical(fill_tokens, token) || nrow(fill) != N * (N + 1) / 2) {
    stop("The parameter fill does not match x.")
  }

  # ===================================================================
  # 6. Pre-compute data for the panels
  # ===================================================================
  if (!is.null(text)) {
    ngram_full <- ngram_frequencies(text, min_freq = min_freq, mapping = mapping)
  } else {
    ngram_full <- ngram_frequencies(x_cunei, min_freq = min_freq, mapping = mapping)
  }

  main_dic <- dic_list[[1]]
  sg       <- sign_grammar(x_cunei, main_dic, mapping = mapping)
  prior    <- prior_probs(main_dic, sentence_prob = sentence_prob)
  gp       <- grammar_probs(sg, prior, main_dic)

  line_str <- paste0(token, collapse = "")

  # N-grams in current line (length >= 2) — used for merged table
  ngram_in_line <- ngram_full[ngram_full$length >= 2 &
    vapply(ngram_full$combination, function(c) {
      grepl(c, line_str, fixed = TRUE)
    }, logical(1)), , drop = FALSE]

  # Multi-sign combination summary (for all substrings of length >= 2)
  multi_sign_df <- NULL
  if (N >= 2) {
    summary_rows <- list()
    for (len in N:2) {
      for (start in 1:(N - len + 1)) {
        substr_cunei <- paste0(token[start:(start + len - 1)], collapse = "")
        substr_sn <- as.character(as.sign_name(substr_cunei, mapping = mapping))

        # Collect types and counts from all dictionaries
        all_types  <- character(0)
        all_counts <- integer(0)
        has_entries <- FALSE

        for (di in seq_along(dic_list)) {
          d <- dic_list[[di]]
          rows <- d[d$sign_name == substr_sn & d$row_type == "trans.", , drop = FALSE]
          if (nrow(rows) > 0) {
            has_entries <- TRUE
            all_types  <- c(all_types, rows$type)
            all_counts <- c(all_counts,
              if ("count" %in% names(rows)) rows$count else rep(1L, nrow(rows)))
          }
        }
        if (!has_entries) next

        # Freq: sum of counts in dictionary 1
        d1 <- dic_list[[1]]
        d1_rows <- d1[d1$sign_name == substr_sn & d1$row_type == "trans.", , drop = FALSE]
        freq_val <- if (nrow(d1_rows) > 0 && "count" %in% names(d1_rows))
          sum(d1_rows$count, na.rm = TRUE) else NA_integer_

        # Type suggestions: unique types ordered by total count
        type_df  <- aggregate(list(n = all_counts), list(type = all_types), sum, na.rm = TRUE)
        type_df  <- type_df[order(-type_df$n), ]
        type_str <- paste(type_df$type, collapse = ", ")

        summary_rows[[length(summary_rows) + 1]] <- data.frame(
          freq           = freq_val,
          sign_names     = substr_sn,
          combination    = substr_cunei,
          type_suggestions = type_str,
          stringsAsFactors = FALSE
        )
      }
    }
    if (length(summary_rows) > 0) {
      multi_sign_df <- do.call(rbind, summary_rows)
      rownames(multi_sign_df) <- NULL
    }
  }

  # ===================================================================
  # 7. Validate bracket input
  # ===================================================================
  validate_bracket_input <- function(s, original_tokens, mapping) {
    # --- Check bracket matching and nesting ---
    chars       <- strsplit(s, "")[[1]]
    openers     <- c("(", "<", "{")
    closers     <- c(")", ">", "}")
    closing_map <- c(")" = "(", ">" = "<", "}" = "{")
    opener_for  <- c("(" = ")", "<" = ">", "{" = "}")
    stack       <- character(0)

    for (ch in chars) {
      if (ch %in% openers) {
        stack <- c(stack, ch)
      } else if (ch %in% closers) {
        if (length(stack) == 0) {
          return(paste0("Unmatched closing bracket '", ch, "'."))
        }
        if (stack[length(stack)] != closing_map[ch]) {
          return(paste0("Bracket mismatch: expected '",
                        opener_for[stack[length(stack)]],
                        "' but found '", ch, "'."))
        }
        stack <- stack[-length(stack)]
      }
    }
    if (length(stack) > 0) {
      return(paste0("Unclosed bracket '", stack[length(stack)], "'."))
    }

    # --- Convert to cuneiform and check tokens ---
    s_cunei <- tryCatch(
      as.cuneiform(s, mapping = mapping),
      error = function(e) NULL
    )
    if (is.null(s_cunei)) {
      return("Cannot convert input to cuneiform. Check for invalid signs.")
    }

    s_tokens <- split_sumerian(s_cunei)$signs
    if (!identical(s_tokens, original_tokens)) {
      return("Tokens do not match the original line. Only brackets and dots may be changed.")
    }

    NULL   # input is valid
  }

  # ===================================================================
  # 7b. get_equation helper
  # ===================================================================
  get_equation <- function(expr) {
    xi  <- info(expr, mapping = mapping)
    res <- paste0(c(paste0(xi$reading, collapse = "-"),
                    paste0(xi$name, collapse = "."),
                    paste0(xi$sign, collapse = "")), collapse = "=")
    return(res)
  }

  # ===================================================================
  # 8. Initial skeleton computation
  # ===================================================================
  x_marked_init <- mark_skeleton_entries(x_cunei)
  df_init       <- extract_skeleton_entries(x_marked_init)
  df_init$fill_idx <- substr_position(df_init$start, df_init$n_tokens, N)

  fill_idx_header <- df_init$fill_idx[1]

  # ===================================================================
  # 9. Sentence boundaries for background colours
  # ===================================================================
  sentence_ends <- detect_sentence_ends(x_raw, mapping)

  # Assign sentence index (1-based) to each cuneiform token
  # Tokens 1..sentence_ends[1] -> sentence 1, etc.
  token_sentence <- integer(N)
  sent_idx <- 1L
  se_ptr   <- 1L
  for (ti in seq_len(N)) {
    token_sentence[ti] <- sent_idx
    if (se_ptr <= length(sentence_ends) && ti == sentence_ends[se_ptr]) {
      sent_idx <- sent_idx + 1L
      se_ptr   <- se_ptr + 1L
    }
  }
  n_sentences <- max(token_sentence)

  # Background colours for sentences (clearly alternating)
  sent_colours <- c("#e0e0e0", "#ececec")

  # Function: determine sentence of a skeleton entry (by its start token)
  entry_sentence <- function(start_tok) {
    token_sentence[start_tok]
  }

  # ===================================================================
  # CSS
  # ===================================================================
  custom_css <- shiny::tags$style(shiny::HTML("
    body { font-family: 'Segoe UI Historic', 'Noto Sans', sans-serif; }

    /* --- Navigation menu at top --- */
    .nav-menu { position: sticky; top: 0; z-index: 100;
                background: #f8f8f8; border-bottom: 1px solid #ccc;
                padding: 6px 12px; display: flex; gap: 12px; }
    .nav-menu a { cursor: pointer; padding: 4px 12px; border-radius: 4px;
                  text-decoration: none; color: #333; font-size: 14px; }
    .nav-menu a:hover { background: #e0e0e0; }

    /* --- Collapsible sections --- */
    .section-header { cursor: pointer; user-select: none; padding: 8px 0;
                      font-size: 16px; font-weight: bold; color: #333;
                      border-bottom: 1px solid #ddd; margin-bottom: 8px; }
    .section-header:hover { color: #337ab7; }
    .section-content { overflow: hidden; }
    .section-content.collapsed { display: none; }

    /* --- Skeleton table for aligned columns --- */
    .skel-table { border-collapse: collapse; width: 100%; table-layout: fixed; }
    .skel-table td { padding: 3px 4px; vertical-align: middle; }
    .skel-table .form-group { margin-bottom: 0; padding: 0; }
    .skel-table .control-label { display: none; }
    .skel-btn-cell { width: 80px; white-space: nowrap; }
    .skel-btn-cell + td { padding-left: 12px !important; }
    .skel-btn-cell .btn { display: inline-block; vertical-align: middle; }
    .skel-btn-cell .btn + .btn { margin-left: 4px; }
    .skel-label { font-size: 13px; white-space: pre-wrap;
                  word-break: break-word; overflow: hidden;
                  width: 35%; }
    .type-input { width: 88px; }
    .type-input .form-control { width: 80px; padding: 2px 4px; height: 28px; }
    .trans-input .shiny-input-container { width: 100% !important; }
    .trans-input .form-control { width: 100%; padding: 2px 4px; height: 28px; }
    .trans-input textarea.form-control { height: auto; resize: vertical; }
    .skel-spacer td { height: 8px; padding: 0 !important; }

    /* --- N-gram merged table (light grey) --- */
    .ngram-table { border-collapse: collapse; width: 100%; margin-bottom: 16px;
                   background-color: #f0f0f0; }
    .ngram-table th { padding: 4px 8px; text-align: left; font-size: 13px;
                      border-bottom: 2px solid #c0c0c0; }
    .ngram-table td { padding: 4px 8px; font-size: 13px;
                      border-bottom: 1px solid #d8d8d8; }

    /* --- Dictionary lookup tables --- */
    .dict-table { border-collapse: collapse; width: 100%; margin-bottom: 16px;
                  background-color: #e8f4fd; }
    .dict-table th { padding: 4px 8px; text-align: left; font-size: 13px;
                     border-bottom: 2px solid #b8d4e8; }
    .dict-table td { padding: 4px 8px; font-size: 13px;
                     border-bottom: 1px solid #d0e4f0; }
    .dict-table tr { cursor: pointer; }
    .dict-table tr:hover td { background-color: #d0e8f5; }

    /* --- Buttons: green book --- */
    .btn-skel { padding: 2px 8px; font-size: 12px; min-width: 30px;
                background-color: #4a9e4a; border-color: #3d8b3d; color: white; }
    .btn-skel:hover { background-color: #3a8a3a; border-color: #2e7a2e; color: white; }
    .btn-skel:focus { background-color: #3a8a3a; border-color: #2e7a2e; color: white; }
    .btn-skel:disabled { background-color: #b5c8b5; border-color: #a8bca8; color: #e0e8e0;
                         cursor: not-allowed; opacity: 0.65; }
    .btn-skel-brown { padding: 2px 8px; font-size: 12px; min-width: 30px;
                background-color: #8B5E3C; border-color: #7A4E2C; color: white; }
    .btn-skel-brown:hover { background-color: #7A4E2C; border-color: #693E1C; color: white; }
    .btn-skel-brown:focus { background-color: #7A4E2C; border-color: #693E1C; color: white; }
    .btn-skel-brown:disabled { background-color: #D9C4A8; border-color: #CCB89C; color: #f5f0eb;
                               cursor: not-allowed; opacity: 0.65; }
    .btn-update { background-color: #5cb85c; border-color: #4cae4c; color: white; }
    .btn-update:hover { background-color: #449d44; border-color: #398439; color: white; }

    /* --- Header row highlight --- */
    .skel-header td { background-color: #d9edf7; }
    .skel-header td:first-child { border-radius: 4px 0 0 4px; }
    .skel-header td:last-child { border-radius: 0 4px 4px 0; }

    /* --- Done button centered at bottom --- */
    .done-bar { text-align: center; padding: 16px 0; margin-top: 20px;
                border-top: 1px solid #ddd; }
    .btn-done { padding: 8px 40px; font-size: 15px;
                background-color: #5cb85c; border-color: #4cae4c; color: white; }
    .btn-done:hover { background-color: #449d44; border-color: #398439; color: white; }
    .btn-cancel { padding: 8px 40px; font-size: 15px; margin-left: 12px;
                  background-color: #d9534f; border-color: #d43f3a; color: white; }
    .btn-cancel:hover { background-color: #c9302c; border-color: #ac2925; color: white; }

    /* --- Equation label with line breaks before = --- */
    .eq-text { word-break: break-all; }
    .eq-text .eq-break { display: inline; }
    @media (min-width: 1px) {
      .eq-text .eq-break::before { content: '\\A'; white-space: pre; }
    }
  "))

  # ===================================================================
  # Helper: format equation with line-break hints before '='
  # ===================================================================
  format_equation_html <- function(eq) {
    # Only add breaks for long equations (> 30 chars)
    if (nchar(eq) <= 30) {
      return(shiny::span(class = "eq-text", eq))
    }
    # Split at '=' and rejoin with wbr tags
    parts <- strsplit(eq, "=", fixed = TRUE)[[1]]
    if (length(parts) <= 1) return(shiny::span(class = "eq-text", eq))
    elems <- list(shiny::span(parts[1]))
    for (k in 2:length(parts)) {
      elems <- c(elems, list(shiny::tags$wbr(), shiny::span(paste0("=", parts[k]))))
    }
    do.call(shiny::span, c(list(class = "eq-text"), elems))
  }

  # ===================================================================
  # UI: single scrollable page with anchored sections
  # ===================================================================
  ui <- shiny::fluidPage(
    custom_css,
    shiny::tags$title("Translate"),

    # --- Navigation menu at top ---
    shiny::div(class = "nav-menu",
      shiny::tags$a("N-grams", href = "#sec_ngrams", onclick = "event.preventDefault(); document.getElementById('sec_ngrams').scrollIntoView({behavior:'smooth'});"),
      shiny::tags$a("Combinations", href = "#sec_combinations", onclick = "event.preventDefault(); document.getElementById('sec_combinations').scrollIntoView({behavior:'smooth'});"),
      shiny::tags$a("Context", href = "#sec_context", onclick = "event.preventDefault(); document.getElementById('sec_context').scrollIntoView({behavior:'smooth'});"),
      shiny::tags$a("Grammar", href = "#sec_grammar", onclick = "event.preventDefault(); document.getElementById('sec_grammar').scrollIntoView({behavior:'smooth'});"),
      shiny::tags$a("Translation", href = "#sec_translation", onclick = "event.preventDefault(); document.getElementById('sec_translation').scrollIntoView({behavior:'smooth'});")
    ),

    # JavaScript for collapsible sections
    shiny::tags$script(shiny::HTML("
      function toggleSection(id) {
        var el = document.getElementById(id);
        el.classList.toggle('collapsed');
        var arrow = document.getElementById(id + '_arrow');
        arrow.textContent = el.classList.contains('collapsed') ? '\\u25B6' : '\\u25BC';
      }
      Shiny.addCustomMessageHandler('scrollTo', function(id) {
        var el = document.getElementById(id);
        if (el) el.scrollIntoView({behavior: 'smooth'});
      });
    ")),

    shiny::div(style = "padding: 12px 20px;",

      # =============================================================
      # Chapter 1: N-grams (collapsible)
      # =============================================================
      shiny::tags$div(id = "sec_ngrams"),
      shiny::div(class = "section-header",
        onclick = "toggleSection('content_ngrams')",
        shiny::span(id = "content_ngrams_arrow", "\u25BC"),
        " N-gram Patterns"
      ),
      shiny::div(id = "content_ngrams", class = "section-content",
        shiny::p(style = "font-size: 13px; color: #555; margin-bottom: 8px;",
          "This table shows recurring sign combinations (n-grams) that appear in the",
          "current line. Frequencies refer to the full text.",
          "A checkmark in the Theme column indicates that the combination also occurs",
          "in one of the preceding or following lines."
        ),
        shiny::uiOutput("ngram_merged_table")
      ),

      shiny::hr(),

      # =============================================================
      # Chapter 1b: Sign Combination Suggestions (collapsible)
      # =============================================================
      shiny::tags$div(id = "sec_combinations"),
      shiny::div(class = "section-header",
        onclick = "toggleSection('content_combinations')",
        shiny::span(id = "content_combinations_arrow", "\u25BC"),
        " Sign Combination Suggestions"
      ),
      shiny::div(id = "content_combinations", class = "section-content",
        shiny::p(style = "font-size: 13px; color: #555; margin-bottom: 8px;",
          "The following table shows sign combinations included in the dictionaries.",
          "Column Freq contains the count of the respective combination in the primary",
          "dictionary (Dictionary 1). Keep in mind that sign combinations are often not",
          "to be treated as fixed terms \u2014 their meanings may derive from the meanings",
          "of the individual signs."
        ),
        shiny::uiOutput("combination_table")
      ),

      shiny::hr(),

      # =============================================================
      # Chapter 2: Context (collapsible)
      # =============================================================
      shiny::tags$div(id = "sec_context"),
      shiny::div(class = "section-header",
        onclick = "toggleSection('content_context')",
        shiny::span(id = "content_context_arrow", "\u25BC"),
        " Context"
      ),
      shiny::div(id = "content_context", class = "section-content",
        shiny::p(style = "font-size: 13px; color: #555; margin-bottom: 8px;",
          "Neighbouring lines are shown here (up to 2 before and after).",
          "Curly braces mark sign sequences that have been identified as n-grams.",
          "The current line is highlighted in bold."
        ),
        shiny::uiOutput("context_text")
      ),

      shiny::hr(),

      # =============================================================
      # Chapter 3: Grammar (collapsible)
      # =============================================================
      shiny::tags$div(id = "sec_grammar"),
      shiny::div(class = "section-header",
        onclick = "toggleSection('content_grammar')",
        shiny::span(id = "content_grammar_arrow", "\u25BC"),
        " Grammar Probabilities"
      ),
      shiny::div(id = "content_grammar", class = "section-content",
        shiny::p(style = "font-size: 13px; color: #555; margin-bottom: 8px;",
          "The chart shows grammar probabilities for each sign in the line.",
          "Tall bars indicate a likely grammatical function",
          "(e.g. verb, noun, case marker). The position of a token in the sentence",
          " is currently not taken into account.."
        ),
        shiny::plotOutput("grammar_plot", height = "400px")
      ),

      shiny::hr(),

      # =============================================================
      # Chapter 4: Translation (Dictionary & Skeleton)
      # =============================================================
      shiny::tags$div(id = "sec_translation"),
      shiny::h3("Translation"),
      shiny::p(style = "font-size: 13px; color: #555; margin-bottom: 8px;",
        "Select a skeleton entry with the green book-button (\U{1F4D6}) to display",
        "matching dictionary entries below.",
        "Click a dictionary row to adopt its type and translation,",
        "or enter the translation manually.",
        "Use the brown button (\U{120FB}) to compose an entry from its components.",
		"The composition must be finished by hand (add articles and conjugate verbs).",
        "Use \u201cUpdate Skeleton\u201d after changing the bracket structure",
        "without losing existing translations."
      ),

      # Section 4.1: Bracket input + Update button
      shiny::tags$div(id = "bracket_area"),
      shiny::fluidRow(
        shiny::column(9,
          shiny::textInput("bracket_input", NULL, value = x_raw, width = "100%")
        ),
        shiny::column(3,
          shiny::actionButton("btn_generate", "Update Skeleton",
                              class = "btn-update")
        )
      ),

      # Section 4.3: Interactive Skeleton
      shiny::uiOutput("skeleton_ui"),

      shiny::hr(),

      # Section 4.4: Dictionary Lookup Display
      shiny::uiOutput("dict_section"),

      # --- Done button at bottom ---
      shiny::div(class = "done-bar",
        shiny::actionButton("btn_done", "Done", class = "btn-done"),
        shiny::actionButton("btn_cancel", "Cancel", class = "btn-cancel")
      )
    )
  )

  # ===================================================================
  # Server
  # ===================================================================
  server <- function(input, output, session) {

    # --- Reactive values ---
    rv <- shiny::reactiveValues(
      selected = 1L,
      df       = df_init,
      rebuild  = 0L
    )

    obs_handles <- list()

    # ---------------------------------------------------------------
    # Chapter 1: Merged N-gram table (patterns + shared)
    # ---------------------------------------------------------------
    output$ngram_merged_table <- shiny::renderUI({
      # 1. Start with ngram_in_line (full-text, length >= 2, in current line)
      merged_combos <- ngram_in_line$combination

      # 2. Compute shared n-grams from context if available
      shared_combos <- character(0)
      if (!is.na(line_index) && !is.null(text)) {
        idx_range       <- max(1, line_index - 2):min(length(text), line_index + 2)
        neighbour_lines <- text[setdiff(idx_range, line_index)]
        local_text      <- c(text[line_index], neighbour_lines)
        ngram_local     <- ngram_frequencies(local_text, min_freq = 2, mapping = mapping)

        current_str    <- paste0(token, collapse = "")
        neighbour_strs <- vapply(neighbour_lines, function(l) {
          paste0(split_sumerian(as.cuneiform(l, mapping = mapping))$signs,
                 collapse = "")
        }, character(1))

        ngram_shared <- ngram_local[vapply(ngram_local$combination, function(c) {
          grepl(c, current_str, fixed = TRUE) &&
            any(vapply(neighbour_strs, function(ns) grepl(c, ns, fixed = TRUE), logical(1)))
        }, logical(1)), , drop = FALSE]

        shared_combos <- ngram_shared$combination
        new_combos    <- setdiff(shared_combos, merged_combos)
        merged_combos <- c(merged_combos, new_combos)
      }

      if (length(merged_combos) == 0) {
        return(shiny::p(style = "color:#999; font-style:italic;",
                        "No n-gram patterns found."))
      }

      # 3. Build merged data frame with frequency from ngram_full
      freq_lookup <- stats::setNames(ngram_full$frequency, ngram_full$combination)
      len_lookup  <- stats::setNames(ngram_full$length, ngram_full$combination)

      merged_df <- data.frame(
        combination = merged_combos,
        frequency   = as.integer(freq_lookup[merged_combos]),
        length      = as.integer(len_lookup[merged_combos]),
        stringsAsFactors = FALSE
      )

      # For shared n-grams not in ngram_full: compute length from tokens
      na_len <- is.na(merged_df$length)
      if (any(na_len)) {
        merged_df$length[na_len] <- vapply(
          merged_df$combination[na_len],
          function(c) length(split_sumerian(c)$signs),
          integer(1))
      }

      # 4. Add sign_names and theme columns
      merged_df$sign_names <- as.character(
        as.sign_name(merged_df$combination, mapping = mapping))
      merged_df$theme <- ifelse(
        merged_df$combination %in% shared_combos, "\u2713", "")

      # 5. Sort by length descending, then frequency descending
      merged_df <- merged_df[order(
        -merged_df$length,
        -ifelse(is.na(merged_df$frequency), -Inf, merged_df$frequency)), ]

      # 6. Build HTML table
      tbl_rows <- lapply(seq_len(nrow(merged_df)), function(r) {
        shiny::tags$tr(
          shiny::tags$td(
            if (!is.na(merged_df$frequency[r]))
              as.character(merged_df$frequency[r]) else ""),
          shiny::tags$td(as.character(merged_df$length[r])),
          shiny::tags$td(merged_df$sign_names[r]),
          shiny::tags$td(merged_df$combination[r]),
          shiny::tags$td(style = "text-align: center;", merged_df$theme[r])
        )
      })

      shiny::tags$table(class = "ngram-table",
        shiny::tags$thead(shiny::tags$tr(
          shiny::tags$th("Freq"),
          shiny::tags$th("Len"),
          shiny::tags$th("Sign names"),
          shiny::tags$th("Combination"),
          shiny::tags$th("Theme")
        )),
        shiny::tags$tbody(tbl_rows)
      )
    })

    # ---------------------------------------------------------------
    # Chapter 1b: Sign Combination Suggestions
    # ---------------------------------------------------------------
    output$combination_table <- shiny::renderUI({
      if (is.null(multi_sign_df) || nrow(multi_sign_df) == 0) {
        return(shiny::p(style = "color:#999; font-style:italic;",
                        "No multi-sign combination suggestions found."))
      }

      tbl_rows <- lapply(seq_len(nrow(multi_sign_df)), function(r) {
        shiny::tags$tr(
          shiny::tags$td(
            if (!is.na(multi_sign_df$freq[r]))
              as.character(multi_sign_df$freq[r]) else ""),
          shiny::tags$td(multi_sign_df$sign_names[r]),
          shiny::tags$td(multi_sign_df$combination[r]),
          shiny::tags$td(multi_sign_df$type_suggestions[r])
        )
      })

      shiny::tags$table(class = "ngram-table",
        shiny::tags$thead(shiny::tags$tr(
          shiny::tags$th("Freq"),
          shiny::tags$th("Sign names"),
          shiny::tags$th("Combination"),
          shiny::tags$th("Type suggestions")
        )),
        shiny::tags$tbody(tbl_rows)
      )
    })

    # ---------------------------------------------------------------
    # Chapter 2: Context
    # ---------------------------------------------------------------
    output$context_text <- shiny::renderUI({
      if (is.na(line_index) || is.null(text)) {
        return(shiny::p("No text context available."))
      }

      idx_range      <- max(1, line_index - 2):min(length(text), line_index + 2)
      context_lines  <- text[idx_range]
      context_marked <- mark_ngrams(context_lines, ngram_full, mapping = mapping)

      rel_pos <- line_index - min(idx_range) + 1
      context_html <- vapply(seq_along(context_marked), function(i) {
        line_esc <- gsub("&", "&amp;", context_marked[i], fixed = TRUE)
        line_esc <- gsub("<", "&lt;", line_esc, fixed = TRUE)
        line_esc <- gsub(">", "&gt;", line_esc, fixed = TRUE)
        if (i == rel_pos) {
          paste0("<b>", line_esc, "</b>")
        } else {
          line_esc
        }
      }, character(1))

      shiny::HTML(paste0(
        "<pre style='font-family: Segoe UI Historic, sans-serif; font-size: 14px;'>",
        paste(context_html, collapse = "\n"), "</pre>"))
    })

    # (shared n-grams are now part of the merged table in Chapter 1)

    # ---------------------------------------------------------------
    # Chapter 3: Grammar
    # ---------------------------------------------------------------
    output$grammar_plot <- shiny::renderPlot({
      plot_sign_grammar(gp, sign_names = FALSE, mapping = mapping)
    })

    # ---------------------------------------------------------------
    # harvest_fill
    # ---------------------------------------------------------------
    harvest_fill <- function() {
      current_df <- rv$df
      for (i in seq_len(nrow(current_df))) {
        fill$type[current_df$fill_idx[i]]        <<- null_to_empty(input[[paste0("type_", i)]])
        fill$translation[current_df$fill_idx[i]] <<- null_to_empty(input[[paste0("trans_", i)]])
      }
    }

    # ---------------------------------------------------------------
    # build_skeleton_ui
    # ---------------------------------------------------------------
    build_skeleton_ui <- function() {
      current_df <- rv$df

      for (h in obs_handles) h$destroy()
      obs_handles <<- list()

      # --- Helper: check if any dictionary has translations for an expression ---
      has_dict_entries <- function(expr) {
        sn <- as.character(as.sign_name(expr, mapping = mapping))
        any(vapply(dic_list, function(d) {
          any(d$sign_name == sn & d$row_type == "trans.")
        }, logical(1)))
      }

      # --- Header row (root = row 1, depth 0): cuneiform tokens ---
      header_type  <- null_to_empty(fill$type[current_df$fill_idx[1]])
      header_trans <- null_to_empty(fill$translation[current_df$fill_idx[1]])
      header_label <- paste0(token, collapse = "")

      # Header has children if there are any depth-1 entries
      header_has_children <- nrow(current_df) > 1 && any(current_df$depth == 1)
      header_has_dict <- has_dict_entries(current_df$expr[1])

      header_row <- shiny::tags$tr(class = "skel-header",
        shiny::tags$td(class = "skel-btn-cell",
          shiny::actionButton("btn_1", label = "\U{1F4D6}",
            class = "btn btn-skel",
            disabled = !header_has_dict),
          shiny::actionButton("comp_1", label = "\U{120FB}",
            class = "btn btn-skel-brown",
            disabled = !header_has_children)
        ),
        shiny::tags$td(class = "skel-label", header_label),
        shiny::tags$td(class = "type-input",
          shiny::textInput("type_1", label = NULL, value = header_type)
        ),
        shiny::tags$td(class = "trans-input",
          shiny::textAreaInput("trans_1", label = NULL, value = header_trans,
                               rows = if (N >= 4) 2 else 1)
        )
      )

      # --- Entry rows (rows 2..n, depth >= 1) ---
      entry_rows <- list()
      if (nrow(current_df) >= 2) for (i in 2:nrow(current_df)) {
        indent_px  <- (current_df$depth[i] - 1) * 30
        fill_type  <- null_to_empty(fill$type[current_df$fill_idx[i]])
        fill_trans <- null_to_empty(fill$translation[current_df$fill_idx[i]])
        eq         <- get_equation(current_df$expr[i])

        # Sentence-based background colour
        sent <- entry_sentence(current_df$start[i])
        bg_col <- sent_colours[((sent - 1) %% length(sent_colours)) + 1]

        # Spacer before depth-1 entries (with sentence background)
        if (current_df$depth[i] == 1) {
          entry_rows <- c(entry_rows, list(
            shiny::tags$tr(class = "skel-spacer",
              style = paste0("background-color:", bg_col, ";"),
              shiny::tags$td(colspan = "4")
            )
          ))
        }

        btn_id   <- paste0("btn_", i)
        comp_id  <- paste0("comp_", i)
        type_id  <- paste0("type_", i)
        trans_id <- paste0("trans_", i)

        # Check if this entry has children (depth+1 entries within its token range)
        e_start <- current_df$start[i]
        e_end   <- e_start + current_df$n_tokens[i] - 1
        e_depth <- current_df$depth[i]
        has_children <- any(current_df$depth == e_depth + 1 &
                            current_df$start >= e_start &
                            current_df$start + current_df$n_tokens - 1 <= e_end)

        entry_has_dict <- has_dict_entries(current_df$expr[i])
        eq_html <- format_equation_html(eq)

        row <- shiny::tags$tr(
          style = paste0("background-color:", bg_col, ";"),
          shiny::tags$td(class = "skel-btn-cell",
            shiny::actionButton(btn_id, label = "\U{1F4D6}",
              class = "btn btn-skel",
              disabled = !entry_has_dict),
            shiny::actionButton(comp_id, label = "\U{120FB}",
              class = "btn btn-skel-brown",
              disabled = !has_children)
          ),
          shiny::tags$td(
            shiny::div(style = paste0("padding-left:", indent_px, "px;"),
              eq_html
            )
          ),
          shiny::tags$td(class = "type-input",
            shiny::textInput(type_id, label = NULL, value = fill_type)
          ),
          shiny::tags$td(class = "trans-input",
            shiny::textAreaInput(trans_id, label = NULL, value = fill_trans,
                                 rows = if (current_df$n_tokens[i] >= 4) 2 else 1)
          )
        )
        entry_rows <- c(entry_rows, list(row))
      }

      # --- Create observers for green buttons (select) ---
      for (i in seq_len(nrow(current_df))) {
        local({
          idx <- i
          h <- shiny::observeEvent(input[[paste0("btn_", idx)]], {
            rv$selected <- idx
            session$sendCustomMessage("scrollTo", "dict_section")
          }, ignoreInit = TRUE)
          obs_handles[[length(obs_handles) + 1]] <<- h
        })
      }

      # --- Create observers for brown buttons (compose) ---
      for (i in seq_len(nrow(current_df))) {
        local({
          idx <- i
          h <- shiny::observeEvent(input[[paste0("comp_", idx)]], {
            harvest_fill()

            # Determine parent entry's token range and depth
            p_start <- current_df$start[idx]
            p_end   <- p_start + current_df$n_tokens[idx] - 1
            p_depth <- current_df$depth[idx]

            # Find direct children (depth+1 within token range)
            child_mask <- current_df$depth == p_depth + 1 &
                          current_df$start >= p_start &
                          current_df$start + current_df$n_tokens - 1 <= p_end
            child_rows <- which(child_mask)

            if (length(child_rows) == 0) return()

            # Build data frame: first row = parent, remaining rows = children
            compose_df <- data.frame(
              expr        = current_df$expr[c(idx, child_rows)],
              n_tokens    = current_df$n_tokens[c(idx, child_rows)],
              start       = current_df$start[c(idx, child_rows)],
              depth       = current_df$depth[c(idx, child_rows)],
              type        = vapply(c(idx, child_rows), function(j)
                null_to_empty(fill$type[current_df$fill_idx[j]]), character(1)),
              translation = vapply(c(idx, child_rows), function(j)
                null_to_empty(fill$translation[current_df$fill_idx[j]]), character(1)),
              stringsAsFactors = FALSE
            )

            # Call helper function
            result <- tryCatch(
              compose_skeleton_entry(compose_df),
              error = function(e) e$message
            )

            if (is.character(result)) {
              shiny::showNotification(result, type = "warning")
              return()
            }

            # Update type and translation for the parent entry (from row 1)
            shiny::updateTextInput(session, paste0("type_", idx), value = result$type[1])
            shiny::updateTextInput(session, paste0("trans_", idx), value = result$translation[1])
          }, ignoreInit = TRUE)
          obs_handles[[length(obs_handles) + 1]] <<- h
        })
      }

      shiny::tags$table(class = "skel-table", header_row, entry_rows)
    }

    # ---------------------------------------------------------------
    # Render skeleton UI
    # ---------------------------------------------------------------
    output$skeleton_ui <- shiny::renderUI({
      rv$rebuild
      build_skeleton_ui()
    })

    # ---------------------------------------------------------------
    # Update Skeleton button
    # ---------------------------------------------------------------
    shiny::observeEvent(input$btn_generate, {
      harvest_fill()

      s <- input$bracket_input

      # Validate brackets and tokens before proceeding
      err <- validate_bracket_input(s, token, mapping)
      if (!is.null(err)) {
        shiny::showNotification(err, type = "error")
        return()
      }

      s_cunei  <- as.cuneiform(s, mapping = mapping)

      s_marked <- tryCatch(
        mark_skeleton_entries(s_cunei),
        error = function(e) {
          shiny::showNotification(
            paste("Skeleton error:", e$message), type = "error")
          return(NULL)
        }
      )
      if (is.null(s_marked)) return()

      new_df   <- extract_skeleton_entries(s_marked)
      new_df$fill_idx <- substr_position(new_df$start, new_df$n_tokens, N)

      # Recompute sentence boundaries from updated input
      sentence_ends <<- detect_sentence_ends(s, mapping)
      sent_idx <- 1L
      se_ptr   <- 1L
      for (ti in seq_len(N)) {
        token_sentence[ti] <<- sent_idx
        if (se_ptr <= length(sentence_ends) && ti == sentence_ends[se_ptr]) {
          sent_idx <- sent_idx + 1L
          se_ptr   <- se_ptr + 1L
        }
      }
      n_sentences <<- max(token_sentence)

      rv$df       <- new_df
      rv$selected <- 1L
      rv$rebuild  <- rv$rebuild + 1L
    })

    # ---------------------------------------------------------------
    # Dictionary lookup: per-dictionary results
    # ---------------------------------------------------------------
    dict_by_dic <- shiny::reactive({
      sel <- rv$selected
      lookup_expr <- rv$df$expr[sel]
      lookup_sign_name <- as.character(as.sign_name(lookup_expr, mapping = mapping))

      ordered_names <- dic_names

      results <- lapply(ordered_names, function(dn) {
        d <- dic_list[[dn]]
        rows <- d[d$sign_name == lookup_sign_name & d$row_type == "trans.", , drop = FALSE]
        if (nrow(rows) > 0) {
          df <- data.frame(
            count   = if ("count" %in% names(rows)) rows$count else NA_integer_,
            type    = rows$type,
            meaning = rows$meaning,
            stringsAsFactors = FALSE
          )
          df[order(-ifelse(is.na(df$count), -Inf, df$count)), ]
        } else {
          data.frame(count = integer(0), type = character(0),
                     meaning = character(0), stringsAsFactors = FALSE)
        }
      })
      names(results) <- ordered_names
      results
    })

    # ---------------------------------------------------------------
    # Render dictionary section (one table per dictionary)
    # ---------------------------------------------------------------
    output$dict_section <- shiny::renderUI({
      results <- dict_by_dic()

      sel <- rv$selected
      expr <- rv$df$expr[sel]
      sn <- as.character(as.sign_name(expr, mapping = mapping))
      expr_clean <- gsub("[(){}]", "", expr)
      header <- shiny::h4(paste0("Dictionary entries for: ", sn, " (", expr_clean, ")"))

      if (length(results) == 0) {
        return(shiny::tagList(header,
          shiny::p(style = "color:#999; font-style:italic;", "No dictionaries selected.")))
      }

      tables <- lapply(seq_along(results), function(d_idx) {
        dn <- names(results)[d_idx]
        df <- results[[d_idx]]

        if (nrow(df) == 0) {
          return(shiny::div(style = "margin-bottom: 12px;",
            shiny::h5(dn),
            shiny::p(style = "color:#999; font-style:italic;", "No entries found.")
          ))
        }

        tbl_rows <- lapply(seq_len(nrow(df)), function(r) {
          click_js <- sprintf(
            "Shiny.setInputValue('dict_click', '%d_%d', {priority: 'event'})",
            d_idx, r)
          shiny::tags$tr(onclick = click_js,
            shiny::tags$td(
              if (!is.na(df$count[r])) as.character(df$count[r]) else ""),
            shiny::tags$td(df$type[r]),
            shiny::tags$td(df$meaning[r])
          )
        })

        tbl <- shiny::tags$table(class = "dict-table",
          shiny::tags$thead(shiny::tags$tr(
            shiny::tags$th("Count"),
            shiny::tags$th("Type"),
            shiny::tags$th("Meaning")
          )),
          shiny::tags$tbody(tbl_rows)
        )

        shiny::div(style = "margin-bottom: 12px;", shiny::h5(dn), tbl)
      })

      shiny::tagList(header, tables)
    })

    # ---------------------------------------------------------------
    # Click-to-adopt from dictionary tables
    # ---------------------------------------------------------------
    shiny::observeEvent(input$dict_click, {
      parts <- strsplit(input$dict_click, "_")[[1]]
      d_idx <- as.integer(parts[1])
      r_idx <- as.integer(parts[2])
      results <- dict_by_dic()
      if (d_idx >= 1 && d_idx <= length(results)) {
        df <- results[[d_idx]]
        if (r_idx >= 1 && r_idx <= nrow(df)) {
          i <- rv$selected
          shiny::updateTextInput(session, paste0("type_", i),
                                 value = null_to_empty(df$type[r_idx]))
          shiny::updateTextInput(session, paste0("trans_", i),
                                 value = null_to_empty(df$meaning[r_idx]))
          session$sendCustomMessage("scrollTo", "bracket_area")
        }
      }
    })

    # ---------------------------------------------------------------
    # Done button
    # ---------------------------------------------------------------
    shiny::observeEvent(input$btn_done, {
      harvest_fill()
      s_final    <- input$bracket_input
      fill_final <- fill
      result <- skeleton(s_final, mapping = mapping, fill = fill_final, space = TRUE)
      shiny::stopApp(result)
    })

    # ---------------------------------------------------------------
    # Cancel button
    # ---------------------------------------------------------------
    shiny::observeEvent(input$btn_cancel, {
      shiny::stopApp(NULL)
    })
  }

  # ===================================================================
  # Run the gadget
  # ===================================================================

  result <- shiny::runGadget(ui, server, viewer = viewer)


  if (is.null(result)) {
    message("Translation cancelled.")
    return(invisible(NULL))
  }

  result
}

merge_dictionaries <- function(...) {

  dics <- list(...)

  # Falls eine einzige Liste übergeben wurde, entpacken
  if (length(dics) == 1 && is.list(dics[[1]]) && !is.data.frame(dics[[1]]))
    dics <- dics[[1]]

  # Dateipfade laden
  for (i in seq_along(dics)) {
    if (is.character(dics[[i]]) && length(dics[[i]]) == 1)
      dics[[i]] <- read_dictionary(dics[[i]])
  }

  # Alle zusammenfügen
  all_df <- do.call(rbind, dics)

  # cunei./reading: erste Vorkommen behalten (= aus dem 1. Wörterbuch)
  meta <- all_df[all_df$row_type %in% c("cunei.", "reading"), ]
  meta <- meta[!duplicated(meta[, c("sign_name", "row_type")]), ]

  # trans.: Einträge mit gleicher sign_name + type + meaning verschmelzen
  trans <- all_df[all_df$row_type == "trans.", ]
  trans <- aggregate(
    list(count = trans$count),
    by = list(sign_name = trans$sign_name, type = trans$type,
              meaning = trans$meaning),
    FUN = sum, na.rm = TRUE
  )
  trans <- data.frame(sign_name = trans$sign_name, row_type = "trans.",
                      count = trans$count, type = trans$type,
                      meaning = trans$meaning)

  # Zusammenführen und sortieren
  result <- rbind(meta, trans)
  ord <- order(result$sign_name, result$row_type, -result$count)
  result <- result[ord, ]
  rownames(result) <- seq_len(nrow(result))

  result
}

translation_context <- function(project_dir, dic = NULL, text = NULL,
                                mapping = NULL, min_freq = c(6, 4, 2),
                                sentence_prob = 1.0) {

  line_folder <- file.path(project_dir, "lines")

  # Add project_dir to text if it is a filename without path
  if (!is.null(text) && is.character(text) && length(text) == 1 &&
      !grepl("[/\\\\]", text)) {
    text <- file.path(project_dir, text)
  }

  # Add project_dir to dic entries that are filenames without path
  if (is.character(dic)) {
    no_path <- !grepl("[/\\\\]", dic)
    dic[no_path] <- file.path(project_dir, dic[no_path])
  }

  structure(
    list(
      project_dir   = project_dir,
      line_folder   = line_folder,
      dic           = dic,
      text          = text,
      mapping       = mapping,
      min_freq      = min_freq,
      sentence_prob = sentence_prob
    ),
    class = "translation_context"
  )
}

# --- Shared constants and helpers ---
DIGITS <- as.character(0:9)

# Read a #N tag starting at position i in character vector chars.
# Returns list(tag, tag_num, next_i).
read_tag <- function(chars, i) {
  tag <- "#"
  i <- i + 1L
  while (i <= length(chars) && chars[i] %in% DIGITS) {
    tag <- paste0(tag, chars[i])
    i <- i + 1L
  }
  list(tag = tag, tag_num = as.integer(sub("#", "", tag)), next_i = i)
}

# --- Helper: parse a type string into a structured list ---
parse_type <- function(type_str) {
  if (is.na(type_str) || type_str == "") {
    return(list(is_operator = FALSE, word_type = NA_character_))
  }
  s <- type_str
  s <- gsub("x", "\u2612", s, fixed = TRUE)
  s <- gsub("->", "\u2192", s, fixed = TRUE)

  if (!grepl("\u2612", s, fixed = TRUE)) {
    return(list(is_operator = FALSE, word_type = s))
  }

  parts <- strsplit(s, "\u2612", fixed = TRUE)[[1]]
  left_arg <- parts[1]

  arrow_parts <- strsplit(parts[2], "\u2192", fixed = TRUE)[[1]]
  right_arg <- arrow_parts[1]
  result    <- arrow_parts[2]

  list(
    is_operator = TRUE,
    left_arg    = left_arg,
    right_arg   = right_arg,
    result      = result
  )
}

# --- Helper: determine the result type of a parsed type ---
get_result_type <- function(parsed) {
  if (parsed$is_operator) parsed$result else parsed$word_type
}

# --- Helper: tokenize a concatenated type string into type tokens ---
# Parses e.g. "SSEN" into c("S", "SEN"), longest match first.
tokenize_types <- function(s) {
  if (nchar(s) == 0) return(character(0))
  known <- c("SEN", "S", "A", "V")
  tokens <- character(0)
  pos <- 1L
  n <- nchar(s)
  while (pos <= n) {
    matched <- FALSE
    for (k in known) {
      end <- pos + nchar(k) - 1L
      if (end <= n && substr(s, pos, end) == k) {
        tokens <- c(tokens, k)
        pos <- end + 1L
        matched <- TRUE
        break
      }
    }
    if (!matched)
      stop(paste0("Unknown type token at position ", pos, " in '", s, "'"))
  }
  tokens
}

# --- Shared rule definitions ---

# Extract the base type of a type string (uppercase prefix).
# E.g. "Vt" -> "V", "So" -> "S", "SEN" -> "SEN", "Vi" -> "V".
base_type <- function(x) {
  sub("[a-z]+$", "", x)
}

# Check whether x has base type y (i.e. the uppercase prefix of x equals y).
has_base_type <- function(x, y) {
  if (is.na(x)) return(FALSE)
  identical(base_type(x), y)
}

# Check if a type matches an expected type (NA-safe).
# If expected is a pure base type (no lowercase suffix), any extension
# of that base type matches (e.g. "Vt" matches "V").
# If expected is an extended type (has lowercase suffix), only an exact
# match counts (e.g. "Vt" matches "Vt" but "V" does not).
type_matches <- function(actual, expected) {
  if (is.na(actual) || is.na(expected)) return(FALSE)
  if (grepl("[a-z]", expected)) {
    actual == expected
  } else {
    has_base_type(actual, expected)
  }
}

# Composition rules: (type1, type2) -> result.
# Used by determine_group_type and add_brackets.
# Note: add_brackets adds separator conditions to S+S rules.
# More specific rules (extended types) must precede general rules.
COMP_RULES <- list(
  list(t1 = "S",   t2 = "A",   result = "S"),
  list(t1 = "S",   t2 = "S",   result = "S"),
  list(t1 = "S",   t2 = "Vt",  result = "Vi"),
  list(t1 = "S",   t2 = "Vi",  result = "SEN"),
  list(t1 = "S",   t2 = "V",   result = "SEN"),
  list(t1 = "SEN", t2 = "SEN", result = "SEN")
)


# --- Helper: build named args list from elements ---
#
# Derives placeholder names from the element base types.  Base types
# appearing once get a plain name (S, A, V, SEN); base types appearing
# more than once get indexed names (S1, S2, ...).
#
# Each arg is c(name, translation, original_type).  The original
# (possibly extended) type is preserved for type inheritance in
# eval_operator.
#
# @param elements List of c(type_str, translation) vectors
# @return List of c(name, translation, original_type) vectors
build_args <- function(elements) {
  types      <- vapply(elements, `[`, character(1), 1)
  base_types <- vapply(types, base_type, character(1), USE.NAMES = FALSE)
  freq       <- table(base_types)
  counter    <- integer(0)
  args       <- vector("list", length(elements))
  for (i in seq_along(elements)) {
    tok <- base_types[i]
    if (freq[[tok]] > 1L) {
      counter[tok] <- if (is.na(counter[tok])) 1L else counter[tok] + 1L
      name <- paste0(tok, counter[tok])
    } else {
      name <- tok
    }
    args[[i]] <- c(name, elements[[i]][2], types[i])
  }
  args
}


compose_skeleton_entry <- function(df) {
  n <- nrow(df)
  if (n < 2) return("No children to compose")

  for (i in 2:n) {
    if (is.na(df$type[i]) || df$type[i] == "")
      return(paste0("Missing type for: ", df$expr[i]))
    if (is.na(df$translation[i]) || df$translation[i] == "")
      return(paste0("Missing translation for: ", df$expr[i]))
  }

  # Step a) Build structure string
  structure_str <- df$expr[1]
  for (i in 2:n) {
    structure_str <- sub(df$expr[i], paste0("#", i - 1), structure_str, fixed = TRUE)
  }
  # Remove token-group wrappers: <> angle brackets (from raw input) and
  # {} curly braces that wrap a single #N tag (artifact of <token> -> {}
  # conversion in mark_skeleton_entries/extract_skeleton_entries pipeline).
  # Multi-child groups like {#1#2} are preserved (user-defined groupings).
  structure_str <- gsub("[<>]", "", structure_str)
  structure_str <- gsub("\\{(#[0-9]+)\\}", "\\1", structure_str)

  # Child type and translation vectors
  child_type <- df$type[-1]
  translations <- df$translation[-1]

  # Validate: operator types must have matching placeholder in translation
  for (i in seq_along(child_type)) {
    parsed <- parse_type(child_type[i])
    if (!parsed$is_operator) next

    raw <- translations[i]
    m <- regmatches(raw, regexpr("(?<=\\{)[^}]+(?=\\})", raw, perl = TRUE))
    check_trans <- if (length(m) > 0 && nchar(m) > 0) m else raw

    # Collect expected placeholder types from operator arguments
    placeholders <- unique(c(
      tokenize_types(parsed$left_arg),
      tokenize_types(parsed$right_arg)
    ))

    found <- any(vapply(placeholders, function(ph)
      grepl(paste0("\\b", ph, "[0-9]*\\b"), check_trans), logical(1)))

    if (!found) {
      return(paste0("Missing placeholder in translation of ", df$expr[i + 1],
                     " (", child_type[i], "): expected '",
                     paste(unique(placeholders), collapse = "' or '"),
                     "' in '", raw, "'"))
    }
  }

  # Validate: Vt translations must contain placeholder S
  for (i in seq_along(child_type)) {
    if (!identical(child_type[i], "Vt")) next
    raw <- translations[i]
    m <- regmatches(raw, regexpr("(?<=\\{)[^}]+(?=\\})", raw, perl = TRUE))
    check_trans <- if (length(m) > 0 && nchar(m) > 0) m else raw
    if (!grepl("\\bS\\b", check_trans)) {
      return(paste0("Missing placeholder in translation of ", df$expr[i + 1],
                     " (Vt): expected 'S' in '", raw, "'"))
    }
  }

  # Step b) Add brackets
  res_b <- add_brackets(structure_str, child_type)
  if (identical(res_b$type, "ERROR")) return(res_b$string)

  # Step c) Apply translation rules
  result <- apply_translation_rules(res_b$string, child_type, translations)

  if (length(result) == 1) return(result)

  df$type[1]        <- result[1]
  df$translation[1] <- result[2]
  df
}

#Legend of utf8-Signs:

#\u011d  g with circumflex
#\u011c  G with circumflex
#\u0161  s with caron
#\u0160  S with caron
#\u1e2b  h with breve below
#\u1e2a  H with breve below

#\u014B  some kind of g
#\u1E63  s with dot below
#\u1E6D  t with dot below


#\u00e1  a acute
#\u00e9  e acute
#\u00ed  i acute
#\u00fa  u acute

#\u00e0  a gravis
#\u00e8  e gravis
#\u00ec  i gravis
#\u00f9  u gravis

#\u00d7  operator: cross
#\u00b7  operator: cdot
#\u00a4  operator: of
#\u2218  operator: chaining

#\u2205  emptyset
#\u2192  arrow
#\u2612  checkbox


standardize_signs <- function(x){
  x <- str_replace(x, "^<(.*)>$", "\\1")

  ## Only relevant for signs of type 1
  x <- str_replace_all(x, "c", "\u0161")
  x <- str_replace_all(x, "j", "\u011d")
  x <- str_replace_all(x, "h", "\u1e2b")

  x <- str_replace_all(x, "\u014B", "\u011d")
  x <- str_replace_all(x, "\u1E63", "s")
  x <- str_replace_all(x, "\u1E6D", "t")
  x <- str_replace_all(x, "\u015B", "\u0161")

  is_variant_2 <- str_detect(x, "\u00e1|\u00e9|\u00ed|\u00fa")
  is_variant_3 <- str_detect(x, "\u00e0|\u00e8|\u00ec|\u00f9")

  x <- str_replace_all(x, "\u00e1", "a")
  x <- str_replace_all(x, "\u00e9", "e")
  x <- str_replace_all(x, "\u00ed", "i")
  x <- str_replace_all(x, "\u00fa", "u")
  x <- str_replace_all(x, "\u00e0", "a")
  x <- str_replace_all(x, "\u00e8", "e")
  x <- str_replace_all(x, "\u00ec", "i")
  x <- str_replace_all(x, "\u00f9", "u")

  x <- ifelse(is_variant_2, paste0(x, "2"), x)
  x <- ifelse(is_variant_3, paste0(x, "3"), x)

  ## Only relevant for signs of type 2

  x <- str_replace_all(x, "PLUS",     "+")
  x <- str_replace_all(x, "TIMES", "\u00d7")
  x <- str_replace_all(x, "OVER",     "/")
  x <- str_replace_all(x, "CROSSING", "%")
  x <- str_replace_all(x, "TENU",     "tenu")
  x <- str_replace_all(x, "GUNU",     "gunu")
  x <- str_replace_all(x, "SHESHIG", "\u0161e\u0161ig")
  x <- str_replace_all(x, "\u0160E\u0160IG", "\u0161e\u0161ig")
  x <- str_replace_all(x, "SESIG", "\u0161e\u0161ig")
  x <- str_replace_all(x, ".OPPOSING.", "\u00b7OPPOSING\u00b7")
  x <- str_replace_all(x, ".INVERTED", "\u00b7INVERTED")
  x <- str_replace_all(x, "ONE.", "ONE\u00b7")
  x <- str_replace_all(x, "TWO.", "TWO\u00b7")
  x <- str_replace_all(x, "THREE.", "THREE\u00b7")
  x <- str_replace_all(x, "FOUR.", "FOUR\u00b7")
  x <- str_replace_all(x, "FIVE.", "FIVE\u00b7")

  x <- str_replace_all(x, " ", "")

  return(x)
}

.onAttach <- function(libname, pkgname) {
  msg <- ""

  if (interactive()) {
    backend <- getOption("RStudioGD.backend", default = "default")
    if (!identical(backend, "ragg")) {
      msg <- paste0(
        "Note: The graphics backend may not be set to AGG, so cuneiform characters may not render correctly. For best results:\n",
        "  - install.packages(\"ragg\")\n",
        "  - In RStudio: Tools > Global Options > General > Graphics > ",
        "Backend > select \"AGG\"\n"
      )
    }
  }

  if(msg==""){
    msg <- switch(sample.int(2L, 1L),
        "Watch for dictionary updates:\n    https://founder-hypothesis.com/en/sumerian-mythology/downloads/",
        "Type 'browseVignettes(\"sumer\")' to get started.")
  }

  packageStartupMessage(msg)
}

plot_sign_grammar <- function(sg,
                              output_file = NULL,
                              width       = 10,
                              height      = 5,
                              sign_names  = FALSE,
                              font_family = NULL,
                              mapping     = NULL) {

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required. Please install it.")
  }

  # --- Detect input format ---
  has_prob <- "prob" %in% names(sg)
  has_n    <- "n"    %in% names(sg)

  if (has_prob) {
    sg$pct <- sg$prob * 100
  } else if (has_n) {
    pos_totals <- tapply(sg$n, sg$position, sum)
    sg$pct <- vapply(seq_len(nrow(sg)), function(i) {
      total <- pos_totals[as.character(sg$position[i])]
      if (is.na(total) || total == 0L) 0 else 100 * sg$n[i] / total
    }, numeric(1))
  } else {
    stop("Input must have column 'n' (from sign_grammar) or 'prob' (from grammar_probs).")
  }

  # --- Detect whether the graphics device supports Unicode + custom fonts ---
  backend   <- getOption("RStudioGD.backend", default = "default")
  knitr_dev <- if (isTRUE(getOption("knitr.in.progress"))) {
    knitr::opts_chunk$get("dev")
  } else NULL
  uses_ragg <- identical(backend, "ragg") ||
    grepl("^ragg", knitr_dev %||% "") ||
    !is.null(output_file)

  # --- X-axis labels: cuneiform characters or sign names ---
  label_df     <- sg[!duplicated(sg$position), c("position", "cuneiform")]
  label_family <- if (is.null(font_family)) "Segoe UI Historic" else font_family

  # --- Fallback for devices that cannot render Unicode / custom fonts ---
  ascii_fallback <- !uses_ragg & is.null(font_family)
  if (sign_names | ascii_fallback) {
    label_df$cuneiform <- as.sign_name(label_df$cuneiform, mapping = mapping)
    label_df$cuneiform <- str_replace_all(label_df$cuneiform, "\u27E8", "<")
    label_df$cuneiform <- str_replace_all(label_df$cuneiform, "\u27E9", ">")
    label_family <- "sans"
  }

  x_labels <- setNames(label_df$cuneiform, label_df$position)

  # --- Helper: replace Unicode operators with ASCII equivalents ---
  ascii_types <- function(x) {
    x <- gsub("\u2612", "x", x)
    x <- gsub("\u2192", "->", x)
    x
  }

  # --- Fixed colour and order for every grammar type ---
  type_colour_table <- c(
    "\u2612S\u2192V"      = "#C0392B",
    "\u2612V\u2192V"      = "#FF6B6B",
    "S\u2612\u2192V"      = "#D35400",
    "S\u2612V\u2192V"     = "#CB4335",
    "V\u2192V"            = "#A93226",
    "V\u2612\u2192V"      = "#922B21",
    "V"                   = "#E74C3C",
    "Vt"                  = "#E74C3C",
    "SEN"                 = "#95A5A6",
    "\u2612S\u2192S"      = "#BDC3C7",
    "\u2612SEN\u2192SEN"  = "#AEB6BF",
    "S\u2612S\u2192S"     = "#808B96",
    "S\u2612V\u2192S"     = "#B2BABB",
    "SEN\u2612\u2192S"    = "#CCD1D1",
    "A"                   = "#2471A3",
    "A\u2612\u2192A"      = "#2E86C1",
    "\u2612S\u2192A"      = "#3498DB",
    "S\u2612\u2192A"      = "#2980B9",
    "S\u2612S\u2192A"     = "#5DADE2",
    "S\u2612SEN\u2192A"   = "#85C1E9",
    "V\u2612\u2192A"      = "#AED6F1",
    "S\u2612\u2192S"      = "#F39C12",
    "S"                   = "#2ECC71"
  )

  if (ascii_fallback) {
    names(type_colour_table) <- ascii_types(names(type_colour_table))
    sg$type <- ascii_types(sg$type)
  }

  fixed_order  <- names(type_colour_table)
  type_colours <- type_colour_table

  all_types <- unique(sg$type)
  unknown   <- setdiff(all_types, fixed_order)
  if (length(unknown) > 0L) {
    fallback <- "#EAECEE"
    for (u in unknown) type_colours[u] <- fallback
    pos <- which(fixed_order == "V")
    fixed_order <- c(fixed_order[seq_len(pos - 1L)],
                     sort(unknown),
                     fixed_order[pos:length(fixed_order)])
  }

  type_order <- fixed_order[fixed_order %in% all_types]
  sg$type    <- factor(sg$type, levels = type_order)

  # --- Sample size per position ---
  sample_sizes <- tapply(sg$n, sg$position, sum)
  sample_df <- data.frame(
    position = names(sample_sizes),
    N        = as.integer(sample_sizes),
    stringsAsFactors = FALSE
  )

  all_positions <- sort(unique(sg$position))
  sg$position        <- factor(sg$position, levels = all_positions)
  sample_df$position <- factor(sample_df$position, levels = all_positions)

  sg <- sg[sg$pct > 0, ]

  # --- Build plot ---
  p <- ggplot2::ggplot(sg,
                       ggplot2::aes(x = .data$position,
                                    y = .data$pct,
                                    fill = .data$type)) +
    ggplot2::geom_col(width = 0.7) +
    ggplot2::geom_text(data    = sample_df,
                       mapping = ggplot2::aes(x = .data$position,
                                              y = 102,
                                              label = paste0("n=", .data$N)),
                       inherit.aes = FALSE,
                       size    = 3.2,
                       colour  = "grey40") +
    ggplot2::scale_fill_manual(values = type_colours, drop = FALSE) +
    ggplot2::scale_x_discrete(labels = x_labels, drop = FALSE) +
    ggplot2::scale_y_continuous(labels = function(x) paste0(x, "%"),
                                expand = ggplot2::expansion(mult = c(0, 0.08))) +
    ggplot2::labs(x    = NULL,
                  y    = "Relative frequency",
                  fill = "Grammar type") +
    ggplot2::theme_minimal(base_size = 13, base_family = "sans") +
    ggplot2::theme(
      axis.text.x         = ggplot2::element_text(family = label_family, size = 16),
      legend.position      = "top",
      panel.grid.major.x   = ggplot2::element_blank()
    )

  # --- Output ---------------------------------------------------------------
  if (!is.null(output_file)) {
    ext <- tolower(tools::file_ext(output_file))

    if (ext == "png") {
      ragg::agg_png(output_file, width = width, height = height,
                      units = "in", res = 150)
    } else if (ext %in% c("jpg", "jpeg")) {
        ragg::agg_jpeg(output_file, width = width, height = height,
                       units = "in", res = 150, quality = 95)
    } else {
      stop("Unknown file extension: ", ext, " (allowed: .png, .jpg, .jpeg)")
    }

    print(p)
    grDevices::dev.off()
    message("Plot saved: ", output_file)

  } else {
    print(p)
  }

  invisible(p)
}

guess_substr_info <- function(x, dic, mapping = NULL) {

  if (is.null(mapping)) {
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep = ";", na.strings = "")
  }

  # Convert dic to a list of dictionaries
  if (is.character(dic)) {
    dic <- lapply(dic, read_dictionary, verbose = FALSE)
  } else if (is.data.frame(dic)) {
    dic <- list(dic)
  }

  # Extract tokens
  token <- split_sumerian(as.cuneiform(x, mapping = mapping))$signs
  N     <- length(token)

  # Initialize empty substring data frame
  df <- init_substr_info(token)

  # Add sign name column
  df$sign_name <- as.sign_name(df$expr, mapping = mapping)

  # For each row: look up the most frequent translation
  for (i in seq_len(nrow(df))) {
    for (d in dic) {
      rows <- d[d$sign_name == df$sign_name[i] & d$row_type == "trans.", , drop = FALSE]
      if (nrow(rows) > 0) {
        if (all(is.na(rows$count))) {
          best <- rows[1, ]
        } else {
          best <- rows[which.max(rows$count), ]
        }
        df$type[i]        <- best$type
        df$translation[i] <- best$meaning
        break
      }
    }
  }

  # Zahlen und N: type -> "S", translation -> Wert aus expr
  sp <- split_sumerian(as.cuneiform(x, mapping = mapping))
  for (i in seq_len(nrow(df))) {
    if (df$n_tokens[i] != 1L) next
    tok_type <- sp$types[df$start[i]]
    if (tok_type == 4L) {
      df$type[i]        <- "S"
      df$translation[i] <- sp$signs[df$start[i]]
    }
  }

  df
}

grammar_probs <- function(sg, prior, dic, alpha0 = 1) {

  if (alpha0 < 0) stop("alpha0 must be >= 0.")

  sentence_prob <- attr(prior, "sentence_prob")
  if (is.null(sentence_prob)) {
    stop("prior must have a 'sentence_prob' attribute (from prior_probs()).")
  }

  # --- Grammar types from the prior ---
  all_types <- names(prior)
  K <- length(all_types)

  # --- Identify verb-like types for correction ---
  is_verb <- vapply(all_types, function(type_str) {
    if (type_str == "V") return(TRUE)
    is_op <- grepl("\u2192", type_str) || grepl("->", type_str, fixed = TRUE)
    if (!is_op) return(FALSE)
    ret <- sub(".*(\u2192|->)\\s*", "", type_str)
    trimws(ret) == "V"
  }, logical(1))

  x <- ifelse(is_verb, 1 / sentence_prob, 1)

  # --- Dictionary: trans. entries for single signs ---
  trans <- dic[dic$row_type == "trans.", ]

  # --- Unique positions from sign_grammar output ---
  positions <- sort(unique(sg$position))
  # sign_name and cuneiform per position
  pos_info <- sg[!duplicated(sg$position), c("position", "sign_name", "cuneiform")]

  # --- Build result: one row per position per grammar type ---
  rows <- vector("list", length(positions) * K)
  idx <- 0L

  for (pos in positions) {
    info <- pos_info[pos_info$position == pos, ]
    s <- info$sign_name

    # Raw counts n[k] from dictionary
    entries <- trans[trans$sign_name == s, ]
    entries <- entries[!is.na(entries$type) & trimws(entries$type) != "", ]

    n_k <- setNames(numeric(K), all_types)
    if (nrow(entries) > 0L) {
      for (j in seq_len(nrow(entries))) {
        t <- trimws(entries$type[j])
        if (t %in% all_types) {
          cnt <- if (is.na(entries$count[j])) 1L else entries$count[j]
          n_k[t] <- n_k[t] + cnt
        }
      }
    }

    # Corrected counts m[k] = n[k] * x[k]
    m_k <- n_k * x
    M <- sum(m_k)

    # Posterior: theta[k] = (alpha0 * p[k] + m[k]) / (alpha0 + M)
    theta <- (alpha0 * prior + m_k) / (alpha0 + M)

    for (ki in seq_len(K)) {
      idx <- idx + 1L
      rows[[idx]] <- data.frame(
        position   = pos,
        sign_name  = s,
        cuneiform  = info$cuneiform,
        type       = all_types[ki],
        prob       = theta[ki],
        n          = as.integer(n_k[ki]),
        stringsAsFactors = FALSE,
        row.names  = NULL
      )
    }
  }

  out <- do.call(rbind, rows[seq_len(idx)])
  rownames(out) <- NULL
  out
}

make_dictionary<- function(file, mapping = NULL){

  if(!is.character(file))stop("file must be a character-vector")

  if(is.null(mapping)){
    path  <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
  }

  df <- read_translated_text(file, mapping)
  df <- beautify_translations(df)
  df <- convert_to_dictionary(df, mapping)

  return(df)
}

# --- determine_group_type: determine the result type of a bracket group ---
# The bracket structure is already set by add_brackets, so this function
# only needs to look at the elements inside a single group:
#   - If an operator is present, return its result type.
#   - Otherwise, apply COMP_RULES to determine the composition type.
determine_group_type <- function(inner, type) {
  SEN_OPEN <- "\u00ab"
  VT_OPEN  <- "\u2039"
  chars <- strsplit(inner, "")[[1]]
  types <- character(0)
  op_result <- NULL
  i <- 1L

  while (i <= length(chars)) {
    ch <- chars[i]

    if (ch == "#") {
      rt <- read_tag(chars, i)
      i <- rt$next_i
      parsed <- parse_type(type[rt$tag_num])
      # If this element is an operator, remember result but continue
      # to collect argument types for type inheritance
      if (parsed$is_operator) {
        op_result <- parsed$result
      } else {
        types <- c(types, parsed$word_type)
      }

    } else if (ch %in% c("(", "<", "[", "{") || ch == SEN_OPEN || ch == VT_OPEN) {
      bracket_type <- switch(ch,
        "(" = "S", "<" = "V", "\u2039" = "Vt", "[" = "A", "{" = "SEN", "SEN")
      close <- switch(ch,
        "(" = ")", "<" = ">", "\u2039" = "\u203a", "[" = "]", "{" = "}", "\u00bb")
      depth <- 1L
      i <- i + 1L
      while (i <= length(chars) && depth > 0) {
        if (chars[i] == ch) depth <- depth + 1L
        if (chars[i] == close) depth <- depth - 1L
        i <- i + 1L
      }
      types <- c(types, bracket_type)

    } else {
      i <- i + 1L
    }
  }

  # Operator with type inheritance
  if (!is.null(op_result)) {
    if (!grepl("[a-z]", op_result)) {
      for (t in types) {
        if (has_base_type(t, op_result) && t != op_result) return(t)
      }
    }
    return(op_result)
  }

  if (length(types) == 0) return("S")
  if (length(types) == 1) return(types[1])

  # No operator: apply composition rules (from COMP_RULES)
  if (length(types) == 2) {
    for (cr in COMP_RULES) {
      if (type_matches(types[1], cr$t1) && type_matches(types[2], cr$t2)) return(cr$result)
    }
  }

  # Fallback
  "SEN"
}


grammatical_structure <- function(s, type, expr = NULL) {
  SEN_OPEN  <- "\u00ab"
  SEN_CLOSE <- "\u00bb"
  VT_OPEN   <- "\u2039"
  VT_CLOSE  <- "\u203a"

  # Step 1: Build structure string by replacing expr[k] with #k
  if (!is.null(expr)) {
    for (k in seq_along(expr)) {
      s <- sub(expr[k], paste0("#", k), s, fixed = TRUE)
    }
  }

  s <- str_replace_all(s, "-","")
  s <- str_replace_all(s, fixed("."),"")
  s <- str_replace_all(s, " ","")

  # Step 2: Add brackets (grouping by operator binding and composition rules)
  res <- add_brackets(s, type)
  if (identical(res$type, "ERROR")) return(res$string)
  s <- res$string

  # Step 3: Replace {} with typed brackets
  while (grepl("\\{[^{}]+\\}", s)) {
    m <- regexpr("\\{[^{}]+\\}", s)
    match_str <- regmatches(s, m)
    inner <- substring(match_str, 2, nchar(match_str) - 1)
    group_type <- determine_group_type(inner, type)
    group_base <- base_type(group_type)
    brackets <- if (group_type == "Vt") {
      c(VT_OPEN, VT_CLOSE)
    } else {
      switch(group_base,
        "S" = c("(", ")"),
        "V" = c("<", ">"),
        "A" = c("[", "]"),
        c(SEN_OPEN, SEN_CLOSE))
    }
    regmatches(s, m) <- paste0(brackets[1], inner, brackets[2])
  }

  s <- gsub(SEN_OPEN, "{", s, fixed = TRUE)
  s <- gsub(SEN_CLOSE, "}", s, fixed = TRUE)
  s <- gsub(VT_OPEN, "<", s, fixed = TRUE)
  s <- gsub(VT_CLOSE, ">", s, fixed = TRUE)

  # Step 4: Replace #N tags with expressions
  if (!is.null(expr)) {
    for (k in seq_along(expr)) {
      s <- sub(paste0("#", k), expr[k], s, fixed = TRUE)
    }
  }

  class(s) <- "grammatical_structure"
  s
}


#' @exportS3Method print grammatical_structure
print.grammatical_structure <- function(x, ...) {

  s <- unclass(x)

  # --- Vivid color styles ---
  col_green <- cli::make_ansi_style("#00CC44")
  col_blue  <- cli::make_ansi_style("#338AFF")
  col_red   <- cli::make_ansi_style("#EE3333")
  col_tan   <- cli::make_ansi_style("#C8A840")
  col_gray  <- cli::make_ansi_style("#AAAAAA")

  apply_color <- function(text, color) {
    switch(color,
      "green" = col_green(text),
      "blue"  = col_blue(text),
      "red"   = col_red(text),
      "tan"   = col_tan(text),
      "gray"  = col_gray(text),
      text
    )
  }

  # --- Bracket definitions ---
  open_brackets  <- c("(", "[", "<", "{")
  close_brackets <- c(")", "]", ">", "}")
  bracket_colors <- c("green", "blue", "red", "tan")

  chars <- strsplit(s, "")[[1]]
  n <- length(chars)

  # === Pass 1: Match bracket pairs and check for bare symbols ===
  pair_id    <- integer(n)
  pair_match <- integer(n)
  stack_pos  <- integer(0)
  pair_count <- 0L

  for (i in seq_len(n)) {
    if (chars[i] %in% open_brackets) {
      stack_pos <- c(stack_pos, i)
    } else if (chars[i] %in% close_brackets) {
      if (length(stack_pos) > 0) {
        open_pos <- stack_pos[length(stack_pos)]
        stack_pos <- stack_pos[-length(stack_pos)]
        pair_count <- pair_count + 1L
        pair_id[open_pos] <- pair_count
        pair_id[i] <- pair_count
        pair_match[open_pos] <- i
        pair_match[i] <- open_pos
      }
    }
  }

  # For each pair check whether it contains a bare (non-bracket) symbol
  # at nesting depth 0 relative to that pair.
  pair_has_bare <- logical(pair_count)

  for (i in seq_len(n)) {
    if (chars[i] %in% open_brackets && pair_id[i] > 0) {
      pid <- pair_id[i]
      if (pair_has_bare[pid]) next
      close_pos <- pair_match[i]
      if (close_pos - i < 2) next
      depth <- 0L
      for (j in (i + 1L):(close_pos - 1L)) {
        if (chars[j] %in% open_brackets) {
          depth <- depth + 1L
        } else if (chars[j] %in% close_brackets) {
          depth <- depth - 1L
        } else if (depth == 0L) {
          pair_has_bare[pid] <- TRUE
          break
        }
      }
    }
  }

  # === Pass 2: Color each character ===
  color_stack <- character(0)
  parts <- character(n)

  for (i in seq_len(n)) {
    ch <- chars[i]

    if (ch %in% open_brackets) {
      pid <- pair_id[i]
      if (pid > 0 && pair_has_bare[pid]) {
        col <- bracket_colors[match(ch, open_brackets)]
      } else {
        col <- "gray"
      }
      color_stack <- c(color_stack, col)
      parts[i] <- apply_color(ch, col)

    } else if (ch %in% close_brackets) {
      col <- if (length(color_stack) > 0) color_stack[length(color_stack)] else ""
      parts[i] <- apply_color(ch, col)
      if (length(color_stack) > 0) {
        color_stack <- color_stack[-length(color_stack)]
      }

    } else {
      col <- if (length(color_stack) > 0) color_stack[length(color_stack)] else ""
      parts[i] <- apply_color(ch, col)
    }
  }

  cat(paste0(parts, collapse = ""), "\n")
  invisible(x)
}

as.sign_name <- function(x, ...) {
  UseMethod("as.sign_name")
}

as.sign_name.default <- function(x, ...) {
  stop("Cannot coerce to cuneiform names")
}

as.sign_name.character <- function(x, mapping = NULL, ...) {
  if(is.null(mapping)){
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
  }
  result <- rep("", length(x))

  for(i in 1: length(x)){
    if(x[i]==""){next}
    this_info  <- info(x[i], mapping=mapping)
    separators <- attr(this_info, "separators")
    names      <- this_info$name

    sign_name_text <- paste0(c("",names), separators, collapse="")
    sign_name_text <- str_replace_all(sign_name_text, "-", ".")

    result[i] <- sign_name_text
  }

  class(result) <- c("sign_name", "character")

  return(result)
}

print.sign_name <- function(x, ...){
  cat(x, sep="\n")
}

prior_probs <- function(dic, sentence_prob = 1.0) {

  if (sentence_prob <= 0 || sentence_prob > 1) {
    stop("sentence_prob must be in (0, 1].")
  }

  trans <- dic[dic$row_type == "trans.", ]

  # --- Keep only single-sign entries with at least one count ---
  has_dot   <- grepl(".", trans$sign_name, fixed = TRUE)
  has_space <- grepl(" ", trans$sign_name, fixed = TRUE)
  single    <- trans[!has_dot & !has_space, ]
  single    <- single[!is.na(single$count) & single$count > 0L, ]
  single    <- single[!is.na(single$type) & trimws(single$type) != "", ]

  if (nrow(single) == 0L) {
    stop("No single-sign entries with counts found in dictionary.")
  }

  # --- Normalise type strings ---
  single$type <- trimws(single$type)

  # --- All unique grammar types ---
  all_types <- sort(unique(single$type))
  K <- length(all_types)

  # --- For each sign: compute relative frequencies across all types ---
  signs <- unique(single$sign_name)
  n_signs <- length(signs)

  # Matrix: one row per sign, one column per grammar type
  rel_freq <- matrix(0, nrow = n_signs, ncol = K,
                     dimnames = list(signs, all_types))

  for (i in seq_along(signs)) {
    entries <- single[single$sign_name == signs[i], ]

    for (j in seq_len(nrow(entries))) {
      col_idx <- match(entries$type[j], all_types)
      if (!is.na(col_idx)) {
        rel_freq[i, col_idx] <- rel_freq[i, col_idx] + entries$count[j]
      }
    }

    # Standardise to sum = 1
    total <- sum(rel_freq[i, ])
    if (total > 0) {
      rel_freq[i, ] <- rel_freq[i, ] / total
    }
  }

  # --- A priori probability: mean relative frequency across all signs ---
  p <- colMeans(rel_freq)

  # --- Correction for verb underrepresentation ---
  # A type is "verb-like" if it is V, or an operator with return type V
  is_verb <- vapply(all_types, function(type_str) {
    if (type_str == "V") return(TRUE)
    is_op <- grepl("\u2192", type_str) || grepl("->", type_str, fixed = TRUE)
    if (!is_op) return(FALSE)
    ret <- sub(".*(\u2192|->)\\s*", "", type_str)
    trimws(ret) == "V"
  }, logical(1))

  x <- ifelse(is_verb, 1 / sentence_prob, 1)
  p_corrected <- p * x / sum(p * x)

  attr(p_corrected, "sentence_prob") <- sentence_prob
  p_corrected
}

# --- split_structure: Tokenize a structure string ---
# Recognizes #N tags and {}-groups (with nesting) as tokens.
# Everything else is a separator.
# Returns list(tokens, seps) with length(seps) == length(tokens) + 1
split_structure <- function(s) {
  chars <- strsplit(s, "")[[1]]
  n <- length(chars)
  tokens <- character(0)
  seps <- character(0)
  buf <- ""
  i <- 1L

  while (i <= n) {
    ch <- chars[i]

    if (ch == "#") {
      seps <- c(seps, buf)
      buf <- ""
      rt <- read_tag(chars, i)
      tokens <- c(tokens, rt$tag)
      i <- rt$next_i

    } else if (ch == "{") {
      seps <- c(seps, buf)
      buf <- ""
      depth <- 1L
      group <- "{"
      i <- i + 1L
      while (i <= n && depth > 0) {
        if (chars[i] == "{") depth <- depth + 1L
        if (chars[i] == "}") depth <- depth - 1L
        group <- paste0(group, chars[i])
        i <- i + 1L
      }
      tokens <- c(tokens, group)

    } else {
      buf <- paste0(buf, ch)
      i <- i + 1L
    }
  }

  seps <- c(seps, buf)
  list(tokens = tokens, seps = seps)
}


# --- merge_at: merge two adjacent positions j, j+1 into one ---
# Used by combine_group() for composition rules.
merge_at <- function(tokens, types, seps, j, new_tok, new_type) {
  tokens <- c(tokens[seq_len(j - 1)],
              new_tok,
              if (j + 1 < length(tokens)) tokens[(j + 2):length(tokens)] else character(0))

  types <- c(types[seq_len(j - 1)],
             new_type,
             if (j + 1 < length(types)) types[(j + 2):length(types)] else character(0))

  rm_sep <- j + 1
  seps <- c(seps[seq_len(rm_sep - 1)],
            if (rm_sep + 1 <= length(seps)) seps[(rm_sep + 1):length(seps)] else character(0))

  list(tokens = tokens, types = types, seps = seps)
}


# --- combine_op: merge operator at j with its arguments ---
# Merges tokens from (j - n_left) to (j + n_right) into one {}-group.
# Arguments are wrapped in {}, operator stays bare.
combine_op <- function(tokens, types, seps, j, n_left, n_right, new_type) {
  start <- j - n_left
  end   <- j + n_right

  # Build parts: wrap arguments, keep operator bare
  parts <- character(end - start + 1)
  for (k in start:end) {
    tok <- tokens[k]
    if (k != j && !startsWith(tok, "{")) tok <- paste0("{", tok, "}")
    parts[k - start + 1] <- tok
  }

  # Assemble with internal separators
  new_tok <- parts[1]
  for (k in 2:length(parts)) {
    new_tok <- paste0(new_tok, seps[start + k - 1], parts[k])
  }
  new_tok <- paste0("{", new_tok, "}")

  # Build result arrays
  new_tokens <- c(if (start > 1) tokens[1:(start - 1)] else character(0),
                  new_tok,
                  if (end < length(tokens)) tokens[(end + 1):length(tokens)] else character(0))
  new_types <- c(if (start > 1) types[1:(start - 1)] else character(0),
                 new_type,
                 if (end < length(types)) types[(end + 1):length(types)] else character(0))
  new_seps <- c(seps[seq_len(start)],
                if (end + 1 <= length(seps)) seps[(end + 1):length(seps)] else character(0))

  list(tokens = new_tokens, types = new_types, seps = new_seps)
}


# --- combine_group: merge tokens[i] and tokens[i+1] for composition ---
# Wraps bare tags in {} so grammatical_structure can type them individually.
combine_group <- function(tokens, types, seps, i, new_type) {
  tok1 <- tokens[i]
  tok2 <- tokens[i + 1]
  if (!startsWith(tok1, "{")) tok1 <- paste0("{", tok1, "}")
  if (!startsWith(tok2, "{")) tok2 <- paste0("{", tok2, "}")
  new_tok <- paste0("{", tok1, seps[i + 1], tok2, "}")

  merge_at(tokens, types, seps, i, new_tok, new_type)
}


add_brackets <- function(s, type) {

  sp <- split_structure(s)
  tokens <- sp$tokens
  seps   <- sp$seps

  if (length(tokens) == 0)
    return(list(string = s, type = NA_character_))

  # Helper: build result from single remaining token
  make_result <- function()
    list(string = paste0(seps[1], tokens[1], seps[2]), type = types[1])

  # Step 0: Determine types, recursively process {} groups,
  #         and build parsed-type lookup for bare tags
  types <- character(length(tokens))
  parsed_list <- list()
  for (k in seq_along(tokens)) {
    if (startsWith(tokens[k], "{")) {
      inner <- substring(tokens[k], 2, nchar(tokens[k]) - 1)
      res <- add_brackets(inner, type)
      if (identical(res$type, "ERROR")) return(res)
      tokens[k] <- paste0("{", res$string, "}")
      types[k] <- res$type
    } else {
      tag_num <- as.integer(sub("#", "", tokens[k]))
      parsed <- parse_type(type[tag_num])
      types[k] <- get_result_type(parsed)
      parsed_list[[tokens[k]]] <- parsed
    }
  }

  if (length(tokens) == 1) return(make_result())

  get_parsed <- function(tok) parsed_list[[tok]]

  # Helper: TRUE if the token is a bare (unbound) operator tag.
  # Bare operator tags must never be used as arguments.
  is_bare_op <- function(tok) {
    p <- get_parsed(tok)
    !is.null(p) && p$is_operator
  }

  # --- Rule factories ---

  # Operator rule: binds an operator to its argument(s).
  # need_left/need_right = number of arguments on that side (0, 1, or 2).
  op_rule <- function(need_left, need_right) {
    function(j) {
      p <- get_parsed(tokens[j])
      if (is.null(p) || !p$is_operator) return(NULL)
      if (nchar(p$left_arg) != need_left) return(NULL)
      if (nchar(p$right_arg) != need_right) return(NULL)
      # Check left neighbors (first char of left_arg = furthest argument)
      for (k in seq_len(need_left)) {
        if (j - k < 1) return(NULL)
        if (is_bare_op(tokens[j - k])) return(NULL)
        if (!type_matches(types[j - k], str_sub(p$left_arg, need_left - k + 1, need_left - k + 1))) return(NULL)
      }
      # Check right neighbors (first char of right_arg = nearest argument)
      for (k in seq_len(need_right)) {
        if (j + k > length(tokens)) return(NULL)
        if (is_bare_op(tokens[j + k])) return(NULL)
        if (!type_matches(types[j + k], str_sub(p$right_arg, k, k))) return(NULL)
      }
      # Type inheritance: if result is a base type, inherit extension
      # from the first argument that has the same base type.
      result_type <- p$result
      if (!grepl("[a-z]", result_type)) {
        arg_positions <- c(
          if (need_left > 0) (j - need_left):(j - 1) else integer(0),
          if (need_right > 0) (j + 1):(j + need_right) else integer(0)
        )
        for (pos in arg_positions) {
          if (has_base_type(types[pos], result_type) && types[pos] != result_type) {
            result_type <- types[pos]
            break
          }
        }
      }
      combine_op(tokens, types, seps, j, need_left, need_right, result_type)
    }
  }


  # Pipe rule: like op_rule, but only matches if the first argument's
  # base type equals the result's base type (the type "flows through").
  # Rejects operators that change the base type (e.g. Sx->V with arg S).
  pipe_rule <- function(need_left, need_right) {
    inner <- op_rule(need_left, need_right)
    function(j) {
      p <- get_parsed(tokens[j])
      if (is.null(p) || !p$is_operator) return(NULL)
      # Determine first argument position
      first_pos <- if (need_left > 0) j - need_left else j + 1L
      if (first_pos < 1 || first_pos > length(tokens)) return(NULL)
      if (!has_base_type(types[first_pos], base_type(p$result))) return(NULL)
      inner(j)
    }
  }

  # Explicit rule: binds one specific operator (e.g. "xS->A").
  # The op_type string is parsed once; only operators whose left_arg,
  # right_arg, and result match exactly are considered.
  explicit_rule <- function(op_type) {
    ref   <- parse_type(op_type)
    inner <- op_rule(nchar(ref$left_arg), nchar(ref$right_arg))
    function(j) {
      p <- get_parsed(tokens[j])
      if (is.null(p) || !p$is_operator) return(NULL)
      if (p$left_arg != ref$left_arg || p$right_arg != ref$right_arg ||
          p$result != ref$result) return(NULL)
      inner(j)
    }
  }

  # Composition rule: groups two adjacent tokens by type.
  # Bare operator tags are rejected -- they must bind via operator rules.
  comp_rule <- function(t1, t2, result, sep_cond = NULL) {
    function(j) {
      if (j >= length(tokens)) return(NULL)
      if (is_bare_op(tokens[j]) || is_bare_op(tokens[j + 1])) return(NULL)
      if (!type_matches(types[j], t1) || !type_matches(types[j + 1], t2)) return(NULL)
      if (!is.null(sep_cond) && !sep_cond(seps[j + 1])) return(NULL)
      combine_group(tokens, types, seps, j, result)
    }
  }

  # --- Rules (applied in priority order) ---
  rules <- list(
    pipe_rule(need_left = 1, need_right = 0),                                   # Sx->S
    pipe_rule(need_left = 0, need_right = 1),                                   # xS->S
    pipe_rule(need_left = 1, need_right = 1),                                   # SxV->S
    pipe_rule(need_left = 2, need_right = 0),                                   # SSx->S
    pipe_rule(need_left = 0, need_right = 2),                                   # xSS->S
	explicit_rule("xS->A"),
    op_rule(need_left = 1, need_right = 0),                                   # Sx->R
    op_rule(need_left = 0, need_right = 1),                                   # xS->R
    op_rule(need_left = 1, need_right = 1),                                   # SxV->R
    op_rule(need_left = 2, need_right = 0),                                   # SSx->R
    op_rule(need_left = 0, need_right = 2),                                   # xSS->R
    comp_rule("S", "A", "S"),                                                 # S + A -> S
    comp_rule("S", "S", "S", function(s) str_detect(s, fixed(","))),          # S, S -> S
    comp_rule("S", "Vt", "Vi"),                                                # S + Vt -> Vi
    comp_rule("S", "S", "S", function(s) !str_detect(s, "[.,;:!?0-9N\u00a4]")),  # S + S -> S
    comp_rule("S", "Vi", "SEN"),                                               # S + Vi -> SEN
    comp_rule("S", "V", "SEN"),                                                # S + V -> SEN
    comp_rule("SEN", "SEN", "SEN")                                            # SEN + SEN -> SEN
  )

  # Apply rules: try each rule in order, restart after each match
  changed <- TRUE
  while (changed && length(tokens) > 1) {
    changed <- FALSE
    for (rule in rules) {
      for (j in seq_along(tokens)) {
        res <- rule(j)
        if (!is.null(res)) {
          tokens <- res$tokens; types <- res$types; seps <- res$seps
          changed <- TRUE
          break
        }
      }
      if (changed) break
    }
  }

  if (length(tokens) != 1) {
    # Check for unbound operators to give a specific error message
    for (k in seq_along(tokens)) {
      if (is_bare_op(tokens[k])) {
        tag_num <- as.integer(sub("#", "", tokens[k]))
        p <- get_parsed(tokens[k])
        side <- if (nchar(p$left_arg) > 0) "left" else "right"
        return(list(string = paste0("Operator ", type[tag_num],
                      " (position ", tag_num, ") has no argument to its ", side),
                    type = "ERROR"))
      }
    }
    remaining <- paste(types, collapse = " + ")
    return(list(string = paste0("Cannot group remaining elements: ", remaining),
                type = "ERROR"))
  }

  make_result()
}

"read_dictionary" <- function(file=NULL, verbose=TRUE){

  if(is.null(file)){
    file  <- system.file("extdata", "sumer-dictionary.txt", package = "sumer")
  }

  dic <- read.table(
         file,
         sep = ";",
         header = TRUE,
         comment.char = "#",
         na.strings="",
         quote="",
         stringsAsFactors = FALSE,
         fileEncoding = "UTF-8"
     )

 dic$type[is.na(dic$type)]<-""

 if(verbose){
  meta <- readLines(file, n=20)
  meta <- meta[grepl("^###", meta)]
  meta <- paste0(meta, "\n")
  meta[1] <- paste0(" ", meta[1])
  cat(meta)
 }

 dic
}

"look_up" <- function(x, dic, lang = "sumer", width = 70, mapping = NULL) {

  if (is.null(mapping)) {
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep = ";", na.strings = "")
  }

  # Hilfsfunktion: Art der Eingabe feststellen

  is_sumerian <- function(x) {
    x <- str_replace_all(x, "[()\\[\\]{}\\.,\\-<>\u00a4\u2218]", " ")
    x <- str_squish(x)
    x <- str_replace_all(x, " ", "-")
    x <- str_split_1(x, "-")[1]

    lowercase <- "[a-z\u011d\u0161\u1e2b\u00e1\u00e9\u00ed\u00fa\u00e0\u00e8\u00ec\u00f9]+[0-9]*"
    uppercase <- "[A-Z\u0160\u011c\u1e2a](?:[A-Z\u0160\u011c\u1e2a0-9\u00d7+/%\u00b7]|gunu|tenu|\u0161e\u0161ig)*"
    cuneiform <- "[\\x{12000}-\\x{12500}\u2205]"
    others    <- paste0("<", lowercase, ">|<", uppercase, ">|", uppercase)

    result <- str_detect(str_sub(x, 1, 1), cuneiform) || str_detect(x, paste0("^(", others, ")$"))

    return(result)
  }

  if (is_sumerian(x) && lang != "sumer" && lang != "sum") {
    lang <- "sumer"
    warning("The language was changed to 'sumer'.")
  }

  # Ergebnisobjekt initialisieren
  result <- list(
    search       = x,
    lang         = lang,
    width        = width,
    mapping      = mapping,
    cuneiform    = NULL,
    sign_name    = NULL,
    translations = NULL,
    substrings   = NULL,
    matches      = NULL
  )

  if (lang == "sumer" || lang == "sum") {
    flat.x <- str_replace_all(x, "[()\\[\\]{}\\.,\\-<>\u00a4\u2218]", " ")
    flat.x <- str_squish(flat.x)
    flat.x <- str_replace_all(flat.x, " ", "-")

    result$cuneiform <- as.cuneiform(flat.x, mapping = mapping)
    result$sign_name <- as.sign_name(flat.x, mapping = mapping)

    # Haupteinträge
    name <- result$sign_name
    result$translations <- dic[(dic$sign_name == name) & (dic$row_type == "trans."), ]

    # Teil-Zeichenketten
    words <- str_split_1(name, "\\.")
    n <- length(words)
    substrings <- unlist(lapply(1:n, function(i) {
      sapply(i:n, function(j) str_c(words[i:j], collapse = "."))
    }))
    substrings <- substrings[substrings != name]
    substrings <- substrings[substrings %in% dic$sign_name]

    if (length(substrings) > 0) {
      result$substrings <- lapply(substrings, function(s) {
        dic[(dic$sign_name == s) & (dic$row_type == "trans."), ]
      })
      names(result$substrings) <- substrings
    }

  } else {
    # Suche nach deutschem/englischem Begriff
    pattern <- paste0("(?<!\\p{L})", str_escape(x), "(?!\\p{L})")
    use     <- str_detect(dic$meaning, regex(pattern, ignore_case = TRUE)) & (dic$row_type == "trans.")
    result$matches <- dic[use, ]
  }

  class(result) <- "look_up"
  return(result)
}


"print.look_up" <- function(x, ...) {


  width <- x$width

  # Hilfsfunktion: Text auf feste Breite auffüllen (Unicode-sicher)
  pad_right <- function(text, width) {
    n <- nchar(text)
    if (n < width) {
      paste0(text, paste(rep(" ", width - n), collapse = ""))
    } else {
      text
    }
  }

  # Hilfsfunktion für blauen Text
  blue <- function(text) {
    cli::col_blue(text)
  }

  # Hilfsfunktion: Suchbegriff im Text hervorheben
  highlight <- function(text, pattern) {
    str_replace_all(
      text,
      regex(paste0("((?<!\\p{L})", str_escape(pattern), "(?!\\p{L}))"), ignore_case = TRUE),
      blue("\\1")
    )
  }

  # Hilfsfunktion: Text umbrechen
  wrap_text <- function(text, width, indent = 10) {
    indent_str <- paste(rep(" ", indent), collapse = "")
    lines <- strwrap(text, width = width)
    paste(lines, collapse = paste0("\n", indent_str))
  }

  # Hilfsfunktion: Einträge formatiert ausgeben (kompakt, für kurze sign_names)
  print_entries_compact <- function(df) {
    if (nrow(df) == 0) {
      cat("  (no entries found)\n")
      return(invisible(NULL))
    }
    for (i in 1:nrow(df)) {
      count_str <- if (is.na(df$count[i])) " " else sprintf("%2d", df$count[i])
      type_str  <- pad_right(df$type[i], 6)
      meaning   <- wrap_text(df$meaning[i], width, indent = 14)
      cat(sprintf("  [%s] %s %s\n", count_str, type_str, meaning))
    }
  }

  # Hilfsfunktion: Einträge formatiert ausgeben (mehrzeilig, für lange sign_names)
  print_entries_full <- function(df, pattern, show_cuneiform = FALSE) {
    if (nrow(df) == 0) {
      cat("  (no entries found)\n")
      return(invisible(NULL))
    }
    for (i in 1:nrow(df)) {
      # Zeile 1: sign_name (und optional Keilschrift)
      if (show_cuneiform) {
        cune <- as.cuneiform(df$sign_name[i], mapping = x$mapping)
        cat(sprintf("\n  %s  %s\n", df$sign_name[i], cune))
      } else {
        cat(sprintf("\n  %s\n", df$sign_name[i]))
      }
      # Zeile 2+: count, type, meaning
      count_str <- if (is.na(df$count[i])) " " else sprintf("%d", df$count[i])
      type_str  <- pad_right(df$type[i], 6)
      meaning   <- highlight(df$meaning[i], pattern)
      meaning   <- wrap_text(meaning, width, indent = 13)

      cat(sprintf("  [%s] %-6s %s\n", count_str, type_str, meaning))
    }
  }

  # Trennlinie
  line <- paste(rep("\u2500", width + 20), collapse = "")

  cat("\n", line, "\n", sep = "")
  cat(" Search: ", x$search, "\n", sep = "")
  cat(line, "\n", sep = "")

  if (x$lang == "sumer" || x$lang == "sum") {

    cat("\n Cuneiform:   ", x$cuneiform, "\n", sep = "")
    cat(" Sign Names:  ", x$sign_name, "\n\n", sep = "")

    # Haupteintrag
    cat(" \u25b6 Translations:\n")
    print_entries_compact(x$translations)

    # Teil-Zeichenketten
    if (!is.null(x$substrings) && length(x$substrings) > 0) {
      cat("\n \u25b6 Individual Signs / Substrings:\n")
      for (s in names(x$substrings)) {
        sub_df <- x$substrings[[s]]
        cune_s <- as.cuneiform(s, mapping = x$mapping)
        cat(sprintf("\n  %s  %s\n", s, cune_s))
        for (j in 1:nrow(sub_df)) {
          count_str <- if (is.na(sub_df$count[j])) " " else pad_right(paste0("[", sub_df$count[j], "]"), 4)
          type_str  <- pad_right(sub_df$type[j], 6)
          meaning   <- wrap_text(sub_df$meaning[j], width, indent = 14)
          cat(sprintf("  %s %-6s %s\n", count_str, type_str, meaning))
        }
      }
    }

  } else {
    # Suche nach deutschem/englischem Begriff
    cat("\n \u25b6 Matches for '", x$search, "':\n", sep = "")
    print_entries_full(x$matches, x$search, show_cuneiform = TRUE)
  }

  cat("\n", line, "\n\n", sep = "")
  invisible(x)
}

info <- function(x, mapping = NULL) {

  if((!is.character(x))|length(x)>1){stop("The parameter x must be a character vector of length 1.")}

  if(is.null(mapping)){
    path  <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
    }

  temp       <- split_sumerian(x)
  signs      <- temp$signs
  separators <- temp$separators
  types      <- temp$types

  signs <- standardize_signs(signs)

  n   <- length(signs)

  cuneiform <- "[\\x{12000}-\\x{12500}\u2205]"



  syllables    <- ifelse(types==3, str_remove_all(signs, "[\\u27E8\\u27E9]"), signs)
  names        <- ifelse(types==3, str_remove_all(signs, "[\\u27E8\\u27E9]"), signs)
  symbols      <- ifelse(types >= 4, signs,
                 ifelse(types == 3, signs, str_c("\u27E8", signs, "\u27E9")))
  readings     <- ifelse(types==3, str_remove_all(signs, "[\\u27E8\\u27E9]"), signs)
  alternatives <- ifelse(types==3, str_remove_all(signs, "[\\u27E8\\u27E9]"), signs)

  syllable_vector  <- paste0(",",str_replace_all(mapping$syllables, " ",""), ",")
  default_syllable <- str_remove(mapping$syllables, ",.*")
  default_syllable[is.na(default_syllable)] <- mapping$name[is.na(default_syllable)]

  for(i in seq_along(signs)){
    k <- switch(
      types[i],
      which(str_detect(syllable_vector, paste0(",", signs[i], ","))),
      which(mapping$name      == signs[i]),
      which(mapping$cuneiform == signs[i]),
      integer(0),
      integer(0)
    )

    if(length(k)>0){
      k <- k[1]
      readings[i]     <- ifelse(types[i]==1, signs[i], default_syllable[k])
      names[i]        <- mapping$name[k]
      symbols[i]      <- mapping$cuneiform[k]
      alternatives[i] <- mapping$syllables[k]
    }
  }

  result <- data.frame(
    reading     = readings,
    sign        = symbols,
    name        = names,
    alternatives = alternatives,
    stringsAsFactors = FALSE
  )

  attr(result, "separators") <- separators

  class(result) <- c("info", "data.frame")

  return(result)
}


#\u00a4  operator: of
#\u2218  operator: chaining

"flatten_cuneiform" <- function(x){
  x <- str_replace_all(x, "[()\\[\\]{}\\.,\\-<>\u00a4\u2218]", "-")
  x <- str_replace_all(x, fixed("--"), "-")
  x <- str_replace_all(x, fixed("--"), "-")
  x <- str_replace_all(x, fixed("--"), "-")
  x <- str_replace_all(x, fixed(" -"), " ")
  x <- str_replace_all(x, fixed("- "), " ")
  if(x[1]=="-"){
    x[1] <- ""
  }
  if(x[length(x)]=="-"){
    x[length(x)] <- ""
  }
  return(x)
}


"print.info" <- function(x, flatten=FALSE, ...){

  separators <- attr(x, "separators")

  if(flatten){
    separators <- flatten_cuneiform(separators)
    }
  cuneiform_text <- paste0(c("",x$sign), separators, collapse="")
  cuneiform_text <- str_replace_all(cuneiform_text, "-(?=[0-9])", " ")
  cuneiform_text <- str_replace_all(cuneiform_text, "-", "")
  cuneiform_text <- str_replace_all(cuneiform_text, fixed("."), "")

  sign_name_text <- paste0(c("",x$name), separators, collapse="")
  sign_name_text <- str_replace_all(sign_name_text, "-", ".")

  syllable_text <- paste0(c("",x$reading), separators, collapse="")
  syllable_text <- str_replace_all(syllable_text, "\\.(?!\\s|$)", "-")

  for(i in 1:nrow(x)){
    cat(paste0(x$sign[i],"\t",x$name[i],"\t",x$alternatives[i]),"\n")
  }
  cat("\n")

  cat(paste0("syllables      : ", syllable_text, "\n"))
  cat(paste0("sign names     : ", sign_name_text, "\n"))
  cat(paste0("cuneiform text : ", cuneiform_text, "\n"))
}

# =============================================================================
# eval_operator: evaluate an operator or composition
# =============================================================================


# --- Helper: extract {specific} from "general {specific}" if present ---
extract_specific <- function(trans) {
  m <- regmatches(trans, regexpr("(?<=\\{)[^}]+(?=\\})", trans, perl = TRUE))
  if (length(m) > 0 && nchar(m) > 0) m else trans
}

# --- Helper: capitalize first letter and ensure trailing period ---
make_sentence <- function(s) {
  s <- sub("^(\\s*)(\\w)", "\\1\\U\\2", s, perl = TRUE)
  if (!endsWith(trimws(s), ".")) s <- paste0(s, ".")
  s
}

# --- Helper: remove <annotations> from translations ---
strip_annotations <- function(trans) {
  trans <- gsub("<[^>]*>", "", trans)
  trans <- gsub("\\s+", " ", trimws(trans))
  trans
}


# --- eval_operator: fully evaluate an operator or composition ---
#
# Contains the complete translation logic: argument processing (article
# insertion, verb-prefix stripping), placeholder substitution, and
# composition rules.
#
# @param operator Type string of the operator (e.g., "Sx->V", "SSx->V"),
#   or NULL for composition (no operator present).
# @param rule_translation Translation string with placeholders
#   (e.g., "to utilize S"), or NULL for composition.
# @param args List of c(name, translation, original_type) vectors:
#   - name: placeholder name derived from the base type (e.g., "S", "S1",
#     "V", "SEN2"). The base type is reconstructed by stripping the
#     trailing index.
#   - translation: resolved translation of the argument
#   - original_type: the full (possibly extended) type (e.g. "So", "Vt")
# @param seps Character vector of separators between elements
# @param trailing_sep Text after the last element
# @return c(result_type, result_translation) or error message (length 1)
eval_operator <- function(operator, rule_translation, args = list(), seps = "", trailing_sep = "") {

  # Apply extract_specific and strip_annotations to all translations upfront
  if (!is.null(rule_translation)) {
    rule_translation <- extract_specific(rule_translation)
    rule_translation <- strip_annotations(rule_translation)
  }
  for (i in seq_along(args)) {
    args[[i]][2] <- extract_specific(args[[i]][2])
    args[[i]][2] <- strip_annotations(args[[i]][2])
  }

  if (!is.null(operator)) {
    # =================================================================
    # Operator evaluation: substitute arguments into rule_translation
    # =================================================================
    parsed <- parse_type(operator)
    result_type <- get_result_type(parsed)

    trans <- rule_translation
    for (arg in args) {
      name      <- arg[1]                         # placeholder name (S, S1, V, ...)
      arg_type  <- sub("[0-9]+$", "", name)        # type reconstructed from name
      arg_trans <- arg[2]

      # Strip "to " from verb-type arguments
      if (!is.na(arg_type) && arg_type == "V")
        arg_trans <- sub("^to ", "", arg_trans)

      # Add "the " for S-type arguments when the operator changes
      # the type (e.g., Sx->V gets "the", but xS->S does not).
      #if (!is.na(arg_type) && arg_type == "S" && !identical("S", result_type))
      #  arg_trans <- paste0("the ", arg_trans)

      # Replace the placeholder (whole-word match)
      trans <- sub(paste0("\\b", name, "\\b"), arg_trans, trans)
    }

    # Type inheritance: if result is a base type and the first argument
    # with that base type carries an extended type, inherit the extension.
    if (!grepl("[a-z]", result_type)) {
      for (arg in args) {
        orig_type <- if (length(arg) >= 3) arg[3] else sub("[0-9]+$", "", arg[1])
        if (has_base_type(orig_type, result_type) && orig_type != result_type) {
          result_type <- orig_type
          break
        }
      }
    }

    # Avoid double period
    if (endsWith(trans, ".") && startsWith(trailing_sep, "."))
      trailing_sep <- substring(trailing_sep, 2)

    trans <- paste0(trans, trailing_sep)
    if (has_base_type(result_type, "SEN")) trans <- make_sentence(trans)
    return(c(result_type, trans))
  }

  # =================================================================
  # No operator: base case or composition
  # =================================================================
  seps <- gsub("-", " ", seps, fixed = TRUE)
  trailing_sep <- gsub("-", " ", trailing_sep, fixed = TRUE)

  # Helper: remove any space immediately before a comma
  fix_comma <- function(s) gsub(" ,", ",", s, fixed = TRUE)

  # Composition: exactly two elements required
  if (length(args) != 2) {
    remaining <- paste(sapply(args, function(a) sub("[0-9]+$", "", a[1])),
                       collapse = " + ")
    return(paste0("Cannot compose: ", remaining))
  }

  t1 <- sub("[0-9]+$", "", args[[1]][1]);  tr1 <- args[[1]][2]
  t2 <- sub("[0-9]+$", "", args[[2]][1]);  tr2 <- args[[2]][2]
  sep <- if (length(seps) > 0) seps[1] else " "

  # Full (extended) types from build_args third element
  full_t1 <- if (length(args[[1]]) >= 3) args[[1]][3] else t1
  full_t2 <- if (length(args[[2]]) >= 3) args[[2]][3] else t2

  # S + A -> S: simple juxtaposition
  if (t1 == "S" && t2 == "A")
    return(c("S", fix_comma(paste0(tr1, " ", tr2, trailing_sep))))

  # S + S -> S: genitive/instrumental ("of/with")
  if (t1 == "S" && t2 == "S" && !grepl(",", sep, fixed = TRUE))
    return(c("S", fix_comma(paste0(tr1, " of/with ", tr2, trailing_sep))))

  # S, S -> S: comma-separated list
  if (t1 == "S" && t2 == "S" && grepl(",", sep, fixed = TRUE))
    return(c("S", fix_comma(paste0(tr1, sep, tr2, trailing_sep))))

  # S + Vt -> Vi: object + transitive verb (swap order)
  if (has_base_type(full_t1, "S") && full_t2 == "Vt") {
    v_trans <- sub("^to ", "", tr2)
    if (grepl("\\bS\\b", v_trans)) {
      v_trans <- sub("\\bS\\b", tr1, v_trans)
      return(c("Vi", fix_comma(paste0(v_trans, trailing_sep))))
    }
    return(c("Vi", fix_comma(paste0(v_trans, " ", tr1, trailing_sep))))
  }

  # S + Vi -> SEN: subject + intransitive verb (strip "to ", keep order)
  if (has_base_type(full_t1, "S") && full_t2 == "Vi") {
    v_trans <- sub("^to ", "", tr2)
    return(c("SEN", make_sentence(fix_comma(paste0(tr1, " ", v_trans, trailing_sep)))))
  }

  # S + V -> SEN: subject-verb (strip "to ")
  if (t1 == "S" && t2 == "V") {
    v_trans <- sub("^to ", "", tr2)
    return(c("SEN", make_sentence(fix_comma(paste0(tr1, " ", v_trans, trailing_sep)))))
  }

  # SEN + SEN -> SEN: sentence concatenation (add period if missing)
  if (t1 == "SEN" && t2 == "SEN") {
    period <- if (endsWith(tr1, ".") || grepl("\\.", sep)) "" else "."
    # Avoid double period when tr1 already ends with "."
    if (endsWith(tr1, ".") && startsWith(sep, "."))
      sep <- substring(sep, 2)
    # Avoid double period when tr2 already ends with "."
    if (endsWith(tr2, ".") && startsWith(trailing_sep, "."))
      trailing_sep <- substring(trailing_sep, 2)
    return(c("SEN", fix_comma(paste0(tr1, period, sep, tr2, trailing_sep))))
  }

  remaining <- paste(sapply(args, function(a) sub("[0-9]+$", "", a[1])),
                     collapse = " + ")
  paste0("Cannot compose: ", remaining)
}

init_substr_info <- function(token) {
  N <- length(token)
  n_rows <- N * (N + 1) / 2

  starts   <- integer(n_rows)
  lengths  <- integer(n_rows)
  exprs    <- character(n_rows)

  row <- 0L
  for (k in seq(N, 1)) {
    for (s in seq_len(N - k + 1)) {
      row <- row + 1L
      starts[row]  <- s
      lengths[row] <- k
      exprs[row]   <- paste0(token[s:(s + k - 1)], collapse = "")
    }
  }

  data.frame(
    start       = starts,
    n_tokens    = lengths,
    expr        = exprs,
    type        = "",
    translation = "",
    stringsAsFactors = FALSE
  )
}

split_sumerian <- function(x) {
  # Regex für sumerische Zeichen

  # a) Kleinbuchstaben a-z, ĝ, š, ḫ, optional gefolgt von Zahl
  lowercase <- "[a-z\u011d\u0161\u1e2b\u00e1\u00e9\u00ed\u00fa\u00e0\u00e8\u00ec\u00f9]+[0-9]*"

  # b) Großbuchstaben, optional gefolgt von Großbuchstaben, Zahlen, × oder +
  uppercase <- "[A-Z\u0160\u011c\u1e2a](?:[A-Z\u0160\u011c\u1e2a0-9\u00d7+/%\u00b7]|gunu|tenu|\u0161e\u0161ig)*"

  # c) Keilschrift: Unicode U+12000 bis U+123FF
  cuneiform <- "[\\x{12000}-\\x{12500}\u2205]"
  cuneiform <- paste0(cuneiform, "|\u27E8", lowercase, "\u27E9|\u27E8", uppercase, "\u27E9")

  # d) Zahlen und N (Platzhalter für beliebige Zahl)
  #    N nur standalone (nicht vor weiterem Uppercase-Zeichen, damit NE, NI, N4 etc. Typ 2 bleiben)
  #    X und N selbst sind aus dem Lookahead ausgenommen, damit XX, XN, NX, NN
  #    als separate Token erkannt werden.
  #    Ziffern bleiben im Lookahead: N4 ohne Separator ist ein Zeichenname (Typ 2).
  #    as.cuneiform bewahrt Separatoren vor Ziffern, damit N-4 getrennt bleibt.
  uc_follow_no_xn <- "[A-MO-WYZ\u0160\u011c\u1e2a0-9\u00d7+/%\u00b7]"
  number  <- paste0("N(?!", uc_follow_no_xn, ")|[0-9]+(?:\\.[0-9]+)?")

  # e) X (Platzhalter für nicht lesbares Zeichen), analog standalone
  unknown <- paste0("X(?!", uc_follow_no_xn, ")")

  # Kombiniertes Pattern (number/unknown VOR lowercase/uppercase für Vorrang)
  pattern <- paste0("(", cuneiform, "|", number, "|", unknown, "|", lowercase, "|", uppercase, ")")

  # Finde alle Matches mit Positionen
  matches <- str_locate_all(x, pattern)[[1]]

  # Sonderfall: keine Matches
  if (nrow(matches) == 0) {
    return(list(signs = character(0), separators = x, type=integer(0)))
  }

  # Extrahiere die sumerischen Zeichen
  signs <- str_sub(x, matches[, "start"], matches[, "end"])

  # Berechne die Trennzeichen (n+1 Elemente für n Zeichen)
  n <- length(signs)
  separators <- character(n + 1)

  # Vor dem ersten Zeichen
  if (matches[1, "start"] > 1) {
    separators[1] <- str_sub(x, 1, matches[1, "start"] - 1)
  }

  # Zwischen den Zeichen
  if (n > 1) {
    for (i in 1:(n - 1)) {
      start_sep <- matches[i, "end"] + 1
      end_sep <- matches[i + 1, "start"] - 1
      if (start_sep <= end_sep) {
        separators[i + 1] <- str_sub(x, start_sep, end_sep)
      }else{
        separators[i + 1]<-"-"
      }
    }
  }

  # Nach dem letzten Zeichen
  if (matches[n, "end"] < str_length(x)) {
    separators[n + 1] <- str_sub(x, matches[n, "end"] + 1, str_length(x))
  }


  # Bestimme den Typ für jedes Zeichen
  # Vollständige Patterns mit Anfang/Ende-Ankern
  pattern_lowercase <- paste0("^", lowercase, "$")
  pattern_uppercase <- paste0("^", uppercase, "$")
  pattern_cuneiform <- paste0("^(?:", cuneiform, ")$")
  pattern_number    <- paste0("^(?:", number, ")$")
  pattern_unknown   <- paste0("^(?:", unknown, ")$")

  types <- integer(length(signs))
  types[str_detect(signs, pattern_lowercase)] <- 1L
  types[str_detect(signs, pattern_uppercase)] <- 2L
  types[str_detect(signs, pattern_cuneiform)] <- 3L
  types[str_detect(signs, pattern_number)]    <- 4L
  types[str_detect(signs, pattern_unknown)]   <- 5L


  list(signs = signs, separators = separators, types=types)
}

read_translated_text <- function(file, mapping=NULL){

  if(!is.character(file))stop("file must be a character-vector")

  if(is.null(mapping)){
    path  <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
  }

  x <- character(0)
  for(i in 1:length(file)){
    if(str_ends(file[i], fixed(".docx"))){
      doc     <- read_docx(file[i])
      doc_xml <- doc$doc_obj$get()
      bookmark_starts <- xml_find_all(doc_xml, "//w:bookmarkStart")
      bookmark_ends   <- xml_find_all(doc_xml, "//w:bookmarkEnd")
      xml_remove(bookmark_starts)
      xml_remove(bookmark_ends)
      text   <- docx_summary(doc)$text
    }else{
      text <- readLines(file[i])
    }
    text <- str_trim(text)
    text <- text[str_starts(text, fixed("|"))]
    x <- c(x, text)
  }

  x <- str_sub(x,2)
  x <- str_squish(x)

  first  <- str_split_fixed(x, ":", 2)
  second <- str_split_fixed(first[,2], ":", 2)

  df <- data.frame(
    sign_name = first[,1],
    type      = ifelse(second[,2] == "", "", second[,1]),
    meaning   = ifelse(second[,2] == "", second[,1], second[,2]),
    stringsAsFactors = FALSE
  )

  if(any(df$type=="")){
    missing <- x[df$type==""]
    warning("The following translations have missing types:\n",
            paste(missing, collapse = "\n"),
            call. = FALSE)
  }

  df$meaning   <- str_split_fixed(df$meaning, fixed("|"), 2)[, 1]
  df$meaning   <- str_squish(df$meaning)

  df$type      <- str_squish(df$type)
  df$type      <- str_replace_all(df$type, " ", "")
  df$type      <- str_replace_all(df$type, "->", "\u2192")
  df$type      <- str_replace_all(df$type, "x", "\u2612")

  df$sign_name <- str_split_fixed(df$sign_name, "=", 2)[, 1]
  df$sign_name <- flatten_cuneiform(df$sign_name)

  # Zeilen mit X-Token (Typ 5) ausschließen
  has_x <- vapply(df$sign_name, function(s) {
    sp <- split_sumerian(s)
    any(sp$types == 5L)
  }, logical(1))
  df <- df[!has_x, , drop = FALSE]

  # Zahlen-Token (Typ 4, aber nicht N) durch N ersetzen
  for (i in seq_len(nrow(df))) {
    sp <- split_sumerian(df$sign_name[i])
    num_idx <- which(sp$types == 4L & sp$signs != "N")
    if (length(num_idx) > 0) {
      sp$signs[num_idx] <- "N"
      df$sign_name[i] <- paste0(sp$separators[1],
        paste0(sp$signs, sp$separators[-1], collapse = ""))
    }
  }

  df <- df[df$sign_name!="",]
  rownames(df) <- 1:nrow(df)

  df$sign_name <- as.sign_name(df$sign_name, mapping=mapping)

  return(df)
}

mark_ngrams <- function(x, ngram, mapping = NULL) {

  if (!is.character(x) || length(x) == 0L) {
    stop("x must be a non-empty character vector.")
  }
  if (!is.data.frame(ngram) || !all(c("combination", "length") %in% names(ngram))) {
    stop("ngram must be a data frame with columns 'combination' and 'length' (from ngram_frequencies).")
  }

  # --- Convert to cuneiform if needed ---
  has_cuneiform <- any(grepl("[\U00012000-\U0001254F]", x, perl = TRUE))
  if (!has_cuneiform) {
    x <- as.cuneiform(x, mapping = mapping)
  }

  # --- Remove spaces and brackets ---
  x <- str_replace_all(x, "[()\\[\\]{} ]", "")

  # --- Sort ngram descending by length ---
  ngram <- ngram[ngram$length>1, ]
  ngram <- ngram[order(-ngram$length), ]

  # --- Replace each combination ---
  for (i in seq_len(nrow(ngram))) {
    com <- ngram$combination[i]
    replacement <- paste0(" {", com, "} ")
    x <- str_replace_all(x, fixed(com), replacement)
  }

  x
}

as.cuneiform <- function(x, ...) {
  UseMethod("as.cuneiform")
}

as.cuneiform.default <- function(x, ...) {
  stop("Cannot coerce to cuneiform")
}

as.cuneiform.character <- function(x, mapping = NULL, ...) {
  if(is.null(mapping)){
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
  }
  result <- rep("", length(x))

  for(i in 1: length(x)){
    if(x[i]==""){next}
    if(str_starts(x[i], "#")){result[i]<-x[i];next}
    this_info  <- info(x[i], mapping=mapping)
    separators <- attr(this_info, "separators")
    symbols    <- this_info$sign

    cuneiform_text <- paste0(c("",symbols), separators, collapse="")
    cuneiform_text <- str_replace_all(cuneiform_text, "-(?=[0-9])", " ")
    cuneiform_text <- str_replace_all(cuneiform_text, "-", "")
    cuneiform_text <- str_replace_all(cuneiform_text, "\\.(?=[\\x{12000}-\\x{1254F}\\u27E8])", "")

    result[i] <- cuneiform_text
  }

  class(result) <- c("cuneiform", "character")

  return(result)
}

print.cuneiform <- function(x, ...){
  cat(x, sep="\n")
}

sign_grammar <- function(x, dic, mapping = NULL) {

  if (is.null(mapping)) {
    path    <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep = ";", na.strings = "")
  }

  # --- Input: single character string ---
  if (length(x) != 1L) stop("x must be a single character string.")

  # --- Convert to cuneiform, then tokenize into individual signs ---
  cune <- as.cuneiform(x, mapping = mapping)
  cune_str <- as.character(cune)

  # Tokenize: extract each cuneiform character (U+12000..U+1254F)
  # and bracket tokens (angle bracket sequences)
  chars <- strsplit(cune_str, "")[[1]]
  n_ch  <- length(chars)
  tokens <- character(0)
  i <- 1L
  while (i <= n_ch) {
    cp <- utf8ToInt(chars[i])
    if (cp == 0x27E8L) {
      # Bracket token
      bracket <- chars[i]
      i <- i + 1L
      while (i <= n_ch && utf8ToInt(chars[i]) != 0x27E9L) {
        bracket <- paste0(bracket, chars[i])
        i <- i + 1L
      }
      if (i <= n_ch) {
        bracket <- paste0(bracket, chars[i])
        i <- i + 1L
      }
      tokens <- c(tokens, bracket)
    } else if (cp >= 0x12000L && cp <= 0x1254FL) {
      tokens <- c(tokens, chars[i])
      i <- i + 1L
    } else {
      i <- i + 1L
    }
  }

  n <- length(tokens)
  if (n == 0L) {
    return(data.frame(position   = integer(0),
                      sign_name  = character(0),
                      cuneiform  = character(0),
                      type       = character(0),
                      n          = integer(0),
                      stringsAsFactors = FALSE))
  }

  # --- Get sign name for each token ---
  sign_names <- vapply(tokens, function(tok) {
    as.character(as.sign_name(tok, mapping = mapping))
  }, character(1), USE.NAMES = FALSE)

  # --- Look up each sign in the dictionary ---
  trans <- dic[dic$row_type == "trans.", ]
  base_types <- c("S", "A", "V")

  # --- Build result: only rows with n > 0, or base types for unknown signs ---
  rows <- list()

  for (i in seq_len(n)) {
    s <- sign_names[i]
    entries <- trans[trans$sign_name == s, ]
    entries <- entries[!is.na(entries$type) & trimws(entries$type) != "", ]

    if (nrow(entries) > 0L) {
      # Known sign: collect counts per type, keep only n > 0
      type_counts <- list()
      for (j in seq_len(nrow(entries))) {
        t <- trimws(entries$type[j])
        cnt <- if (is.na(entries$count[j])) 1L else as.integer(entries$count[j])
        if (is.null(type_counts[[t]])) type_counts[[t]] <- 0L
        type_counts[[t]] <- type_counts[[t]] + cnt
      }
      for (t in names(type_counts)) {
        if (type_counts[[t]] > 0L) {
          rows[[length(rows) + 1L]] <- data.frame(
            position = i, sign_name = s, cuneiform = tokens[i],
            type = t, n = type_counts[[t]],
            stringsAsFactors = FALSE, row.names = NULL
          )
        }
      }
    } else {
      # Unknown sign: output base types with n = 0
      for (bt in base_types) {
        rows[[length(rows) + 1L]] <- data.frame(
          position = i, sign_name = s, cuneiform = tokens[i],
          type = bt, n = 0L,
          stringsAsFactors = FALSE, row.names = NULL
        )
      }
    }
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

ngram_frequencies <- function(x, min_freq = c(6, 4, 2), mapping = NULL) {

  # --- Eingabe in Keilschrift konvertieren (falls noetig) ---
  # Pruefen ob bereits Keilschrift enthalten ist (U+12000..U+1254F)
  has_cuneiform <- any(grepl("[\U00012000-\U0001254F]", x, perl = TRUE))
  if (!has_cuneiform && length(x) > 0) {
    x <- as.cuneiform(x, mapping = mapping)
  }

  x <- str_replace_all(x, " ", "")

  # --- Vorverarbeitung ---
  # Kommentarzeilen (beginnen mit #) und Leerzeilen entfernen
  x <- x[!grepl("^\\s*#", x)]
  x <- x[nchar(trimws(x)) > 0]

  # Zeilennummern entfernen: "42)\t" oder "42) " am Zeilenanfang
  x <- sub("^\\s*\\d+\\)\\s*", "", x)

  # --- Tokenisierung ---
  # Jedes Keilschriftzeichen (U+12000..U+1254F) ist ein Token.
  # Jede Zeichenfolge in \u27E8...\u27E9 ist ein Token.
  # Alles andere wird ignoriert.
  tokenize_line <- function(line) {
    chars  <- strsplit(line, "")[[1]]
    n_ch   <- length(chars)
    tokens <- character(0)
    i      <- 1L

    while (i <= n_ch) {
      cp <- utf8ToInt(chars[i])

      if (cp == 0x27E8L) {
        # \u27E8 gefunden - alles bis \u27E9 sammeln
        bracket <- chars[i]
        i <- i + 1L
        while (i <= n_ch && utf8ToInt(chars[i]) != 0x27E9L) {
          bracket <- paste0(bracket, chars[i])
          i <- i + 1L
        }
        if (i <= n_ch) {
          bracket <- paste0(bracket, chars[i])  # \u27E9 anhaengen
          i <- i + 1L
        }
        tokens <- c(tokens, bracket)

      } else if (cp >= 0x12000L && cp <= 0x1254FL) {
        tokens <- c(tokens, chars[i])
        i <- i + 1L

      } else {
        i <- i + 1L
      }
    }
    tokens
  }

  lines_tok <- lapply(x, tokenize_line)
  lines_tok <- lines_tok[lengths(lines_tok) > 0L]

  if (length(lines_tok) == 0L) {
    return(data.frame(frequency   = integer(0),
                      length      = integer(0),
                      combination = character(0),
                      stringsAsFactors = FALSE))
  }

  # --- max_n automatisch bestimmen ---
  max_n <- max(lengths(lines_tok))

  # --- min_freq aufbereiten ---
  if (is.null(min_freq)) {
    min_freq <- c(6L, 4L, 2L)
  }
  min_freq <- as.integer(min_freq)

  # --- C-Kernschleife: N-Gramm-Extraktion, Zaehlung und Maskierung ---
  result <- .Call(C_ngram_core, lines_tok, as.integer(max_n), min_freq)

  # --- Ergebnis zusammenbauen ---
  out <- data.frame(
    frequency   = result$frequency,
    length      = result$length,
    combination = result$combination,
    stringsAsFactors = FALSE
  )

  if (nrow(out) == 0L) {
    return(out)
  }

  out <- out[order(-out$length, -out$frequency), ]
  rownames(out) <- NULL

  # Spaltenreihenfolge: frequency, length, combination
  out[, c("frequency", "length", "combination")]
}

convert_to_dictionary <- function(df, mapping=NULL){

  if(is.null(mapping)){
    path  <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
    mapping <- read.csv2(path, sep=";", na.strings="")
  }

  signs <- list(cuneiform=as.cuneiform(df$sign_name, mapping=mapping), sign_name=df$sign_name)

  df <- aggregate(
    list(count = rep(1, nrow(df))),
    by = list(sign_name = df$sign_name, type = df$type, meaning = df$meaning),
    FUN = sum
  )

  df <- data.frame(sign_name=df$sign_name, row_type="trans.", count=df$count, type=df$type, meaning=df$meaning)

  df_signs <- as.data.frame(signs)
  df_signs <- df_signs[!duplicated(df_signs$cuneiform),]
  df_signs$reading <- ""
  for(i in 1:nrow(df_signs)){
    sign_names <- split_sumerian(df_signs$sign_name[i])$signs
    n <- length(sign_names)
    this_reading <- rep("?", n)
    for(j in 1:n){
      k <- which(mapping$name == sign_names[j])
      if(length(k)>0){
        k <- k[1]
        syllables <- mapping$syllables[k[1]]
        if(n>1){syllables <- str_replace(syllables, "^((?:[^,]*,){2}[^,]*),.*$", "\\1, \u2026")}
        if(str_count(syllables, fixed(","))>0){
          syllables <- paste0("{",syllables,"}")
        }
        this_reading[j] <- syllables
      }
      df_signs[i,"reading"] <- paste0(this_reading, collapse="-")
    }
  }

  df_names    <- data.frame(sign_name=df_signs$sign_name, row_type="cunei.", count=NA, type="", meaning=df_signs$cuneiform)
  df_readings <- data.frame(sign_name=df_signs$sign_name, row_type="reading", count=NA, type="", meaning=df_signs$reading)

  df <- rbind(df, df_names, df_readings)

  ord <- order(df$sign_name, df$row_type,  -df$count)
  df <- df[ord,]
  rownames(df) <- 1:nrow(df)

  return(df)
}

tokenize <- function(x){

  part <- split_sumerian(x)

  if(str_detect(x,"<")){
    df <- data.frame(
      start = which(str_ends(  part$separators, "<")),
      end   = which(str_starts(part$separators, ">"))-1,
      stringsAsFactors = FALSE)

    tok_sign   <- part$signs
    tok_length <- rep(1, length(tok_sign))
    tok_sep    <- part$separators

    for(i in nrow(df):1){
      N <- length(tok_sign)
      start <- df$start[i]
      end   <- df$end[i]
      v <- tok_sign[start:end]
      w <- c(tok_sep[(start+1):end], "")
      this_token <- paste0(v, w, collapse="")

      first_ones <- integer(0)
      last_ones  <- integer(0)

      if(start>1)first_ones <- 1:(start-1)
      if(end<N) last_ones <- (end+1):N

      tok_sign   <- c(tok_sign[first_ones],           this_token,  tok_sign[last_ones])
      tok_sep    <- c(tok_sep[c(first_ones, start)],               tok_sep[c(last_ones, N+1)])
      tok_length <- c(tok_length[first_ones],         end-start+1, tok_length[last_ones])
    }
  }else{
    tok_sign     <- part$signs
    tok_length <- rep(1,length(part$signs))
    tok_sep      <- part$separators
  }

  token <- list(
    sign   = tok_sign,
    length = tok_length,
    sep    = tok_sep
  )
  return(token)
}



mark_skeleton_entries <- function(x){
  token <- tokenize(x)
  
  token$sep <- str_replace_all(token$sep, fixed("<"), "(")
  token$sep <- str_replace_all(token$sep, fixed(">"), ")")

  token$sep <- str_replace_all(token$sep, fixed("("), "{(")
  token$sep <- str_replace_all(token$sep, fixed(")"), ")}")

  N <- length(token$sign)
  for(i in 1:N){
    has_brackets <- str_ends(token$sep[i],fixed("(")) && str_starts(token$sep[i+1],fixed(")"))
    if(!has_brackets){
      token$sep[i] <- paste0(token$sep[i], "(")
      token$sep[i+1] <- paste0(")", token$sep[i+1])
    }
  }
  token$sep <- str_replace_all(token$sep, "  ", " ")

  result <- paste0(c("", token$sign), token$sep, collapse = "")

  return(result)
}





mark_skeleton_entries_alt <- function(x){
  token <- tokenize(x)
  token$sep <- str_replace_all(token$sep, "[<>]", " ")

  N <- length(token$sign)
  for(i in 1:N){
    has_brackets <- str_ends(token$sep[i],fixed("(")) && str_starts(token$sep[i+1],fixed(")"))
    if(!has_brackets){
      token$sep[i] <- paste0(token$sep[i], "(")
      token$sep[i+1] <- paste0(")", token$sep[i+1])
    }
  }
  token$sep <- str_replace_all(token$sep, "  ", " ")

  result <- paste0(c("", token$sign), token$sep, collapse = "")

  return(result)
}

# Hilfsfunktion: Extrahiert die Inhalte der äußersten Klammerpaare
extract_top_level <- function(s) {
  chars <- unlist(strsplit(s, ""))
  depth <- 0
  groups <- character(0)
  start <- NA
  for (i in seq_along(chars)) {
    if (chars[i] == "(") {
      if (depth == 0) start <- i
      depth <- depth + 1
    } else if (chars[i] == ")") {
      depth <- depth - 1
      if (depth == 0) {
        groups <- c(groups, substr(s, start + 1, i - 1))
      }
    }
  }
  groups
}

# Rekursive Funktion
extract_skeleton_entries_recursive <- function(s, depth = 0, token_start = 1) {
  groups <- extract_top_level(s)

  # Unwrap single groups to avoid duplicate entries
  while (length(groups) == 1) {
    groups <- extract_top_level(groups[[1]])
  }

  if (length(groups) == 0) {
    return(data.frame(start=integer(0), n_tokens=integer(0), depth=integer(0), expr=character(0)))
  }

  result <- data.frame(start=integer(0), n_tokens=integer(0), depth=integer(0), expr=character(0))
  current_token <- token_start

  for (g in groups) {
    n <- length(split_sumerian(g)$signs)
    result <- rbind(result, data.frame(start=current_token, n_tokens=n, depth=depth+1, expr=g))
    # Falls noch Klammern enthalten: Rekursion
    if (grepl("\\(", g)) {
      result <- rbind(result, extract_skeleton_entries_recursive(g, depth + 1, current_token))
    }
    current_token <- current_token + n
  }
  result
}

extract_skeleton_entries <- function(x){
  n <- length(split_sumerian(x)$signs)
  root <- data.frame(start = 1L, n_tokens = n, depth = 0L, expr = x,
                     stringsAsFactors = FALSE)
  entries <- extract_skeleton_entries_recursive(x, 0, 1)
  result <- rbind(root, entries)
  result$expr <- gsub("[()]", "", result$expr)
  result
}

translate_line <- function(n, context) {

  line_folder   <- context$line_folder
  dic           <- context$dic
  text          <- context$text
  mapping       <- context$mapping
  min_freq      <- context$min_freq
  sentence_prob <- context$sentence_prob

  line_file <- paste0(line_folder, "/Line_", n, ".txt")

  # 1. Gespeicherte Übersetzung laden falls vorhanden
  if (file.exists(line_file)) {
    x       <- sub("^Structure: ", "", readLines(line_file, n = 1))
    df_fill <- fill_substr_info(line_file)
  } else {
    x       <- n
    df_fill <- NULL
  }

  # 2. Primärwörterbuch aufbauen
  line_files <- list.files(line_folder, "^Line_[0-9]+\\.txt$", full.names = TRUE)
  line_files <- line_files[!str_ends(line_files, paste0("Line_", n, ".txt"))]

  if (is.character(dic) && length(dic) >= 1) {
    primary <- read_dictionary(dic[1])
  } else if (is.data.frame(dic)) {
    primary <- dic
  } else if (is.list(dic) && length(dic) >= 1 && is.data.frame(dic[[1]])) {
    primary <- dic[[1]]
  } else {
    primary <- read_dictionary()
  }

  if (length(line_files) > 0) {
    dic_from_text <- make_dictionary(line_files)
    dics <- list(primary, dic_from_text)
  }else{
    dics <- list(primary)
  }

  # 3. Wörterbuchliste (Primär + Nachschlagewerke)

  if (is.character(dic) && length(dic) >= 2) {
    for (i in 2:length(dic))
      dics[[length(dics) + 1]] <- read_dictionary(dic[i])
  } else if (is.list(dic) && length(dic) >= 2) {
    for (i in 2:length(dic))
      dics[[length(dics) + 1]] <- if (is.character(dic[[i]])) read_dictionary(dic[[i]]) else dic[[i]]
  }

  # 4. Übersetzen
  result <- translate(x, text = text, dic = dics, mapping = mapping,
                      fill = df_fill, min_freq = min_freq,
                      sentence_prob = sentence_prob)

  # 5. Ergebnis speichern
  if (is.null(result)) {
    cat("Note: The translation wasn't saved.\n")
  } else {
    print(result)
    writeLines(result, line_file)
    cat(paste0("\nNote: Result is saved to file '", line_file, "'.\n\n"))
  }
}

# The data frame df has the following columns:
# sign_name: chr: sign names
# type:      chr: grammatical type of the word
# meaning:   chr: the english translation of the word/phrase

beautify_translations <- function(df) {

  # Helper: get the effective base type for a type string
  # For plain types like "S", "Vt" -> base type ("S", "V")
  # For operators like "xS->V" -> result type's base ("V")
  effective_base <- function(type_str) {
    if (is.na(type_str) || type_str == "") return(NA_character_)
    parsed <- parse_type(type_str)
    rt <- get_result_type(parsed)
    if (is.na(rt)) return(NA_character_)
    base_type(rt)
  }

  # Replace operator "S☒→V" with "Vt"
  df$type[df$type == paste0("S\u2612\u2192V")] <- "Vt"

  for (i in seq_len(nrow(df))) {
    eb <- effective_base(df$type[i])
    if (is.na(eb)) next

    meaning <- df$meaning[i]

    if (eb == "S") {
      # Strip leading article from meaning
      meaning <- sub("^(the|a|an|The|A|This|this)\\s+", "", meaning)
      # Strip leading article from inside {} blocks
      meaning <- gsub("\\{(the|a|an|The|A|This|this)\\s+([^}]*)\\}", "{\\2}", meaning)
    } else if (eb == "V") {
      # Strip leading "to" from meaning
      meaning <- sub("^to\\s+", "", meaning)
      # Strip leading "to" from inside {} blocks
      meaning <- gsub("\\{to\\s+([^}]*)\\}", "{\\1}", meaning)
    }

    meaning <- str_replace_all(trimws(meaning), "^,+|,+$", "")

    df$meaning[i] <- trimws(meaning)
  }

  return(df)
}

---
title: "Translating Sumerian Texts"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Translating Sumerian Texts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "ragg_png"
)
```

## 1. Introduction

The first vignette ("Getting Started with sumer") introduced the basic concepts: cuneiform sign representations, dictionary lookup, the type system, text analysis, and the interactive `translate()` function. This vignette describes the complete workflow for translating an entire document and building a custom dictionary from the results.

The workflow consists of the following steps:

1. Set up a project folder for the text
2. Translate line by line using `translate_line()`
3. Build a custom dictionary from the translations
4. Merge it with an existing dictionary for future use

Each translated line improves the dictionary, and the improved dictionary makes the next translation easier. This creates a virtuous cycle.

```{r}
library(sumer)
```


## 2. Setting Up a Project

A translation project consists of a folder with the following structure:

```
project/
  complete_cuneiform_text.txt    # cuneiform text
  lines/                         # translated lines
    Line_1.txt
    Line_2.txt
    ...
```

The **text file** contains the cuneiform text. If you have a transliterated text, you can convert it with `as.cuneiform()`. Each line can optionally begin with a line number (e.g. `8)\t...`). Lines starting with `#` are treated as comments and ignored during analysis.

The **`lines/`** subfolder stores one file per translated line. The file for line *n* is called `Line_n.txt`. These files are created automatically by `translate_line()` when you click "Done".

The package includes an example project for the Sumerian myth "Enki and the World Order". Since the project folder inside the installed package is read-only, we first copy it to a temporary directory:

```{r}
pkg_path <- system.file("extdata", "project", package = "sumer")
file.copy(from = pkg_path, to = tempdir(), recursive = TRUE)
project_dir <- file.path(tempdir(), "project")

cat(head(list.files(project_dir, recursive=TRUE)), sep="\n")
```

Now we can set up the translation context and load the text:

```{r}
ctx <- translation_context(
  project_dir   = project_dir,
  text          = "enki_and_the_world_order.txt",
  dic           = file.path(pkg_path, "sumer-dictionary.txt"),
  mapping       = NULL,
  sentence_prob = 0.25
)

text <- readLines(ctx$text, encoding = "UTF-8")
```

The parameters of `translation_context()` are:

- **`project_dir`**: The project directory with subfolder `lines`.
- **`text`**: The file path (or filename relative to project_dir) of the file with the full cuneiform text.
- **`dic`**: One or more dictionary files. The first has priority for automatic suggestions.
- **`mapping`**: A custom sign mapping table (data frame or file path). If `NULL`, the package's built-in mapping is used. A custom mapping is needed when working with texts that contain signs not covered by the default table.
- **`sentence_prob`**: Corrects for verb underrepresentation in the dictionary. A value of 0.25 means that an estimated 25% of the dictionary entries come from complete sentences; verb probabilities are upweighted accordingly.


## 3. Interactive Translation with `translate_line()`

To translate a specific line, call:

```{r, eval = FALSE}
translate_line(8, ctx)
```

This opens the interactive translation tool. If a file `Line_8.txt` already exists in the `lines/` folder, the previous translation is loaded so that you can continue where you left off. When you click "Done", the result is saved back to the file.

Additionally, `translate_line()` builds on the fly a project-specific dictionary from all previously translated lines in the `lines/` folder that can be used in the Shiny App. The current line is excluded to avoid confirmation bias. This project dictionary appears alongside the primary dictionary in the lookup panel.

The gadget displays several sections on a scrollable page. The following sections describe each of them.


### 3.1 N-gram patterns

The first section displays frequent sign combinations (n-grams) computed from the entire text that appear in the current line. Recurring patterns point to fixed terms or compound words. Combinations that also appear in neighbouring lines are marked with a checkmark in the "Theme" column -- these are thematic connections across lines.

Outside the gadget, the same analysis is available through the functions `ngram_frequencies()` and `mark_ngrams()` (see Vignette 1, Section 5.1).


### 3.2 Sign combination suggestions

This section lists sign combinations from the current line for which one of the dictionaries offers a translation. This helps identify multi-sign expressions that have known meanings and can guide you in setting up bracket structures.


### 3.3 Context

The neighbouring lines (up to 2 before and 2 after the current line) are shown with frequent n-grams marked in curly braces. This reveals patterns that repeat across line boundaries and helps understand the thematic flow of the text.

Outside the gadget, you can mark n-grams in any text with `mark_ngrams()` (see Vignette 1, Section 5.1).


### 3.4 Grammar probabilities

A bar chart shows the probability of each grammatical type for each sign in the current line, based on the dictionary. This is the same visualization produced by `plot_sign_grammar()` (see Vignette 1, Section 5.2). Tall green bars suggest a noun (S), red bars suggest a verb (V), and blue bars suggest an operator producing an attribute (A).


### 3.5 Translation

The main interactive section is the translation area. Here you see the skeleton template with input fields for type and translation. This is where the actual translation work happens -- assigning types, looking up dictionary entries, adjusting the bracket structure, and composing translations. The basic mechanics (green lookup button, brown compose button, bracket input, "Update Skeleton") are described in Vignette 1, Section 6. Verb prefixes and suffixes are explained in Vignette 1, Section 4.3.

Let us demonstrate the translation process on **line 8**, which features verb prefixes.


#### Line 8 in detail

```{r}
i <- which(startsWith(text, "8)"))
cat(text[i], sep = "\n")
cat(as.sign_name(text[i]), sep = "\n")
```

Line 8 contains two sentences:

**First sentence: 𒀭𒂗𒆠𒂗𒃶𒅅𒆷** 

Here, 𒀭𒂗𒆠 forms the subject (Enki), 𒂗 is the object ("cultural leader"), and 𒃶𒅅𒆷 is a complex verb with two prefixes:

| Sign          |  Type  | Translation                                          |
|---------------|--------|------------------------------------------------------|
| an=AN=𒀭      | ☒S→S  | the god of heaven who is S                           |
| en=EN=𒂗      | ☒S→S  | the cultural leader of S                             |
| ki=KI=𒆠      | S      | the Earth                                            |
| en=EN=𒂗      | S      | cultural leader                                      |
| gan=GAN=𒃶    | ☒V→V  | may V                                                 |
| ig=IG=𒅅      | ☒V→V  | V with the task of establishing sustenance of human existence |
| la=LA=𒆷      | Vt     | to equip S                                           |

The verb builds up from the core outward: 𒆷 (Vt) is the core verb, 𒅅 wraps it with additional meaning, and 𒃶 adds modality. The final composed verb is: "may equip S with the task of establishing sustenance of human existence" (Vt).

The bracket structure `((𒀭𒂗𒆠)𒂗(𒃶𒅅𒆷))` groups the subject `(𒀭𒂗𒆠)` and the verb `(𒃶𒅅𒆷)` as units within the sentence. This grouping guides the skeleton hierarchy and makes the compose button work correctly for each unit. Sign combinations that should later be included in a dictionary must be written in brackets.

**Second sentence: 𒀭𒀀𒉣𒈾𒆤𒉈** 

This sentence demonstrates the operator type `S☒→A`, which produces an attribute:

| Sign   Type   | Translation                            |
|---------------|--------|-------------------------------|
| an=AN=𒀭      | ☒S→S  | god of heaven with S          |
| a=A=𒀀        | S      | transformative power          |
| nun=NUN=𒉣    | S      | exaltedness                   |
| na=NA=𒈾      | S☒→S  | being bound to S              |
| ke4=KID=𒆤    | S☒→A  | who is defined as S           |
| ne=NE=𒉈      | V      | to be used as a resource      |

The sign sequence AN.A.NUN.NA.KID denotes the **Anunnaki**, the gods of the Sumerian pantheon. The compositional translation of this sign sequence is: "the gods of heaven with transformative power who are defined as being bound to exaltedness". Here, KID (`S☒→A`) transforms the noun phrase to its left into an attribute (A). The attribute then combines with the remaining noun phrase (S + A -> S) before meeting the verb.


#### The completed Line_8.txt

```
Structure: ((𒀭𒂗𒆠)𒂗(𒃶𒅅𒆷)). ((𒀭𒀀𒉣𒈾𒆤)𒉈).

|an-en-ki-en-gan-ig-la-an-a-nun-na-ke4-ne: SEN: Enki, the god of heaven
  who is the cultural leader of the Earth may equip cultural leaders with
  the task of establishing sustenance of human existence. The Anunnaki,
  the gods of heaven with transformative power who are defined as being
  bound to exaltedness are used as a resource.

|an-en-ki-en-gan-ig-la=AN.EN.KI.EN.GAN.IG.LA=𒀭𒂗𒆠𒂗𒃶𒅅𒆷: SEN: ...
|	an-en-ki=AN.EN.KI=𒀭𒂗𒆠: S: Enki, the god of heaven who is the
	  cultural leader of the Earth
|		an=AN=𒀭: ☒S→S: the god of heaven who is S
|		en=EN=𒂗: ☒S→S: the cultural leader of S
|		ki=KI=𒆠: S: the Earth
|	en=EN=𒂗: S: cultural leader
|	gan-ig-la=GAN.IG.LA=𒃶𒅅𒆷: Vt: may equip S with the task of
	  establishing sustenance of human existence
|		gan=GAN=𒃶: ☒V→V: may V
|		ig=IG=𒅅: ☒V→V: V with the task of establishing sustenance
		  of human existence
|		la=LA=𒆷: Vt: to equip S

|an-a-nun-na-ke4-ne=AN.A.NUN.NA.KID.NE=𒀭𒀀𒉣𒈾𒆤𒉈: SEN: ...
|	an-a-nun-na-ke4=AN.A.NUN.NA.KID=𒀭𒀀𒉣𒈾𒆤: S: The Anunnaki, the gods
	  of heaven with transformative power who are defined as being bound
	  to exaltedness
|		an=AN=𒀭: ☒S→S: god of heaven with S
|		a=A=𒀀: S: transformative power
|		nun=NUN=𒉣: S: exaltedness
|		na=NA=𒈾: S☒→S: being bound to S
|		ke4=KID=𒆤: S☒→A: who is defined as S
|	ne=NE=𒉈: V: to be used as a resource
```


## 4. Working with Dictionaries

### 4.1 Formatting rules for dictionary entries

The line files produced by `translate()` and `translate_line()` use a pipe format where each entry starting with `|` becomes a dictionary entry. When building a dictionary from these files, some automatic normalization is applied. It is helpful to understand these conventions when writing translations:

**Curly braces `{specific meaning}`** in a translation indicate a context-specific interpretation. For example, "container {country}" means that the general compositional meaning is "container" but in this context it refers to "country". When composing entries with the compose button, only the specific meaning inside the curly braces is used for substitution.

**Angle brackets `<comment>`** in a translation contain comments or annotations. The text inside angle brackets is stripped out from translations. This can be used to add explanatory notes, for example: "S &lt;the agent of the transitive verb&gt;".

**Leading articles are removed.** Nouns and noun phrases should be translated as they fit into the sentence, including articles where appropriate. When the dictionary is built, leading articles ("the", "a", "an") at the beginning of a translation string are automatically stripped. This ensures clean dictionary entries while allowing natural English in the line files.

**Verbs should be in base form.** Verb translations should be in their base form, optionally preceded by "to" (e.g. "to create" or "create"). A leading "to" is automatically removed when building the dictionary.


### 4.2 Creating a dictionary

Once you have translated several lines, the line files in the `lines/` folder can be combined into a dictionary. The function `make_dictionary()` reads all line files and aggregates the entries:

```{r}
line_files <- list.files(ctx$line_folder, full.names = TRUE)
head(basename(line_files))

project_dic <- make_dictionary(line_files)
```

The function counts how often each combination of sign name, type, and translation occurs across all files. Signs that appear frequently with the same meaning get higher counts, making them more reliable dictionary entries.

Let us inspect some entries:

```{r}
look_up("AN", project_dic)
```

The sign AN appears with multiple types and meanings, each with a count reflecting how often that particular usage was attested in the translated lines.


### 4.3 Merging dictionaries

A project dictionary built from a single text is most useful when combined with a broader dictionary. The function `merge_dictionaries()` combines two or more dictionaries:

```{r}
dic1 <- read_dictionary()

merged_dic <- merge_dictionaries(dic1, project_dic)
```

Translation entries that agree in sign name, type, and meaning are merged by summing their counts. Cuneiform and reading rows are taken from the first dictionary.

We can verify that the merged dictionary contains entries from both sources:

```{r}
look_up("LAM", dic1)
look_up("LAM", project_dic)
look_up("LAM", merged_dic)
```

**A note on combining dictionaries:** Merging is most meaningful when the underlying texts come from comparable periods and regions of Mesopotamia. The same sign can carry different meanings across time and place, so combining dictionaries from widely different epochs may produce misleading frequency counts.


### 4.4 Saving dictionaries

The completed dictionary can be saved with metadata:

```{r, eval = FALSE}
save_dictionary(
  dic     = merged,
  file    = "my_dictionary.txt",
  author  = "My Name",
  year    = "2026",
  version = "1.0",
  url     = "https://example.com/dictionary"
)
```

The saved dictionary can be loaded in future sessions with `read_dictionary("my_dictionary.txt")` and used as the primary dictionary for new translation projects.


## 5. The Cycle

The workflow described in this vignette forms a self-reinforcing cycle:

1. **Translate** a line using `translate_line()`, guided by the dictionary and n-gram analysis.
2. **Save** the result. The line file is written automatically when you click "Done".
3. **Reuse** the translations. The next time you call `translate_line()`, it automatically builds a project dictionary from all saved lines (excluding the current one). This project dictionary appears alongside the primary dictionary, providing suggestions based on your own previous work.
4. **Build** a dictionary with `make_dictionary()` and merge it with an existing one using `merge_dictionaries()`.
5. **Translate another text** with the improved dictionary.

With each translated line, the dictionary grows. Frequent signs and expressions accumulate higher counts, and the automatic pre-filling of translation templates becomes increasingly accurate. Over time, you build a comprehensive dictionary grounded in your own texts.

---
title: "Getting Started with sumer"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Getting Started with sumer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "ragg_png"
)
```

## 1. Introduction

The `sumer` package provides tools for translating and analyzing Sumerian cuneiform texts. It converts between different text representations, offers dictionary lookup, and includes an interactive translation tool.

Modern scholars typically work with Sumerian texts in **transliterated** form -- a phonetic rendering in Latin characters. For example, the Sumerian word for king is transliterated as `lugal`. However, translation does not depend on pronunciation. The meaning of a sign depends on the sign itself, not on how it is read aloud. Even when there is reason to believe that words with similar pronunciations have similar meanings, dictionaries can be based solely on the cuneiform characters. The same cuneiform sign can have several different readings (transliterations), but it always has exactly one sign name and one cuneiform character.

```{r}
library(sumer)
```


## 2. Representations of Cuneiform Signs

Each cuneiform sign has three representations:

| Representation    | Example             | Description                                        |
|-------------------|---------------------|----------------------------------------------------|
| Transliteration   | `lugal`             | Phonetic transcription in lowercase letters        |
| Sign name         | `LUGAL`             | Canonical name in uppercase letters                |
| Cuneiform         | &#x12217;           | Unicode character (U+12000 to U+12500)             |

The package works internally with cuneiform characters and sign names. Transliteration serves as a convenient input method.

**Note on display:** Cuneiform characters require a font supporting the Unicode Cuneiform block (U+12000--U+12500). In RStudio, the AGG graphics backend should be enabled (Tools > Global Options > General > Graphics > Backend > AGG).


### 2.1 Retrieving sign information

The function `info()` retrieves all available information about a sign or sign sequence:

```{r}
info("lugal")
```

For compound expressions, all contained signs are analyzed:

```{r}
info("d-en-lil2")
```

Each sign (d, en, lil2) is shown with its sign name (AN, EN, KID) and cuneiform character. The `alternatives` column lists all possible readings -- for instance, EN can also be read as `ru12` or `uru16`.


### 2.2 Conversion between representations

Two functions convert entire texts:

```{r}
# Transliteration -> Cuneiform
as.cuneiform("lugal-e")
as.cuneiform(c("d-en-lil2", "an-ki"))

# Transliteration -> Sign names
as.sign_name("lugal-e")
as.sign_name(c("d-en-lil2", "an-ki"))
```

Within a word, hyphens (`-`) separate syllables; dots (`.`) separate sign names; spaces separate words.


## 3. Dictionary Lookup

### 3.1 Loading a dictionary

The package includes a built-in dictionary:

```{r}
dic <- read_dictionary()
```

The vignette "Translating Sumerian Texts" describes how you can create your own dictionary.

### 3.2 Forward lookup: Sumerian -> English

```{r}
look_up("lugal", dic)
```

The output shows the sign name, cuneiform character, translations with frequency counts and grammatical types, and entries for individual signs and substrings. For compound expressions, all partial combinations are looked up as well:

```{r}
look_up("d-suen", dic)
```

### 3.3 Reverse lookup: English -> Sumerian

To find the Sumerian sign for an English term, use `lang = "en"`:

```{r}
look_up("Agga", dic, "en")
```

The reverse lookup searches all translations and displays matching entries with their sign names and cuneiform characters.


## 4. The Type System

Each dictionary entry has a **grammatical type** in addition to its translation. These types describe the function of a sign in a sentence. Since the same sign can serve different functions depending on context, it may have multiple entries with different types.

### 4.1 Basic types

There are three basic types:

| Type    | Name        | Description                                     | Example                  |
|---------|-------------|-------------------------------------------------|--------------------------|
| **S**   | Substantive | Noun phrases and substantives                   | "king", "Earth"          |
| **V**   | Verb        | Verbs and verbal expressions                    | "create", "go"           |
| **A**   | Attribute   | Modifying clauses                               | "who is strong"          |

You can see the different types of a sign with `look_up()`:

```{r}
look_up("an", dic)
```

The sign AN appears both as a noun (S: "sky/heaven") and as an operator that transforms other expressions -- which brings us to the next topic.


### 4.2 Operators

In addition to basic types, there are **operators**. An operator takes one or two expressions of a certain type as arguments and produces an expression of a (possibly different) type. The notation describes where the arguments stand and what type is produced:

```{r, echo = FALSE}
op_df <- data.frame(
  Notation    = c("Sx->V", "xS->S", "Sx->A", "Sx->S", "xV->V", "Vx->V", "SSx->V"),
  Alternative = c("S\u2612\u2192V", "\u2612S\u2192S", "S\u2612\u2192A", "S\u2612\u2192S",
                   "\u2612V\u2192V", "V\u2612\u2192V", "SS\u2612\u2192V"),
  Meaning     = c("Takes an S to the left, produces a V",
                   "Takes an S to the right, produces an S",
                   "Takes an S to the left, produces an A",
                   "Takes an S to the left, produces an S",
                   "Takes a V to the right, produces a V",
                   "Takes a V to the left, produces a V",
                   "Takes two S to the left, produces a V")
)
knitr::kable(op_df, col.names = c("Notation", "Alternative notation", "Meaning"))
```

The `x` marks the position of the operator itself. In `Sx->V`, the argument S stands to the left of the operator; in `xS->S`, the argument stands to the right. Some operators with two arguments (like `SSx->V`) take both arguments from the same side. The symbols `x` and `->` have the alternative Unicode representations `☒` and `→` that may appear in dictionary entries and line files.

In translations, the placeholder **S** (or **V**) stands for the argument. For example, an operator `xS->S` with the translation "community of S" means: take the noun to the right of this sign and insert it where the S placeholder stands. For operators with two arguments of the same type, the placeholders are numbered: `S1` and `S2`.

Let us trace through a concrete example. Consider the expression `un-ma-gi` from "Enki and the World Order" (line 16), which consists of three signs:

```{r, echo = FALSE}
ex_df <- data.frame(
  Syllable  = c("un", "ma", "gi"),
  Sign      = c("\U00012327", "\U00012220", "\U00012100"),
  Type      = c("\u2612S\u2192S", "S", "S\u2612\u2192S"),
  Translation = c("community of S", "container", "the permanent S")
)
knitr::kable(ex_df, col.names = c("Syllable", "Cuneiform Sign", "Type", "Translation"))
```

We build up the noun phrase step by step:

1. **𒈠** is a simple noun (S): "container".
2. **𒄀** is an operator `S☒→S` that takes the S to its left. It binds 𒈠 and replaces the placeholder: "the permanent **container**". The result is again an S.
3. **𒌧**  is an operator `☒S→S` that takes the S to its right. It binds the result from step 2: "community of **the permanent container**". The result is again an S.

Note that operators with arguments at the right-hand side bind stronger than operators with arguments  at the left-hand side. The final noun phrase is: "community of the permanent container" (type S). In this context, "community" refers to the people and "permanent container" refers to the land of Sumer. The expression `un-ma-gi` thus means "the people of Sumer". In the dictionary, such context-dependent meanings are recorded using the notation **"literal meaning {specific meaning}"**. The dictionary entry for this expression would be: "community of the permanent container {people of Sumer}". The literal meaning documents the compositional structure, while the specific meaning in curly braces gives a contextual interpretation.


### 4.3 Verb types

A simple verb (type **V**) stands on its own -- for example, "to be used as a resource". It combines directly with a subject noun phrase (S + V -> SEN) to form a sentence.

Many Sumerian verbs, however, take a noun phrase as their object. The type **Vt** describes such a **transitive verb**: it takes an S as its object and produces a complex intransitive verb (V). The translation contains an S placeholder for the object, for example "to equip S". The resulting V must then be combined with a subject (S) to form a complete sentence (SEN). `Vt` is a generalization of `Sx->V` that also works correctly when the verb has prefixes or suffixes.

In Sumerian, verbs are often preceded by **verb prefixes** -- signs that modify the verb's meaning (expressing modality, aspect, or other nuances). A verb prefix has the type `☒V→V`: it takes a verb to its right and produces a modified verb. Since each prefix produces a V, multiple prefixes can chain together. They bind from right to left, wrapping around the core verb like layers. Conversely, a **verb suffix** has the type `V☒→V` and binds from left to right. Prefixes and suffixes can co-occur.

Consider the verb chain `gan-ig-la` from line 8 of "Enki and the World Order":

```{r, echo = FALSE}
vt_df <- data.frame(
  Syllable   = c("gan", "ig", "la"),
  Sign       = c("\U000120f6", "\U00012145", "\U000121b7"),
  Type      = c("\u2612V\u2192V", "\u2612V\u2192V", "Vt"),
  Translation = c("may V",
                   "V with the task of establishing sustenance of human existence",
                   "to equip S")
)
knitr::kable(vt_df, col.names = c("Syllable", "Cuneiform Sign", "Type", "Translation"))
```

The verb builds up from the core outward:

```
𒃶          𒅅                         𒆷
☒V→V         ☒V→V                        Vt
"may V"      "V with the task ..."        "to equip S"
              └────────────────────────────┘
              𒅅(𒆷) = "equip S with the task ..."
   └──────────────────────┘
𒃶(𒅅(𒆷)) = "may equip S with the task ..."
```

𒆷 (`Vt`: "to equip S") is the core verb. 𒅅 (`☒V→V`) wraps it with additional meaning. 𒃶 (`☒V→V`) adds modality. The final composed verb is: "may equip S with the task of establishing sustenance of human existence" (`Vt`). This complex verb still has its S placeholder -- it will be filled when the verb meets its object during sentence composition.


### 4.4 Composition rules

When two expressions stand side by side and neither is an operator, they are combined by **composition rules**:

| Left | Right | Result | Translation pattern                          |
|------|-------|--------|----------------------------------------------|
| S    | S     | S      | "X of/with Y"                                |
| S    | A     | S      | "X Y" (juxtaposition)                        |
| S    | V     | SEN    | "X Y" (subject + verb, "to" stripped)        |
| SEN  | SEN   | SEN    | "X. Y" (sentences joined with period)        |

The type **SEN** stands for a complete sentence. It arises when a noun phrase (S) meets a verb (V). For example, if "separated groups" (S) is followed by "to be created" (V), the composition strips the "to" and produces the sentence "Separated groups be created." (SEN). This raw composition must then be finished by hand -- in this case, adjusting the verb form to produce "Separated groups are created."

These rules, together with the operator types from the previous sections, are sufficient to translate entire Sumerian sentences from their individual parts. Since in Old Sumerian texts, signs decode phrases rather than syllables, this implies that the Old Sumerian language is actually not a natural language. It is a formal language that can be pronounced in any natural language that is sufficiently complex.


## 5. Text Analysis

### 5.1 N-gram analysis

A good first step when working with a new text is to search for frequently recurring sign combinations (n-grams). Such patterns are valuable clues: if a certain sequence of cuneiform signs appears repeatedly, it is likely a fixed term, a compound word, or an idiomatic expression.

The package includes the example text "Enki and the World Order":

```{r}
path <- system.file("extdata", "project", "enki_and_the_world_order.txt", package = "sumer")
text <- readLines(path, encoding = "UTF-8")
```

The function `ngram_frequencies()` finds recurring combinations:

```{r}
freq <- ngram_frequencies(text, min_freq = c(6, 4, 2))
head(freq, 10)
```

The `min_freq` parameter controls the minimum frequency for different n-gram lengths. The default value `c(6, 4, 2)` means: single signs must occur at least 6 times, pairs at least 4 times, and all longer combinations at least 2 times.

The analysis works from the longest combinations down to the shortest. When a long combination is identified as frequent, duplicated occurrences are masked so that shorter sub-combinations are not falsely counted as frequent just because they are part of the longer combination.

With `mark_ngrams()`, the identified patterns are marked in the text with curly braces:

```{r}
text_marked <- mark_ngrams(text, freq)
cat(text_marked[1:5], sep = "\n")
```

You can also search for a specific pattern in the annotated text:

```{r}
term    <- "IGI.DIB.TU"
pattern <- mark_ngrams(term, freq)
pattern
result  <- text_marked[grepl(pattern, text_marked, fixed = TRUE)]
cat(result, sep = "\n")
```


### 5.2 Grammar probabilities

To understand the structure of a sentence, it is helpful to know which grammatical role each individual sign is likely to play. The function `sign_grammar()` looks up each sign of a string in the dictionary and counts how often it occurs with each grammatical type:

```{r}
sg  <- sign_grammar("a-ma-ru ba-ur3 ra ta", dic)
```

The raw frequencies can be refined into probabilities using a Bayesian model. First, compute the prior distribution of types across all signs in the dictionary:

```{r}
prior <- prior_probs(dic, sentence_prob = 0.25)
```

The `sentence_prob` parameter corrects a systematic bias: if a dictionary was primarily built from noun phrases (rather than complete sentences), verbs are underrepresented. A value of 0.25 means that an estimated 25% of the dictionary entries come from complete sentences. Verb probabilities are then upweighted accordingly.

Next, `grammar_probs()` computes the posterior probabilities for each sign:

```{r}
gp <- grammar_probs(sg, prior, dic)
```

For signs with many dictionary entries, the observed frequencies dominate; for rare signs, the result falls back to the prior distribution. The position of a sign in the sequence is currently not taken into account for calculating probabilities.

The function `plot_sign_grammar()` presents the results as a stacked bar chart:

```{r, fig.width = 7, fig.height = 4}
plot_sign_grammar(gp, sign_names = TRUE)
```

Each bar represents a sign position in the sentence. The colours represent grammatical types: green for nouns (S), red shades for verbs (V) and verb operators, blue shades for attribute operators, orange for adjective-like operators (S☒→S), and grey shades for all other operators. A tall bar in a particular colour indicates that the sign likely has that grammatical function.


The chart can also be saved to a file:

```{r, eval = FALSE}
plot_sign_grammar(gp, output_file = "grammar.png")
```


### 5.3 Grammatical structure of a cuneiform text

Once you have assigned grammatical types to each sign, the function `grammatical_structure()` shows how the parts are grouped according to the operator binding and composition rules. The output uses typed brackets to indicate the role of each group: `()` for substantives (S), `<>` for verbs (V), `[]` for attributes (A), and `{}` for sentences (SEN).

Consider the expression `mec3-ki-aj2-ga-ce-er ce du`:

```{r}
x <- "mec3-ki-aj2-ga-ce-er-ce-du"
x <- paste0(info(x)$reading, collapse = "-")
x
expr <- split_sumerian(x)$signs
type <- c("S", "S", "Sx->A", "xS->A", "S", "Sx->S", "S", "Sx->V")

grammatical_structure(x, type, expr)
```

The following figure shows the same result with colour coding:

```{r, echo = FALSE, out.width = "100%"}
img_path <- system.file("extdata", "grammatical_structure.png", package = "sumer")
if (file.exists(img_path)) knitr::include_graphics(img_path)
```

The figure shows that the sentence has the typical structure of an Old Sumerian sentence with the subject (mec3) at its beginning, followed by some specifications of the subject (here in square brackets), followed by the object (ce), and the verb (du) that absorbs the object. This example demonstrates that many Sumerian proper names are self-explanatory. The term "mec3-ki-aj2-ga-ce-er" stands for the proper name "Meskiagasher", but can also be read as a noun phrase. 

This visualization makes the grammatical structure explicit and can help verify that the type assignments produce a sensible grouping.


## 6. Interactive Translation with `translate()`

The function `translate()` opens an interactive Shiny gadget for translating Sumerian text. To demonstrate, we use a fragment from line 16 of "Enki and the World Order":

```{r, eval = FALSE}
x <- as.cuneiform("cag4-kalam-ma-gi-hal. hal-la-gin7.")
result <- translate(x)
```

This expression contains eight cuneiform signs. Our task is to assign each sign a grammatical type and translation, and then compose them into coherent English sentences.


### 6.1 Recognizing sentence boundaries

The input actually contains two sentences. You must recognize sentence boundaries yourself -- they are not detected automatically. In general, sentence boundaries follow directly after verbs.

A striking feature of this Old Sumerian text is that duplicated signs often mark sentence boundaries: the left occurrence functions as a verb at the end of one sentence, while the right occurrence functions as a noun at the beginning of the next sentence. In our example, the sign HAL (𒄬) appears twice. The first HAL is a verb (Vt: "to split S into separate groups") ending the first sentence, while the second HAL is a noun (S: "separated groups") beginning the second sentence.

The two sentences are:

1. `cag4-kalam-ma-gi-hal` (𒊮𒌧𒈠𒄀𒄬): "The central administration splits the people of Sumer into separate groups."
2. `hal-la-gin7` (𒄬𒆷𒁶): "Places for the separated groups are created."


### 6.2 Structure of the translate gadget

When `translate()` opens, you see a scrollable page with the following sections. The gadget is described in more detail in the vignette "Translating Sumerian Texts".

- **N-gram patterns** -- Frequent sign combinations from the text that appear in the current line. This section is only available when used together with a full text.
- **Sign combination suggestions** -- Sign combinations from the current line for which a dictionary offers a translation.
- **Context** -- Neighbouring lines with frequent n-grams marked. This section is only available when used together with a full text.
- **Grammar probabilities** -- A bar chart showing the likelihood of each grammatical type for each sign (see Section 5.2).
- **Translation** -- The main interactive section, containing:
    - A **bracket input field** with an "Update Skeleton" button for defining the sentence structure (see below).
    - An **interactive skeleton** with input fields for type and translation. Each entry has a green lookup button and a brown compose button.
    - A **dictionary panel** that displays lookup results when the green button is clicked.


### 6.3 Looking up and adopting dictionary entries

When the gadget opens, each sign is pre-filled with its most frequent translation from the dictionary. These suggestions are not always correct -- they are simply the entries with the highest count.

Consider the sign `gi=GI=𒄀` in our example. The automatic suggestion may show a noun entry (S) if that is the most frequent type for 𒄀 in the dictionary. However, in this context, 𒄀 functions as an adjective operator `S☒→S` meaning "permanent S".

To correct this:

1. **Click the green arrow button** next to the 𒄀 entry. This opens the dictionary panel below, showing all entries for 𒄀 with their counts and types.
2. **Find the correct entry** -- in this case, `S☒→S`: "permanent S".
3. **Click the dictionary row** to adopt its type and translation into the skeleton.

If you use multiple dictionaries, the first one has priority for the automatic suggestions. All dictionaries are displayed in the lookup panel, so you can choose from any of them.


### 6.4 Defining structure with brackets

In the bracket input field (next to the "Update Skeleton" button), you can control how the skeleton is structured by inserting brackets:

**Round brackets `( )`** group signs into a compound expression. The skeleton will show an entry for the group in addition to entries for its individual signs. Hence, the brackets tell the tool that these signs form a coherent phrase and adds a line to the skeleton where its translation can be entered.

**Angle brackets `< >`** mark a fixed term (typically a proper name). No individual entries are generated for the signs inside. For instance, `<d-en-ki>` would be treated as a single unit "Enki" without breaking it into AN, EN, KI.

**Curly braces `{ }`** mark operator arguments. In most cases this is not necessary, because operators and their arguments are detected automatically. Only when the automatic detection fails -- for instance in ambiguous groupings -- do you need to specify operator arguments explicitly with curly braces.

After editing the brackets, click **"Update Skeleton"** to rebuild the template. All previously entered translations are preserved.


### 6.5 Composing entries with the compose button

Once you have assigned types and translations to the individual signs of a group, you can click the **brown compose button** (&#x120FB;) next to the parent entry. This automatically combines the children into a composed translation, applying the operator and composition rules described in Section 4.

For example, after filling in the three children of `(un-ma-gi)`:

```{r, echo = FALSE}
ex_df <- data.frame(
  Syllable  = c("un", "ma", "gi"),
  Sign      = c("\U00012327", "\U00012220", "\U00012100"),
  Type      = c("\u2612S\u2192S", "S", "S\u2612\u2192S"),
  Translation = c("community of S", "container", "the permanent S")
)
knitr::kable(ex_df, col.names = c("Syllable", "Cuneiform Sign", "Type", "Translation"))
```

clicking the compose button on the parent entry produces: type **S**, translation **"community of the permanent container"**.

The composed translation often needs manual finishing. In this case, you would edit the translation to add the specific meaning: **"community of the permanent container {people of Sumer}"**. Other common adjustments include adding articles or conjugating verbs. The compose button provides a starting point that you then refine.


### 6.6 Result and next steps

When you click **"Done"**, `translate()` returns a `skeleton` object -- a character vector containing the completed translation in pipe format. This can be saved as a text file:

```{r, eval = FALSE}
result <- translate(x)
writeLines(result, "my_translation.txt")
```

The saved file serves as input for building a custom dictionary (see Vignette 2).

A completed translation for our example looks like this:

```
Structure: (𒊮(𒌧𒈠𒄀)𒄬). (𒄬𒆷𒁶).

|cag4-kalam-ma-gi-hal-hal-la-gin7: SEN: The central administration splits
  the people of Sumer into separate groups. Places for the separated
  groups are created.

|cag4-kalam-ma-gi-hal=ŠA3.UN.MA.GI.HAL: SEN: The central administration
  splits the people of Sumer into separate groups.
|	cag4=ŠA3=𒊮: S: center {the central administration}
|	kalam-ma-gi=UN.MA.GI=𒌧𒈠𒄀: S: community of the permanent container {people of Sumer}
|		kalam=UN=𒌧: ☒S→S: community of S
|		ma=MA=𒈠: S: container
|		gi=GI=𒄀: S☒→S: the permanent S
|	hal=HAL=𒄬: Vt: to split S into separate groups

|hal-la-gin7=HAL.LA.DIM2=𒄬𒆷𒁶: SEN: Places for the separated groups
  are created.
|	hal=HAL=𒄬: S: separated groups
|	la=LA=𒆷: S☒→S: place for S
|	gin7=DIM2=𒁶: V: to be created
```

Each line starting with `|` is a dictionary entry. The indentation reflects the hierarchical structure: the overall sentence at the top, word groups below, and individual signs at the deepest level.

**Learning by example.** The package includes an example project with lines 1--31 of "Enki and the World Order" already translated. You can open any of these lines to study the translations and learn how the type system works in practice:

```{r, eval = FALSE}
path <- system.file("extdata", package = "sumer")

file.copy(
  from = file.path(path, "project"),
  to   = tempdir(),
  recursive = TRUE
)

ctx <- translation_context(
  line_folder   = file.path(tempdir(), "project/lines"),
  text          = file.path(tempdir(), "project/enki_and_the_world_order.txt"),
  dic           = file.path(path, "sumer-dictionary.txt"),
  sentence_prob = 0.25
)

# Open line 16 to see the full translation of our example
translate_line(16, ctx)
```

The second vignette ("Translating Sumerian Texts") describes the complete workflow for translating a document line by line and building a dictionary from the results.

\name{split_sumerian}
\alias{split_sumerian}
\title{Split a String into Sumerian Signs and Separators}
\description{
Splits a transliterated Sumerian text string into its constituent signs and the separators between them. The function recognizes five types of Sumerian sign representations: lowercase transliterations, uppercase sign names, Unicode cuneiform characters, numbers (including the placeholder \code{N}), and the unknown-sign placeholder \code{X}.
}
\usage{
split_sumerian(x)
}
\arguments{
  \item{x}{A character string containing transliterated Sumerian text.}
}
\details{
The function identifies Sumerian signs based on three patterns:

\enumerate{
  \item \strong{Lowercase transliterations} (type 1): Sequences of lowercase letters (a-z) including special characters (\enc{ĝ}{g}, \enc{š}{sz}, ...) and accented vowels (\enc{á}{a}, \enc{é}{e}, \enc{í}{i}, \enc{ú}{u}, \enc{à}{a}, \enc{è}{e}, \enc{ì}{i}, \enc{ù}{u}), optionally followed by a numeric index.

  \item \strong{Uppercase sign names} (type 2): Sequences starting with an uppercase letter, optionally followed by additional uppercase letters, digits, or the characters \code{+}, \code{/}, and \enc{×}{x}.

  \item \strong{Cuneiform characters} (type 3): Unicode characters in the Cuneiform block (U+12000 to U+12500).

  \item \strong{Numbers} (type 4): Integer or decimal numbers (e.g. \code{4}, \code{3.5}), and the standalone letter \code{N} which serves as a placeholder for an arbitrary number.

  \item \strong{Unknown signs} (type 5): The standalone letter \code{X}, which serves as a placeholder for an unreadable sign.
}

The function returns the signs and separators in a format that allows exact reconstruction of the original string using \code{paste0(c("", signs), separators, collapse = "")}.
}
\value{
A list with three components:
  \item{signs}{A character vector containing the extracted Sumerian signs.}
  \item{separators}{A character vector of length \code{length(signs) + 1} containing the separators. The first element contains any text before the first sign, subsequent elements contain text between consecutive signs, and the last element contains any text after the final sign. Empty strings indicate no separator at that position.}
  \item{types}{An integer vector of the same length as \code{signs} indicating the type of each sign: \code{1} for lowercase transliterations, \code{2} for uppercase sign names, \code{3} for cuneiform characters, \code{4} for numbers and the placeholder \code{N}, and \code{5} for the unknown-sign placeholder \code{X}.}
}

\examples{

# Example 1

set.seed(4)

x <- "en-tarah-an-na-ke4"

result <- split_sumerian(x)

result

# Example 2

x <- "en-DARA3.AN.na-ke4"

result <- split_sumerian(x)

result

# Reconstruct the original string
paste0(c("", result$signs), result$separators, collapse = "")

}
\keyword{character}
\keyword{utilities}

\name{make_dictionary}
\alias{make_dictionary}
\title{Create a Sumerian Dictionary from Annotated Text Files}
\description{
Parses Word documents (.docx) or plain text files containing annotated
Sumerian translations and creates a structured dictionary data frame.
The function extracts sign names, their cuneiform representations,
possible readings, and translations with grammatical types.
}
\usage{
make_dictionary(file, mapping = NULL)
}
\arguments{
\item{file}{A character vector of file paths to .docx or text files.
Files must contain translation lines that are formatted as described below.}
\item{mapping}{A data frame containing sign-to-reading mappings with columns
\code{name}, \code{cuneiform} and \code{syllables}. If \code{NULL} (default), the package's built-in
mapping file \code{etcsl_mapping.txt} is used.}
}
\details{
\subsection{Input Format}{
The input files must contain lines starting with \code{|} in the following format:

\code{|sign_name: TYPE: meaning}

or

\code{|equation for sign_name: TYPE: meaning}

 For example:
\preformatted{
|a2-tab: S: the double amount of work performance
|me=ME: S: divine force
|AN: S: god of heaven
|na=NA: Sx->A: whose existence is bound to S
}

Lines not starting with \code{|} are ignored. Only the first entry in an equation of sign names is used for the dictionary. The following notation is suggested for grammatical types:

\itemize{
\item \code{S} for substantives and noun phrases,  (e.g., "the old man in the temple")
\item \code{V} for verbs and decorated verbs (e.g., "to go", "to bring the delivery into the temple")
\item \code{A} for adjectives, attributes and subordinate clauses that further define the subject (e.g., "who/which is weak", "whose resource for sustaining life is grain")
\item \code{Sx->A} for a symbol that transforms the preceding noun phrase into an attribute (e.g., "whose resource for sustaining life is \code{S}"). Other transformations are denoted accordingly.
\item \code{N} for numbers,
\item \code{D} for everything else.
}
}

\subsection{Processing Steps}{
\enumerate{
\item Extracts text from .docx files or reads plain text
\item Filters lines starting with \code{|}
\item Excludes lines containing the unknown-sign placeholder \code{X}
\item Replaces standalone numbers in sign names with \code{N} (suffix digits like the \code{2} in \code{jal2} are not affected)
\item Normalizes sign names and looks up possible readings from the mapping table
\item Aggregates translations and counts occurrences
}
}

\subsection{Output Structure}{
For each unique sign, the output contains:
\itemize{
\item One \code{cunei.} row with the cuneiform character(s)
\item One \code{reading} row with possible phonetic readings
\item One or more \code{trans.} rows with translations, sorted by frequency
}
}
}
\value{
A data frame with the following columns:
\describe{
\item{sign_name}{The normalized Sumerian sign name (e.g., "A", "AN", "ME")}
\item{row_type}{Type of entry: \code{"cunei."} (cuneiform), \code{"reading"} (phonetic readings), or \code{"trans."} (translation)}
\item{count}{Number of occurrences for translations; \code{NA} for cuneiform and reading entries}
\item{type}{Grammatical type (e.g., "S", "V", "Sx->A") for translations; empty for other line types}
\item{meaning}{The cuneiform character(s), reading(s), or translated meaning depending on line_type}
}
}
\seealso{
\code{\link{read_translated_text}} for reading translation files,
\code{\link{convert_to_dictionary}} for the aggregation step,
\code{\link{read_dictionary}} for loading a saved dictionary,
\code{\link{save_dictionary}} for saving a dictionary to file,
\code{\link{look_up}} for searching a dictionary
}
\examples{

# Create a dictionary from a single text document
filename  <- system.file("extdata", "text_with_translations.txt", package = "sumer")
dict <- make_dictionary(filename)

# Use the dictionary
look_up("an", dict)
}

\name{merge_dictionaries}
\alias{merge_dictionaries}
\title{Merge Two or More Sumerian Dictionaries}
\description{
Combines two or more dictionaries (as produced by \code{\link{make_dictionary}}) into a single dictionary. Translation entries that agree in sign name, grammatical type, and meaning are merged by summing their counts. Cuneiform and reading rows are taken from the first dictionary that contains them.
}
\usage{
merge_dictionaries(...)
}
\arguments{
  \item{...}{Two or more dictionaries, each either a data frame (as returned by \code{\link{make_dictionary}} or \code{\link{read_dictionary}}) or a file path to a dictionary file. A single list of dictionaries is also accepted.}
}
\details{
The function processes the three row types differently:

\describe{
  \item{\code{"cunei."} and \code{"reading"}}{Each sign name has at most one row of each type. If multiple input dictionaries contain a cuneiform or reading row for the same sign, the row from the first dictionary (in argument order) is kept.}
  \item{\code{"trans."}}{Rows that share the same \code{sign_name}, \code{type}, and \code{meaning} are merged into a single row whose \code{count} is the sum of the individual counts. Rows that differ in type or meaning are kept as separate entries.}
}

The result is sorted by \code{sign_name}, then \code{row_type}, then descending \code{count}, matching the order produced by \code{\link{make_dictionary}}.

\bold{Note:} Merging dictionaries is most meaningful when the underlying text corpora come from comparable periods and regions of Mesopotamia. Combining dictionaries from widely different epochs or dialects may produce misleading frequency counts, since the same sign can carry different meanings across time and place.
}
\value{
A data frame with five columns in the standard dictionary format:
\describe{
  \item{sign_name}{Character. The sign name (e.g. \code{"A"}, \code{"LUGAL"}).}
  \item{row_type}{Character. One of \code{"cunei."}, \code{"reading"}, or \code{"trans."}.}
  \item{count}{Numeric. The (merged) occurrence count for translation rows; \code{NA} for cuneiform and reading rows.}
  \item{type}{Character. The grammatical type for translation rows (e.g. \code{"S"}, \code{"V"}); empty for cuneiform and reading rows.}
  \item{meaning}{Character. The cuneiform glyph, the reading notation, or the translation text, depending on \code{row_type}.}
}
}
\seealso{
\code{\link{make_dictionary}} for creating a dictionary from a translated text,
\code{\link{read_dictionary}} for loading a dictionary from file,
\code{\link{save_dictionary}} for saving a dictionary to file,
\code{\link{convert_to_dictionary}} for the aggregation step inside \code{make_dictionary}
}
\examples{
# Load the built-in dictionary
dic1 <- read_dictionary()

# Make a dictionary from some translations of "Enki and the World Order"
path <- system.file("extdata", "project/lines", package = "sumer")
dic2 <- make_dictionary(list.files(path, full.names=TRUE)[1:10])

# Merge both dictionaries
dic <- merge_dictionaries(dic1, dic2)

#Test the function
look_up("IL2", dic1)
look_up("IL2", dic2)
look_up("IL2", dic)

}
\keyword{utilities}

\name{info}
\alias{info}
\alias{print.info}
\title{Retrieve Information About Sumerian Signs}
\description{
Analyzes a transliterated Sumerian text string and retrieves detailed information about each sign, including syllabic readings, sign names, cuneiform symbols, and alternative readings.

The function \code{info} computes the result and returns an object of class \code{"info"}. The \code{print} method displays a summary of different text representations in the console.
}
\usage{
info(x, mapping = NULL)

\method{print}{info}(x, flatten = FALSE, ...)
}
\arguments{
\item{x}{
    For \code{info}: a character string of length 1 containing transliterated Sumerian text.

    For \code{print.info}: an object of class \code{"info"}.
  }
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
  \item{flatten}{Logical. If \code{TRUE}, grammar indicators in the text are removed (such as parentheses, brackets, braces, and operators). If \code{FALSE} (the default), the original separators are preserved.}
  \item{...}{Additional arguments passed to the print method (currently unused).}
}
\details{
The function \code{info} performs the following steps:

\enumerate{
  \item Splits the input string into signs and separators using \code{\link{split_sumerian}}.
  \item Standardizes the signs.
  \item Looks up each sign in the mapping table based on its type:
    \itemize{
      \item Type 1 (lowercase): Searches for a matching syllable reading.
      \item Type 2 (uppercase): Searches for a matching sign name.
      \item Type 3 (cuneiform): Searches for a matching cuneiform character.
      \item Type 4 (numbers and \code{N}): Passed through unchanged.
      \item Type 5 (unknown \code{X}): Passed through unchanged.
    }
  \item Returns a data frame with the results, along with the separators stored as an attribute.
}

The mapping table must contain the following columns:
\describe{
  \item{syllables}{Comma-separated list of possible syllabic readings for the sign. The first reading is used as the default.}
  \item{name}{The canonical sign name in uppercase.}
  \item{cuneiform}{The Unicode cuneiform character.}
}

The \code{print} method displays each sign with its name and alternative readings, followed by three text representations: syllables, sign names, and cuneiform text.
}
\value{
\code{info} returns a data frame of class \code{c("info", "data.frame")} with one row per sign and the following columns:
  \item{reading}{The syllabic reading of the sign. For lowercase input, this is the standardized input; for other types, this is the default syllable from the mapping.}
  \item{sign}{The Unicode cuneiform character corresponding to the sign.}
  \item{name}{The canonical sign name in uppercase.}
  \item{alternatives}{A comma-separated string of all possible syllabic readings for the sign.}

The data frame has an attribute \code{"separators"} containing the separator characters between signs.

\code{print.info} prints the following to the console and returns \code{x} invisibly:
\describe{
  \item{Sign table}{Each sign with its cuneiform symbol, name, and alternative readings.}
  \item{syllables}{The text with syllabic readings, using hyphens as separators within words.}
  \item{sign names}{The text with sign names, using periods as separators within words.}
  \item{cuneiform text}{The text rendered in Unicode cuneiform characters, with hyphens and periods removed.}
}
}
\note{
If no custom mapping is provided, the function loads the internal mapping file included with the \pkg{sumer} package.
}
\seealso{
\code{\link{split_sumerian}} for splitting Sumerian text into signs,
\code{\link{as.cuneiform}} for converting to cuneiform characters,
\code{\link{as.sign_name}} for converting to sign names
}
\examples{
library(stringr)

# Basic usage - compute and print
info("lugal-e")

# Store the result for further processing
result <- info("an-ki")
result

# Access the underlying data frame
result$sign
result$name

# Print with and without flattened separators
result <- info("(an)na")
print(result)
print(result, flatten = TRUE)

# Using a custom mapping table
path <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
my_mapping <- read.csv2(path, sep=";", na.strings="")
info("an-ki", mapping = my_mapping)
}
\keyword{utilities}
\keyword{character}

\name{sign_grammar}
\alias{sign_grammar}
\title{Grammatical Type Frequencies for Each Sign in a Sumerian Sentence}
\description{
For each cuneiform sign in a Sumerian sentence, looks up the dictionary to
determine the frequency of each individual grammatical type (e.g., \code{S},
\code{V}, \code{Sx->S}, \code{xS->A}). Returns a data frame with one row
per sign per grammatical type.
}
\usage{
sign_grammar(x, dic, mapping = NULL)
}
\arguments{
  \item{x}{A single character string containing a Sumerian sentence
    (cuneiform, sign names, or transliteration).}

  \item{dic}{A dictionary data frame as returned by
    \code{\link{read_dictionary}}.}

  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
}
\details{
The function converts the input to cuneiform, splits it into individual
signs, and looks up each sign in the dictionary. For each sign, the
translations are grouped by their individual type string
(e.g., \code{"S"}, \code{"V"}, \code{"Sx->S"}, \code{"xS->A"}).

For each type the dictionary \code{count} values are summed. If a
translation entry has no count, it is treated as 1.

The set of types returned is the union of all types found across all
signs in the sentence. Each sign gets one row per type, even if the
count is 0 for that type.
}
\value{
A data frame with columns:
\describe{
  \item{position}{Integer. Position of the sign in the sentence.}
  \item{sign_name}{Character. The sign name (e.g., \code{"KA"}).}
  \item{cuneiform}{Character. The cuneiform character.}
  \item{type}{Character. The grammar type string (e.g., \code{"S"},
    \code{"V"}, \code{"Sx->S"}).}
  \item{n}{Integer. Sum of dictionary counts for this sign and this type.}
}
}
\examples{
dic <- read_dictionary()

# Analyse a sentence
sg <- sign_grammar("a-ma-ru ba-ur3 ra", dic)
print(sg)

# Use with cuneiform input
x<-"\U00012000\U000121AD"
print(x)
sg <- sign_grammar(x, dic)
print(sg)
}
\seealso{
\code{\link{grammar_probs}} for Bayesian posterior probabilities,
\code{\link{plot_sign_grammar}} for visualising the result,
\code{\link{read_dictionary}} for loading a dictionary,
\code{\link{as.cuneiform}} for cuneiform conversion.
}
\keyword{utilities}

\name{as.sign_name}
\alias{as.sign_name}
\alias{as.sign_name.default}
\alias{as.sign_name.character}
\alias{print.sign_name}

\title{Convert Transliterated Sumerian Text to Sign Names}
\description{
Converts transliterated Sumerian text to canonical sign names in uppercase notation. This is a generic function with a method for character vectors.
}
\usage{
as.sign_name(x, ...)

\method{as.sign_name}{default}(x, ...)

\method{as.sign_name}{character}(x, mapping = NULL, ...)

\method{print}{sign_name}(x, ...)
}
\arguments{
  \item{x}{
    For \code{as.sign_name}: An object to be converted to sign names. Currently, only character vectors are supported.

    For \code{print.sign_name}: An object of class \code{"sign_name"}.
  }
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded. Only used by the character method.}
  \item{...}{Additional arguments passed to methods.}
}
\details{
The function processes each element of the input character vector by:
\enumerate{
  \item Calling \code{\link{info}} to look up sign information for each transliterated sign.
  \item Extracting the canonical sign names for each sign.
  \item Reconstructing the text using the original separators, but replacing hyphens with periods to follow standard sign name notation.
}

The default method throws an error for unsupported input types.
}
\value{
  \code{as.sign_name} returns a character vector of class \code{c("sign_name", "character")} with the sign name representation of each input element.

  \code{print.sign_name} displays a character vector of class \code{"sign_name"}.
}
\seealso{
\code{\link{as.cuneiform}} for converting to cuneiform characters,
\code{\link{info}} for retrieving detailed sign information,
\code{\link{split_sumerian}} for splitting Sumerian text into signs
}
\examples{
# Convert transliterated text to sign names
as.sign_name(c("lugal-e", "an-ki"))

# Load transliterated text from a file
file <- system.file("extdata", "transliterated-text.txt", package = "sumer")
x <- readLines(file)
cat(x, sep="\n")

# Convert transliterated text to sign names
as.sign_name(x)

# Using a custom mapping table
path <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
my_mapping <- read.csv2(path, sep=";", na.strings="")
as.sign_name("lugal", mapping = my_mapping)

}
\keyword{utilities}
\keyword{character}
\keyword{methods}

\name{mark_ngrams}
\alias{mark_ngrams}
\title{Mark N-gram Combinations in Cuneiform Text}
\description{
Takes a character vector of Sumerian text and marks all n-gram
combinations (from \code{\link{ngram_frequencies}}) with curly braces.
Longer combinations are marked first, shorter ones afterwards
(including inside already-marked regions).
}
\usage{
mark_ngrams(x, ngram, mapping = NULL)
}
\arguments{
\item{x}{A character vector of Sumerian text (transliteration,
sign names, or cuneiform). Will be converted to cuneiform
internally.}

\item{ngram}{A data frame as returned by \code{\link{ngram_frequencies}},
with at least columns \code{combination} and \code{length}.}

\item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
}
\value{
A character vector of cuneiform text with n-gram combinations
enclosed in curly braces and surrounded by spaces.
}
\details{
The function first converts \code{x} to cuneiform (if not already)
and removes spaces and brackets \code{()[]{}}.

Single-sign n-grams (\code{length == 1}) are excluded from marking.
The remaining n-grams are sorted descending by \code{length} and each
occurrence of a combination is replaced with \code{ \{combination\} }
(space, open brace, combination, close brace, space).

Shorter n-grams may be marked inside already-marked longer n-grams
(nesting is allowed).
}
\examples{

# Load the example text of "Enki and the World Order"
path  <- system.file("extdata", "project", "enki_and_the_world_order.txt", package = "sumer")
text <- readLines(path, encoding="UTF-8")
cat(text[1:10],sep="\n")

# Find combinations that appear at least 6 times in the text
freq <- ngram_frequencies(text, min_freq = 6)
freq[1:10,]

# Mark these combinations in the text
text_marked <- mark_ngrams(text, freq)
cat(text_marked[1:10], sep="\n")

# You can enter transliterated text
x <- "kij2-sig unu2 gal d-re-e-ne-ka me-te-ac im-mi-ib-jal2"
mark_ngrams(x, freq)

# Find all occurences of a pattern in the annotated text
term     <- "IGI.DIB.TU"
(pattern <- mark_ngrams(term, freq))
result   <- text_marked[grepl(pattern, text_marked, fixed=TRUE)]
cat(result, sep="\n")


}
\seealso{
\code{\link{ngram_frequencies}}
}

\name{guess_substr_info}
\alias{guess_substr_info}
\title{Look Up Translations for All Substrings of a Sumerian Text}
\description{
Converts a Sumerian text string into cuneiform tokens, generates all contiguous substrings, and looks up the most frequent translation for each substring in one or more dictionaries.
}
\usage{
guess_substr_info(x, dic, mapping = NULL)
}
\arguments{
  \item{x}{A character string of length 1 containing Sumerian text (transliteration, sign names, or cuneiform characters). May contain brackets as used by \code{\link{skeleton}}.}
  \item{dic}{A dictionary, a list of dictionaries, or a character vector of file paths to dictionary files. If file paths are given, each file is loaded with \code{\link{read_dictionary}}. Dictionaries are tried in order: the first dictionary that contains a translation for a given substring wins.}
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
}
\details{
The function performs the following steps:

\enumerate{
  \item If \code{dic} is a character vector of file paths, the dictionaries are loaded with \code{\link{read_dictionary}}. If \code{dic} is a single data frame, it is wrapped in a list.
  \item The input string \code{x} is converted to cuneiform with \code{\link{as.cuneiform}} and split into individual tokens with \code{\link{split_sumerian}}.
  \item A data frame of all contiguous substrings is created with \code{\link{init_substr_info}}.
  \item A \code{sign_name} column is added by converting each substring expression with \code{\link{as.sign_name}}.
  \item For each substring, the dictionaries are searched in order. The most frequent translation (highest \code{count} among rows with \code{row_type == "trans."}) from the first dictionary that contains a match is used to fill in the \code{type} and \code{translation} columns.
  \item Single-token entries of type 4 (numbers and \code{N}) receive type \code{"S"} and their numeric value as translation, regardless of dictionary content.
}
}
\value{
A data frame with one row per substring and the following columns:
  \item{start}{Integer. The token position of the first token in the substring (1-based).}
  \item{n_tokens}{Integer. The number of tokens in the substring.}
  \item{expr}{Character. The concatenated cuneiform tokens of the substring.}
  \item{type}{Character. The grammatical type of the most frequent translation (e.g. \code{"S"}, \code{"V"}), or \code{""} if no translation was found.}
  \item{translation}{Character. The most frequent translation from the dictionaries, or \code{""} if no translation was found.}
  \item{sign_name}{Character. The sign name representation of the substring.}

The rows are ordered as in \code{\link{init_substr_info}} (by \code{n_tokens} descending, then \code{start} ascending), so that row indices can be computed with \code{\link{substr_position}}.
}
\seealso{
\code{\link{init_substr_info}} for creating the substring data frame,
\code{\link{substr_position}} for computing row indices,
\code{\link{read_dictionary}} for loading dictionaries,
\code{\link{look_up}} for interactive dictionary lookup,
\code{\link{skeleton}} for creating translation templates
}
\examples{
# Load the built-in dictionary
dic <- read_dictionary()

# Look up translations for all substrings
x <- "lugal kur-ra-ke4"
df <- guess_substr_info(x, dic)

# Show rows that have a translation
df[df$translation != "", ]

# Use multiple dictionaries (ordered by reliability -> first match wins)
file1 <- system.file("extdata", "sumer-dictionary.txt", package = "sumer")
df <- guess_substr_info(x, file1)

}
\keyword{utilities}
\keyword{character}

\name{ngram_frequencies}
\alias{ngram_frequencies}
\title{Frequency Analysis of Cuneiform Sign Combinations (N-grams)}
\description{
Analyzes a Sumerian text for frequently occurring cuneiform sign combinations
(n-grams). The input can be either cuneiform text or transliterated text
(which is automatically converted to cuneiform via \code{\link{as.cuneiform}}).
The analysis starts with the longest combinations and works down to single
signs, masking already-counted occurrences to avoid reporting subsequences
that are only frequent because they are part of a longer frequent combination.
N-grams are searched within lines only (not across line boundaries).
}
\usage{
ngram_frequencies(x, min_freq = c(6, 4, 2), mapping = NULL)
}
\arguments{
  \item{x}{Character vector whose elements are the lines of a Sumerian text.
    The input can be either cuneiform characters or transliterated text. If no
    cuneiform characters (U+12000 to U+1254F) are detected, the input is
    automatically converted using \code{\link{as.cuneiform}}.
    Lines starting with \code{#} are treated as comments and ignored.
    Optional line numbers at the beginning of a line (e.g., \code{"42)\t"})
    are automatically removed. Spaces are removed before tokenization.}

  \item{min_freq}{Integer vector specifying minimum frequencies (default:
    \code{c(6, 4, 2)}). The i-th value specifies the minimum frequency for
    combinations of length i. For lengths beyond the vector's length, the last
    value is used.

    The default \code{c(6, 4, 2)} means: single signs must occur at least 6
    times, pairs at least 4 times, and all longer combinations at least 2
    times.}

  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
}
\details{
A \dQuote{sign} is defined as either a single cuneiform Unicode character
(U+12000 to U+1254F) or a character sequence enclosed in mathematical
angle brackets (U+27E8 ... U+27E9), which is treated as a single token.
All other characters (spaces, X, numbers, punctuation, etc.) are skipped
during tokenization.

The maximum n-gram length is automatically determined as the length of the
longest tokenized line in the input.

The analysis proceeds from the longest combinations down to single signs.
When a combination is identified as frequent (i.e., meets the minimum
frequency threshold), all occurrences except the first are masked before
continuing with shorter combinations. This prevents subsequences from being
reported as frequent when their frequency is solely due to a longer frequent
combination.
}
\value{
A data frame with three columns, sorted by descending length, then descending
frequency:
  \item{frequency}{Integer. The number of occurrences of the combination.}
  \item{length}{Integer. The number of signs in the combination.}
  \item{combination}{Character. The cuneiform sign combination
    (e.g., \code{"\U0001202D\U00012097\U000120A0"}).}
}
\examples{
# Read the text "Enki and the World Order"

path  <- system.file("extdata", "project", "enki_and_the_world_order.txt", package = "sumer")
text <- readLines(path, encoding="UTF-8")

cat(text[1:10],sep="\n")

# Find combinations that appear at least 6 times in the text
freq <- ngram_frequencies(text, min_freq = 6)

freq[1:10,]
}
\seealso{
\code{\link{as.sign_name}} for converting cuneiform to sign names,
\code{\link{as.cuneiform}} for converting transliterations to cuneiform,
\code{\link{split_sumerian}} for tokenizing transliterated text.
}
\keyword{utilities}
\keyword{univar}

\name{grammar_probs}
\alias{grammar_probs}
\title{Posterior Probabilities of Grammatical Types for Each Sign}
\description{
For each cuneiform sign in a sentence, computes Bayesian posterior
probabilities for all grammatical types, combining prior beliefs from
\code{\link{prior_probs}} with observed dictionary frequencies. The
dictionary counts are corrected for verb underrepresentation using the
\code{sentence_prob} stored in the prior.
}
\usage{
grammar_probs(sg, prior, dic, alpha0 = 1)
}
\arguments{
  \item{sg}{A data frame as returned by \code{\link{sign_grammar}}.}

  \item{prior}{A named numeric vector as returned by
    \code{\link{prior_probs}}, with a \code{sentence_prob} attribute.}

  \item{dic}{A dictionary data frame as returned by
    \code{\link{read_dictionary}}.}

  \item{alpha0}{Numeric (>= 0). Strength of the prior (pseudo sample
    size). Larger values pull the posterior towards the prior. When
    \code{alpha0 = 0}, the result is purely data-driven. Default: 1.}
}
\details{
For each sign at position \eqn{i} in the sentence, the function computes:

\enumerate{
  \item The raw dictionary counts \eqn{n_k} for each grammar type \eqn{k}.
  \item A correction factor \eqn{x_k = 1 / \mathrm{sentence\_prob}} for
    verb-like types, \eqn{x_k = 1} otherwise. The corrected counts are
    \eqn{m_k = n_k \cdot x_k} with total \eqn{M = \sum_k m_k}.
  \item The posterior probability (Dirichlet-Multinomial model):
    \deqn{\theta_k = \frac{\alpha_0 \, p_k + m_k}{\alpha_0 + M}}
    where \eqn{p_k} is the prior probability from \code{prior_probs()}.
}

For signs not in the dictionary (\eqn{M = 0}), the posterior equals the
prior. For signs with many observations (\eqn{M \gg \alpha_0}), the
posterior is dominated by the data.
}
\value{
A data frame with columns:
\describe{
  \item{position}{Integer. Position of the sign in the sentence.}
  \item{sign_name}{Character. The sign name.}
  \item{cuneiform}{Character. The cuneiform character.}
  \item{type}{Character. The grammar type (e.g., \code{"S"}, \code{"V"},
    \code{"Sx->S"}).}
  \item{prob}{Numeric. Posterior probability for this type at this
    position.}
  \item{n}{Numeric. Number of counts in the dictionary.}
}
}
\examples{
dic   <- read_dictionary()
sg    <- sign_grammar("a-ma-ru ba-ur3 ra", dic)
prior <- prior_probs(dic, sentence_prob = 0.25)
gp    <- grammar_probs(sg, prior, dic, alpha0 = 1)
print(gp)
}
\seealso{
\code{\link{prior_probs}} for computing the prior,
\code{\link{sign_grammar}} for the input data,
\code{\link{plot_sign_grammar}} for visualisation.
}
\keyword{utilities}

\name{translate}
\alias{translate}
\title{Interactive Translation Tool for Sumerian Text}
\description{
Opens an interactive Shiny gadget for translating a single line of Sumerian cuneiform text. The page displays four sections on a single scrollable page: n-gram patterns, context with neighbouring lines, grammar probabilities, and an interactive skeleton with dictionary lookup. When the user clicks \dQuote{Done}, the function returns a \code{\link{skeleton}} object with the updated translations.
}
\usage{
translate(x, text = NULL, dic = NULL, mapping = NULL, fill = NULL,
          min_freq = c(6, 4, 2), sentence_prob = 1.0,
          viewer = shiny::paneViewer())
}
\arguments{
  \item{x}{A single Sumerian text string (transliteration, sign names, or cuneiform), or an integer line number indexing into \code{text}.}
  \item{text}{A character vector containing the full text being translated (one line per element), a file path to load with \code{readLines()}, or \code{NULL}. Lines may start with numbering like \code{"12)\t..."} or \code{"12. ..."}. Required when \code{x} is an integer; optional otherwise. If a single string that is an existing file path, it is loaded automatically.}
  \item{dic}{A dictionary (data.frame), a list of dictionaries, or a character vector of file paths to dictionary files. If file paths are given, each is loaded with \code{\link{read_dictionary}}. If \code{NULL}, the built-in dictionary is loaded via \code{read_dictionary()}.}
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}, or a file path to a semicolon-separated mapping file. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
  \item{fill}{A pre-computed substring info data frame (as from \code{\link{init_substr_info}} or \code{\link{guess_substr_info}}). If \code{NULL}, it is computed automatically via \code{\link{guess_substr_info}}.}
  \item{min_freq}{Minimum frequency thresholds passed to \code{\link{ngram_frequencies}}. A numeric vector where the \emph{i}-th element is the minimum frequency for n-grams of length \emph{i}. Default is \code{c(6, 4, 2)}.}
  \item{sentence_prob}{Probability that a randomly chosen sign is part of a sentence with a verb, passed to \code{\link{prior_probs}}. Default is \code{1.0}.}
  \item{viewer}{A Shiny viewer function that controls where the gadget window is opened. The default is \code{shiny::paneViewer()} which uses the RStudio Viewer pane. Use \code{shiny::browserViewer()} to open in the system browser, or \code{shiny::dialogViewer("Translate", width = 842, height = 900)} to open a fixed-size dialog in RStudio}
}
\details{
The gadget opens in the viewer specified by the \code{viewer} parameter (by default the RStudio Viewer pane) and displays four sections on a single scrollable page. The first three sections (N-grams, Context, Grammar) can be collapsed individually. A sticky navigation menu at the top allows jumping to each section.

\describe{
  \item{N-gram Patterns}{Displays a merged table of n-gram combinations that appear in the current line: n-grams of length 2 or more from the full text (controlled by \code{min_freq}), combined with shared n-grams found in neighbouring lines. A \dQuote{Theme} column marks n-grams shared with the context. Frequencies refer to the full text.}
  \item{Context}{Shows neighbouring lines (up to 2 before and after) with frequent n-grams marked. Only available when \code{text} is provided and the line index is known.}
  \item{Grammar Probabilities}{Displays a bar chart of grammar probabilities for each sign in the line, computed via \code{\link{grammar_probs}} with the given \code{sentence_prob}.}
  \item{Translation}{The main interactive section with dictionary selection checkboxes, a bracket input field for editing the skeleton structure, an interactive skeleton display with type and translation fields, and a dictionary lookup panel. Clicking a dictionary row adopts its type and translation into the selected skeleton entry.}
}

When the line contains multiple sentences (separated by dots in the transliteration), skeleton entries belonging to different sentences are displayed with alternating background colours.

The bracket input field allows the user to add or modify brackets \code{()}, \code{<>}, \code{\{\}} to control the grouping structure of the skeleton. Pressing \dQuote{Update Skeleton} rebuilds the skeleton display while preserving all translations in the \code{fill} data frame.
}
\value{
A \code{skeleton} object (character vector of class \code{c("skeleton", "character")}), generated by calling \code{\link{skeleton}} with the final bracket string and updated \code{fill} data frame. Returns \code{invisible(NULL)} if the user closes the window without clicking \dQuote{Done}.
}
\note{
Requires \pkg{shiny} (listed in \code{Imports}).
By default, the gadget opens in the RStudio Viewer pane. To get a resizable window or a more stable connection (e.g. when the computer may enter standby), use \code{viewer = shiny::browserViewer()}.
}
\seealso{
\code{\link{skeleton}} for creating translation templates,
\code{\link{guess_substr_info}} for pre-computing substring translations,
\code{\link{look_up}} for interactive dictionary lookup,
\code{\link{ngram_frequencies}} for n-gram analysis,
\code{\link{grammar_probs}} for grammar probability computation,
\code{\link{prior_probs}} for prior probability computation,
\code{\link{mark_ngrams}} for marking n-grams in text
}
\examples{
\dontrun{
# Basic usage with a transliterated string
result <- translate("lugal kur-ra-ke4")


# Full example with package data
x <- "<d-nu-dim2-mud> ki a. jal2 (e2-kur) ra. gaba jal2. an ki a"

dict_file <- system.file("extdata", "sumer-dictionary.txt", package = "sumer")
text_file <- system.file("extdata", "project", "enki_and_the_world_order.txt", package = "sumer")

result <- translate(x,
                text = text_file,
                dic = dict_file,
                min_freq = c(6, 4, 2),
                sentence_prob = 0.25)
print(result)



# Open in system browser (resizable, survives standby)

x <- 9

result <- translate(x,
                text = text_file,
                dic = dict_file,
                min_freq = c(6, 4, 2),
                sentence_prob = 0.25,
                viewer = shiny::browserViewer())
print(result)

}
}
\keyword{utilities}

\name{skeleton}
\alias{skeleton}
\alias{print.skeleton}
\title{Create a Translation Template for Sumerian Text}
\description{
Creates a structured template (skeleton) for translating Sumerian text. The template displays each token and subexpression with its syllabic reading, sign name, and cuneiform representation, providing a framework for adding translations.

The input may contain three types of brackets to control how the template is generated (see Details). Optionally, the template can be pre-filled with translations from one or more dictionaries using \code{\link{guess_substr_info}}.

The function \code{skeleton} computes the template and returns an object of class \code{"skeleton"}. The \code{print} method displays the template in the console.
}
\usage{
skeleton(x, mapping = NULL, fill = NULL, space = FALSE)

\method{print}{skeleton}(x, ...)
}
\arguments{
  \item{x}{For \code{skeleton}: A character string of length 1 containing transliterated Sumerian text (transliteration, sign names, or cuneiform characters). Tokens may be grouped with brackets to control template generation (see Details).

  For \code{print.skeleton}: An object of class \code{"skeleton"} as returned by \code{skeleton}.}
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
  \item{fill}{A data frame as returned by \code{\link{guess_substr_info}}, containing translations and grammatical types for all substrings of \code{x}. If provided, the template lines are pre-filled with the corresponding type and translation. If \code{NULL} (the default), the template lines are left empty.}
  \item{space}{Logical. If \code{TRUE}, an empty line is inserted before each entry at nesting depth 1, visually separating top-level groups. Defaults to \code{FALSE}.}
  \item{...}{Additional arguments passed to the print method (currently unused).}
}
\details{
The function generates a hierarchical template from a Sumerian text string. The input is first converted to cuneiform with \code{\link{as.cuneiform}}. The input string may contain three types of brackets that control how entries in the template are generated:

\describe{
  \item{Angle brackets \code{< >}}{The enclosed token sequence is treated as a fixed term. No individual skeleton entries are generated for the tokens inside. For example, \code{<d-nu-dim2-mud>} is treated as a single unit.}
  \item{Round brackets \code{( )}}{The enclosed token sequence is a coherent term for which a single skeleton entry is generated, in addition to entries for its individual tokens. Nesting is allowed.}
  \item{Curly braces \code{\{ \}}}{Ignored during skeleton generation. They can be used in the input to indicate which tokens serve as arguments to an operator, but this information is not needed for the skeleton.}
}

In addition, a skeleton entry is generated for every individual token that does not appear inside angle brackets.

Each line in the resulting template follows the format:

\code{|[tabs]reading=SIGN.NAME=cuneiform:type:translation}

When \code{fill} is not provided, the type and translation fields are left empty:

\code{|[tabs]reading=SIGN.NAME=cuneiform::}

The template should then be filled in as follows:
\itemize{
  \item Between the two colons: the grammatical type of the expression (e.g., \code{S} for noun phrases, \code{V} for verbs). See \code{\link{make_dictionary}} for details.
  \item After the second colon: the translation.
}

The indentation level (number of tabs) reflects the nesting depth: top-level entries have no indentation, their sub-entries have one tab, and so on.

The template format is designed to be saved as a text file (.txt) or Word document (.docx), edited manually, and then used as input for \code{\link{make_dictionary}} to create a custom dictionary.

If \code{fill} is provided, the function validates that \code{fill} matches \code{x}: the cuneiform tokens of the first row in \code{fill} must be identical to the tokens of \code{x}, and the number of rows must equal \eqn{N(N+1)/2} where \eqn{N} is the number of tokens.
}
\value{
\code{skeleton} returns a character vector of class \code{c("skeleton", "character")} containing the template lines. The first line is the header with the full reading of the input, followed by one line per skeleton entry. If \code{space = TRUE}, empty strings are inserted as separator lines.

\code{print.skeleton} prints the template to the console (one line per element) and returns \code{x} invisibly.
}
\seealso{
\code{\link{guess_substr_info}} for generating the \code{fill} data frame,
\code{\link{mark_skeleton_entries}} for the bracket normalization step,
\code{\link{extract_skeleton_entries}} for the hierarchical extraction step,
\code{\link{substr_position}} for computing row indices in the fill data frame,
\code{\link{look_up}} for looking up translations of Sumerian signs and words,
\code{\link{make_dictionary}} for creating a dictionary from filled-in templates,
\code{\link{info}} for retrieving detailed sign information
}
\examples{
# Create an empty template
x <- "<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra. gaba jal2. an ki a"
skeleton(x)

# Pre-fill the template with dictionary translations
dic <- read_dictionary()
fill <- guess_substr_info(x, dic)
skeleton(x, fill = fill)

# Use spacing to visually separate top-level groups
skeleton(x, fill = fill, space = TRUE)
}
\keyword{utilities}
\keyword{character}

\name{read_translated_text}
\alias{read_translated_text}
\title{Read Annotated Sumerian Translations from Text Files}
\description{
Reads Word documents (.docx) or plain text files containing annotated
Sumerian translations and extracts sign names, grammatical types, and
meanings into a structured data frame.
}
\usage{
read_translated_text(file, mapping=NULL)
}
\arguments{
\item{file}{A character vector of file paths to .docx or text files.
Files must contain translation lines that are formatted as described below.}
\item{mapping}{A data frame containing sign-to-reading mappings with columns
\code{name}, \code{cuneiform} and \code{syllables}. If \code{NULL} (default), the package's built-in
mapping file \code{etcsl_mapping.txt} is used.}
}

\details{
\subsection{Input Format}{
The input files must contain lines starting with \code{|} in the following format:

\code{|sign_name: TYPE: meaning}

or

\code{|equation for sign_name: TYPE: meaning}

 For example:
\preformatted{
|a2-tab: S: the double amount of work performance
|me=ME: S: divine force
|AN: S: god of heaven
|na=NA: Sx->A: whose existence is bound to S
}

Lines not starting with \code{|} are ignored. Only the first entry in an equation of sign names is extracted. The following notation is suggested for grammatical types:

\itemize{
\item \code{S} for substantives and noun phrases,  (e.g., "the old man in the temple")
\item \code{V} for verbs and decorated verbs (e.g., "to go", "to bring the delivery into the temple")
\item \code{A} for adjectives, attributes and subordinate clauses that further define the subject (e.g., "who/which is weak", "whose resource for sustaining life is grain")
\item \code{Sx->A} for a symbol that transforms the preceding noun phrase into an attribute (e.g., "whose resource for sustaining life is \code{S}"). Other transformations are denoted accordingly.
\item \code{N} for numbers,
\item \code{D} for everything else.
}
}

\subsection{Processing Steps}{
\enumerate{
\item Reads text from .docx files or plain text files
\item Filters lines starting with \code{|}
\item Parses each line into sign name, type, and meaning components
\item Cleans meaning field by removing content after \code{;} or \code{|} delimiters
\item Issues a warning for entries with missing type annotations
\item Excludes lines containing the unknown-sign placeholder \code{X}
\item Replaces standalone numbers in sign names with \code{N} (suffix digits like the \code{2} in \code{jal2} are not affected)
\item Normalizes transliterated text by removing separators and looking up the sign names from the \code{mapping}
\item Excludes empty sign names from the result
}
}

}
\value{
A data frame with the following columns:
\describe{
\item{sign_name}{The normalized sign name with components separated by hyphens (e.g., \code{"A"}, \code{"AN"}, \code{"X-NA"})}
\item{type}{Grammatical type (e.g., \code{"S"}, \code{"V"}, \code{"A"}, \code{"Sx->A"})}
\item{meaning}{The translated meaning of the sign}
}
}
\note{
If any translations have missing type annotations, the function prints a
warning message listing the affected entries.
}
\seealso{
\code{\link{convert_to_dictionary}} for converting the result into a dictionary,
\code{\link{make_dictionary}} for creating a complete dictionary with
cuneiform representations and readings in a single step.
}
\examples{

# Read translations from a single text document
filename     <- system.file("extdata", "text_with_translations.txt", package = "sumer")
translations <- read_translated_text(filename)

# View the structure
head(translations)

# Filter by grammatical type
nouns <- translations[translations$type == "S", ]
nouns

#Make some custom unifications (here: removing the word "the")
translations$meaning <- gsub("\\\\bthe\\\\b", "", translations$meaning, ignore.case = TRUE)
translations$meaning <- trimws(gsub("\\\\s+", " ", translations$meaning))

# View the structure
head(translations)

#Convert the result into a dictionary
dictionary   <- convert_to_dictionary(translations)

# View the structure
head(dictionary)

}

\name{look_up}
\alias{look_up}
\alias{print.look_up}
\title{Look Up Sumerian Signs or Search for Translations}
\description{
Searches a Sumerian dictionary either by sign name (forward lookup) or by
translation text (reverse lookup).

The function \code{look_up} computes the search results and returns an object of class \code{"look_up"}. The \code{print} method displays formatted results with cuneiform representations, grammatical types, and translation counts.
}
\usage{
look_up(x, dic, lang = "sumer", width = 70, mapping = NULL)

\method{print}{look_up}(x, ...)
}
\arguments{
\item{x}{For \code{look_up}: A character string specifying the search term. Can be either:
\itemize{
\item A Sumerian sign name (e.g., \code{"AN"}, \code{"AN.EN.ZU"})
\item A cuneiform character string
\item A word or phrase to search in translations (e.g., \code{"Gilgamesh"}, \code{"heaven"})
}

For \code{print.look_up}: An object of class \code{"look_up"} as returned by \code{look_up}.}

\item{dic}{A dictionary data frame, typically created by
\code{\link{make_dictionary}} or loaded with \code{\link{read_dictionary}}.
Must contain columns \code{sign_name}, \code{row_type}, \code{count},
\code{type}, and \code{meaning}.}
\item{lang}{Character string specifying whether \code{x} is a Sumerian expression (\code{"sumer"}) or an English expression (\code{"en"}).}
\item{width}{Integer specifying the text width for line wrapping. Default is 70.}
\item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
\item{...}{Additional arguments passed to the print method (currently unused).}
}
\value{
\code{look_up} returns an object of class \code{"look_up"}, which is a list containing:
\item{search}{The original search term.}
\item{lang}{The language setting used for the search.}
\item{width}{The text width for formatting.}
\item{cuneiform}{The cuneiform representation (only for Sumerian searches).}
\item{sign_name}{The canonical sign name (only for Sumerian searches).}
\item{translations}{A data frame with translations for the exact sign combination (only for Sumerian searches).}
\item{substrings}{A named list of data frames with translations for individual signs and substrings (only for Sumerian searches).}
\item{matches}{A data frame with matching entries (only for non-Sumerian searches).}

\code{print.look_up} prints formatted dictionary entries to the console and returns \code{x} invisibly.
}
\details{
\subsection{Search Modes}{
The function operates in two modes depending on the input:

\strong{Forward Lookup} (Sumerian input detected):
\enumerate{
\item Converts the sign name to cuneiform
\item Retrieves all translations for the exact sign combination
\item Retrieves translations for all individual signs and substrings
}

\strong{Reverse Lookup} (non-Sumerian input):
\enumerate{
\item Searches for the term in all translation meanings
\item Retrieves matching entries with sign names and cuneiform
}
}

\subsection{Output Format}{
The print method displays results with:
\itemize{
\item Sign names with cuneiform representations
\item Occurrence counts in brackets (e.g., \code{[29]})
\item Grammatical type abbreviations (e.g., \code{S}, \code{V})
\item Translation meanings with automatic line wrapping
\item Search term highlighting in blue for reverse lookups (only for ANSI-compatible terminals)
}
}


}
\seealso{
\code{\link{read_dictionary}} for loading dictionaries,
\code{\link{make_dictionary}} for creating dictionaries,
\code{\link{as.cuneiform}} for cuneiform conversion.
}
\examples{
# Load dictionary
dic <- read_dictionary()

# Forward lookup: search by phonetic spelling
look_up("d-suen", dic)

# Forward lookup: search by Sumerian sign name
look_up("AN", dic)
look_up("AN.EN.ZU", dic)

# Forward lookup: search by cuneiform character string
AN.NA <- paste0(intToUtf8(0x1202D), intToUtf8(0x1223E))
AN.NA
look_up(AN.NA, dic)

# Reverse lookup: search in translations
look_up("Gilgamesh", dic, "en")

# Adjust output width for narrow terminals
look_up("water", dic, "en", width = 50)

# Store results for later use
result <- look_up("lugal", dic)
result$cuneiform
result$translations

# Print stored results
print(result)
}
\keyword{utilities}
\keyword{database}

\name{extract_skeleton_entries}
\alias{extract_skeleton_entries}
\title{Extract Hierarchical Skeleton Entries from Bracketed Text}
\description{
Recursively extracts the contents of nested round brackets from a normalized Sumerian text string and returns them as a data frame with position, length, nesting depth, and expression for each entry.

This is an internal helper function used by \code{\link{skeleton}}.
}
\usage{
extract_skeleton_entries(x)
}
\arguments{
  \item{x}{A character string containing Sumerian text with round brackets, as returned by \code{\link{mark_skeleton_entries}}.}
}
\details{
The first row of the result always represents the entire input expression at depth 0 (the root entry). The function then extracts the contents of all outermost (top-level) bracket pairs using an internal helper function. For each extracted group, a row is added to the result data frame at depth 1. If a group itself contains further nested brackets, the function recurses into it to extract deeper levels.

The \code{depth} value of each entry reflects the nesting level: the root entry has depth 0, entries from the outermost brackets have depth 1, entries nested one level deeper have depth 2, and so on.

The \code{start} column records the position (in tokens) of the first token in each group, relative to the full input. The \code{n_tokens} column gives the number of tokens in the group as determined by \code{\link{split_sumerian}}.
}
\value{
A data frame with one row per extracted entry and the following columns:
  \item{start}{Integer. The token position of the first token in the group (1-based).}
  \item{n_tokens}{Integer. The number of Sumerian tokens (signs) in the group.}
  \item{depth}{Integer. The nesting depth of the entry (0 for the root entry representing the full expression, 1 for top-level groups, 2 for groups nested one level deeper, etc.).}
  \item{expr}{Character. The text content of the bracket group (without the surrounding brackets). For the root entry (row 1), this is the full input string.}

The result always has at least one row (the root entry).
}
\seealso{
\code{\link{skeleton}} which calls this function,
\code{\link{mark_skeleton_entries}} for the preceding normalization step,
\code{\link{split_sumerian}} for determining the number of tokens
}
\examples{
# First normalize the input with mark_skeleton_entries
x <- "<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra. gaba jal2. an ki a"
normalized <- sumer:::mark_skeleton_entries(x)
normalized

# Then extract the hierarchical structure
sumer:::extract_skeleton_entries(normalized)
}
\keyword{internal}

\name{compose_skeleton_entry}
\alias{compose_skeleton_entry}
\title{Compose a Skeleton Entry from its Children}
\description{
Composes the \code{type} and \code{translation} fields of a parent skeleton entry from the corresponding fields of its direct children, following the rules of the Sumerian type system. This function is called by \code{\link{translate}} when the user clicks the brown compose button next to a skeleton entry.
}
\usage{
compose_skeleton_entry(df)
}
\arguments{
  \item{df}{A data frame with \eqn{n + 1} rows, where the first row represents the parent entry and the remaining \eqn{n} rows represent its direct children in the skeleton hierarchy.

  The data frame has the following columns:
  \describe{
    \item{expr}{Character. The expression of the entry.}
    \item{n_tokens}{Integer. Number of tokens covered by the entry.}
    \item{start}{Integer. Position (1-based) of the first token.}
    \item{depth}{Integer. Nesting depth in the skeleton hierarchy.}
    \item{type}{Character. The grammatical type annotation: \code{"S"} (substantive), \code{"V"} (verb), \code{"A"} (adjective/attribute), or an operator type like \code{"xS->S"}.}
    \item{translation}{Character. The translation. Operator translations contain type-letter placeholders (e.g. \code{"S"} in \code{"supplier of energy from S"}).}
  }
  }
}
\details{
The algorithm works in three steps:
\enumerate{
  \item \strong{Structure string:} Build a structure string by replacing each child's expression in the parent expression with a numbered tag (\code{#1}, \code{#2}, etc.). Token-group wrappers (\code{<>} and single-tag \code{\{\}}) are removed.
  \item \strong{Bracket insertion:} \code{\link{add_brackets}} groups operators with their arguments and applies compositional rules (S+A, S+S, S+V, SEN+SEN).
  \item \strong{Translation:} \code{\link{apply_translation_rules}} recursively evaluates the bracketed string to produce the final translation.
}
}
\value{
Either the input data frame \code{df} with the \code{type} and \code{translation} fields of the first row filled in, or a character string containing an error message.
}
\examples{
# Minimal example: S + Sx->V
df <- data.frame(
  expr = c("#1 #2", "#1", "#2"),
  n_tokens = c(2L, 1L, 1L),
  start = c(1L, 1L, 2L),
  depth = c(0L, 1L, 1L),
  type = c("", "S", "Sx->V"),
  translation = c("", "water", "to utilize S"),
  stringsAsFactors = FALSE
)
result <- sumer:::compose_skeleton_entry(df)
result
}

\seealso{
\code{\link{add_brackets}}, \code{\link{apply_translation_rules}} for the individual pipeline steps,
\code{\link{translate}} which calls this function
}
\keyword{internal}

\name{plot_sign_grammar}
\alias{plot_sign_grammar}
\title{Stacked Bar Chart of Grammatical Type Frequencies}
\description{
Creates a stacked bar chart from the output of \code{\link{sign_grammar}} or
\code{\link{grammar_probs}}. Each bar represents one sign position in the
sentence. The colours indicate the relative frequency or posterior
probability of each individual grammatical type.
}
\usage{
plot_sign_grammar(sg,
                  output_file = NULL,
                  width       = 10,
                  height      = 5,
                  sign_names  = FALSE,
                  font_family = NULL,
                  mapping     = NULL)
}
\arguments{
  \item{sg}{A data frame as returned by \code{\link{sign_grammar}} (with
    column \code{n}) or \code{\link{grammar_probs}} (with column
    \code{prob}).}

  \item{output_file}{Character. File path for saving the plot (PNG or JPG).
    If \code{NULL} (default), the plot is displayed on the current device.}

  \item{width}{Numeric. Plot width in inches. Default: 10.}

  \item{height}{Numeric. Plot height in inches. Default: 5.}

  \item{sign_names}{Logical. Whether sign names or cuneiform characters should be used as labels of the x-axis. Default: FALSE.}

  \item{font_family}{Character. Font family for cuneiform x-axis labels.
    If \code{NULL} (default), \code{"Segoe UI Historic"} is used.}

  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded.}
}
\details{
When the input comes from \code{sign_grammar()} (column \code{n}),
absolute frequencies are converted to percentages so that bars sum to
100\%. When the input comes from \code{grammar_probs()} (column
\code{prob}), posterior probabilities are used directly.

Colours are assigned per grammatical type, grouped by class:
\itemize{
  \item Red shades: Verbs (\code{V}) and operators returning verbs
  \item Blue shades: Operators returning attributes \code{A}
  \item Orange: Adjectives and other signs with grammatical type (\code{Sx->S})
  \item Green: Nouns

  \item Grey/other shades: All other types
}
}
\value{
Invisibly returns the \pkg{ggplot2} plot object.
}
\examples{
dic   <- read_dictionary()
sg    <- sign_grammar("a-ma-ru ba-ur3 ra", dic)

# Plot raw frequencies
file <- file.path(tempdir(), "test.png")
plot_sign_grammar(sg, file)

# Plot probabilities
prior <- prior_probs(dic, sentence_prob = 0.25)
gp    <- grammar_probs(sg, prior, dic, alpha0 = 1)
file  <- file.path(tempdir(), "test2.png")
plot_sign_grammar(gp, file)
}
\seealso{
\code{\link{sign_grammar}} for generating raw frequency data,
\code{\link{grammar_probs}} for Bayesian posterior probabilities,
\code{\link{prior_probs}} for computing the prior.
}
\keyword{hplot}

\name{read_dictionary}
\alias{read_dictionary}
\title{Read a Sumerian Dictionary from File}
\description{
Reads a Sumerian dictionary from a semicolon-separated text file, optionally
displaying the metadata header with author, version, and update information.
}
\usage{
read_dictionary(file = NULL, verbose = TRUE)
}
\arguments{
\item{file}{A character string specifying the path to the dictionary file.
If \code{NULL} (default), the package's built-in dictionary
\code{sumer-dictionary.txt} is loaded.}

\item{verbose}{Logical. If \code{TRUE} (default), the metadata header
(author, year, version, URL) is printed to the console.}
}
\value{
A data frame with the following columns:
\describe{
\item{sign_name}{The Sumerian sign name (e.g., \code{"A"}, \code{"AN"}, \code{"ME"})}
\item{row_type}{Type of entry: \code{"cunei."} (cuneiform character), \code{"reading"} (phonetic readings), or \code{"trans."} (translation)}
\item{count}{Number of occurrences for translations; \code{NA} for cuneiform and reading entries}
\item{type}{Grammatical type (e.g., \code{"S"}, \code{"V"}) for translations; empty string for other row types}
\item{meaning}{The cuneiform character(s), phonetic reading(s), or translated meaning depending on \code{row_type}}
}
}
\details{
\subsection{File Format}{
The function expects a semicolon-separated file with a metadata header.
Lines starting with \code{#} are treated as comments. The expected format is:

\preformatted{
###---------------------------------------------------------------
###                Sumerian Dictionary
###
### Author:  Robin Wellmann
### Year:    2026
### Version: 0.5
### Watch for Updates:
###   https://founder-hypothesis.com/en/sumerian-mythology/downloads/
###---------------------------------------------------------------
sign_name;row_type;count;type;meaning
A;cunei.;;;<here would be the cuneiform sign for A>
A;reading;;;{a, dur5, duru5}
A;trans.;3;S;water
}
}

\subsection{Encoding}{
The file is read with UTF-8 encoding to properly handle cuneiform characters.
}
}
\seealso{
\code{\link{save_dictionary}} for saving dictionaries to file,
\code{\link{make_dictionary}} and \code{\link{convert_to_dictionary}} for
creating dictionaries.
}
\examples{
# Load the built-in dictionary
dic <- read_dictionary()

# Load a custom dictionary
filename <- system.file("extdata", "sumer-dictionary.txt", package = "sumer")
dic <- read_dictionary(filename)

# Look up an entry
look_up("d-suen", dic)

}

\name{convert_to_dictionary}
\alias{convert_to_dictionary}
\title{Convert Translation Data to a Sumerian Dictionary}
\description{
Converts a data frame of Sumerian translations into a structured dictionary
format, adding cuneiform representations and phonetic readings for each sign.
}
\usage{
convert_to_dictionary(df, mapping = NULL)
}
\arguments{
\item{df}{A data frame with columns \code{sign_name}, \code{type}, and
\code{meaning}, typically produced by \code{\link{read_translated_text}}.}

\item{mapping}{A data frame containing sign-to-reading mappings with columns
\code{name}, \code{cuneiform} and \code{syllables}. If \code{NULL} (default), the package's
built-in mapping file \code{etcsl_mapping.txt} is used.}
}
\value{
A data frame with the following columns:
\describe{
\item{sign_name}{The normalized Sumerian text (e.g., \code{"A"}, \code{"AN"}, \code{"A2.TAB"})}
\item{row_type}{Type of entry: \code{"cunei."} (cuneiform character), \code{"reading"} (phonetic readings), or \code{"trans."} (translation)}
\item{count}{Number of occurrences for translations; \code{NA} for cuneiform and reading entries}
\item{type}{Grammatical type (e.g., \code{"S"}, \code{"V"}, \code{"A"}) for translations; empty string for other row types}
\item{meaning}{The cuneiform character(s), phonetic reading(s), or translated meaning depending on \code{row_type}}
}

The data frame is sorted by \code{sign_name}, \code{row_type}, and
descending \code{count}.
}
\details{
\subsection{Processing Steps}{
\enumerate{
\item Aggregates translations and counts occurrences of each unique combination in \code{df}
\item Looks up phonetic readings and cuneiform signs for each sign component
\item Combines cuneiform, reading, and translation rows into a single data frame
\item Sorts the result by sign name and row type
}
}

\subsection{Reading Format}{
Phonetic readings are formatted as follows:
\itemize{
\item Multiple possible readings are enclosed in braces: \code{\{a, dur5, duru5\}}
\item For compound signs, readings of individual components are joined with hyphens
\item If a sign has more than three possible readings in a compound, only the first three are shown followed by \code{...}
\item Unknown readings are marked with \code{?}
}
}
}
\seealso{
\code{\link{read_translated_text}} for reading translation files,
\code{\link{make_dictionary}} for creating a complete dictionary with
cuneiform representations and readings in a single step.
}
\examples{
# Read translations from a single text document
filename     <- system.file("extdata", "text_with_translations.txt", package = "sumer")
translations <- read_translated_text(filename)

# View the structure
head(translations)

#Make some custom unifications (here: removing the word "the")
translations$meaning <- gsub("\\\\bthe\\\\b", "", translations$meaning, ignore.case = TRUE)
translations$meaning <- trimws(gsub("\\\\s+", " ", translations$meaning))

# View the structure
head(translations)

#Convert the result into a dictionary
dictionary   <- convert_to_dictionary(translations)

# View the structure
head(dictionary)

# View entries for a specific sign
dictionary[dictionary$sign_name == "EN", ]

# With custom mapping
path  <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
mapping <- read.csv2(path, sep=";", na.strings="")
translations <- read_translated_text(filename, mapping = mapping)
dictionary <- convert_to_dictionary(translations, mapping = mapping)
head(dictionary)
}

\name{prior_probs}
\alias{prior_probs}
\title{Prior Probabilities of Grammatical Types}
\description{
Computes prior probabilities for each grammatical type (e.g., \code{S},
\code{V}, \code{Sx->S}, \code{xS->A}, etc.) from a dictionary. The priors
can be corrected for verb underrepresentation in the dictionary data.
}
\usage{
prior_probs(dic, sentence_prob = 1.0)
}
\arguments{
  \item{dic}{A dictionary data frame as returned by
    \code{\link{read_dictionary}}.}

  \item{sentence_prob}{Numeric in (0, 1]. The estimated proportion of
    complete sentences (as opposed to noun phrases) in the training data
    from which the dictionary was created. Verbs appear in complete
    sentences, so a value less than 1 upweights verb-like types.
    Default: 1.0.}
}
\details{
The function proceeds in three steps:

\enumerate{
  \item For each single-sign dictionary entry with at least one count,
    the counts per grammatical type are normalised to sum to 1.
  \item The prior probability of each type is the mean of these
    normalised frequencies across all signs.
  \item A correction is applied: counts of verb-like types (\code{V} and all
    operators with return type \code{V}, such as \code{Vx->V} or
    \code{xV->V}) are multiplied by \code{1/sentence_prob}, then all
    probabilities are renormalised. This compensates for the fact that
    verbs are underrepresented when most dictionary entries are obtained from noun
    phrases rather than complete sentences.
}

When \code{sentence_prob = 1}, no correction is applied.
}
\value{
A named numeric vector with one element per grammatical type found in
the dictionary, summing to 1. The names are the type strings as they
appear in the dictionary (e.g., \code{"S"}, \code{"V"}, \code{"Sx->S"}).
The \code{sentence_prob} parameter is stored as an attribute.
}
\examples{
dic   <- read_dictionary()

# Default usage
prior_probs(dic)

# Applying correction (only 25\% sentences in training data)
prior_probs(dic, sentence_prob = 0.25)
}
\seealso{
\code{\link{sign_grammar}} for per-sign grammatical type frequencies.
}
\keyword{utilities}

\name{mark_skeleton_entries}
\alias{mark_skeleton_entries}
\title{Normalize Brackets for Skeleton Generation}
\description{
Transforms a transliterated Sumerian text string into a normalized form that contains only round brackets. This prepares the input for hierarchical extraction by \code{\link{extract_skeleton_entries}}.

This is an internal helper function used by \code{\link{skeleton}}.
}
\usage{
mark_skeleton_entries(x)
}
\arguments{
  \item{x}{A character string of length 1 containing transliterated Sumerian text. The string may contain angle brackets (\code{< >}), round brackets (\code{( )}), and curly braces (\code{\{ \}}) to annotate token groups (see Details).}
}
\details{
The function performs the following transformations:

\enumerate{
  \item Tokenizes the input using an internal helper function. Tokens enclosed in angle brackets are merged into a single token.
  \item Removes angle brackets from the separators, replacing them with spaces. Curly braces are preserved.
  \item Wraps every token that is not already enclosed in round brackets with round brackets.
}

The result is a string in which every token is enclosed in round brackets. Existing round brackets from the input are preserved, so the nesting structure reflects the grouping specified in the original input.

For example, the input

\code{"<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra"}

is transformed into a string where \code{d-nu-dim2-mud} appears as a single bracketed token, \code{e2} and \code{kur} are individually bracketed inside the existing round brackets around \code{e2\{kur\}}, and all other tokens (\code{ki}, \code{a}, \code{jal2}, \code{ra}) are each wrapped in their own round brackets.
}
\value{
A character string of length 1 in which all tokens are enclosed in round brackets. Angle brackets are removed; curly braces from the input are preserved.
}
\seealso{
\code{\link{skeleton}} which calls this function,
\code{\link{extract_skeleton_entries}} for the subsequent extraction step,
\code{\link{split_sumerian}} for the underlying sign-splitting logic
}
\examples{
# Input with all three bracket types
x <- "<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra. gaba jal2. an ki a"
sumer:::mark_skeleton_entries(x)

# Input without any brackets: each token gets wrapped in round brackets
sumer:::mark_skeleton_entries("LUGAL.E")

# Angle brackets merge tokens into a single unit
sumer:::mark_skeleton_entries("<an-ki> lugal")
}
\keyword{internal}

\name{save_dictionary}
\alias{save_dictionary}
\title{Save a Sumerian Dictionary to File}
\description{
Saves a Sumerian dictionary data frame to a semicolon-separated text file
with a metadata header containing author, year, version, and URL information.
}
\usage{
save_dictionary(dic, file, author = "", year = "", version = "", url = "")
}
\arguments{
\item{dic}{A dictionary data frame, typically created by
\code{\link{make_dictionary}} or \code{\link{convert_to_dictionary}}.
Must contain columns \code{sign_name}, \code{row_type}, \code{count},
\code{type}, and \code{meaning}.}

\item{file}{A character string specifying the output file path.}

\item{author}{A character string with the author name(s) for the metadata header.}

\item{year}{A character string with the year of creation for the metadata header.}

\item{version}{A character string with the version number for the metadata header.}

\item{url}{A character string with a URL where updates can be found.}
}
\value{
No return value. The function is called for its side effect of writing the
dictionary to a file.
}
\details{
\subsection{Output Format}{
The output file consists of two parts:

\enumerate{
\item A metadata header with lines starting with \code{###}, containing
author, year, version, and URL information
\item The dictionary data in semicolon-separated format with columns:
\code{sign_name}, \code{row_type}, \code{count}, \code{type}, \code{meaning}
}

Example output:
\preformatted{
###---------------------------------------------------------------
###                Sumerian Dictionary
###
### Author:  Robin Wellmann
### Year:    2026
### Version: 1.0
### Watch for Updates: https://founder-hypothesis.com/sumer/
###---------------------------------------------------------------
sign_name;row_type;count;type;meaning
A;cunei.;;;<cuneiform sign for A>
A;reading;;;{a, dur5, duru5}
A;trans.;3;S;water
}
}


}

\seealso{
\code{\link{make_dictionary}} and \code{\link{convert_to_dictionary}} for
creating dictionaries, \code{\link{read_dictionary}} for reading saved
dictionaries.
}
\examples{
# Create and save a dictionary

filename  <- system.file("extdata", "text_with_translations.txt", package = "sumer")
dictionary <- make_dictionary(filename)

save_dictionary(
  dic     = dictionary,
  file    = file.path(tempdir(), "sumerian_dictionary.txt"),
  author  = "John Doe",
  year    = "2026",
  version = "1.0",
  url     = "https://example.com/dictionary"
)
}

\name{apply_translation_rules}
\alias{apply_translation_rules}
\title{Apply Translation Rules to a Bracketed Structure String}
\description{
Translates a bracketed structure string into English by evaluating sumerian operators (substituting arguments into translation templates) and composing adjacent elements according to grammatical rules. The input is a structure string as produced by \code{\link{add_brackets}}, together with  vectors of types and translations for each tag.
}
\usage{
apply_translation_rules(s, type, translation)
}
\arguments{
  \item{s}{Character string showing the order of evaluation, as produced by \code{\link{add_brackets}}.}
  \item{type}{Character vector of grammatical types. \code{type[k]} is the type string for tag \code{#k} (e.g. \code{"S"}, \code{"V"}, \code{"Sx->V"}).}
  \item{translation}{Character vector of translations. \code{translation[k]} is the translation for tag \code{#k}. Operator translations contain placeholders (e.g. \code{"to deliver S"}).}
}
\details{
Nested \code{\{...\}} groups are evaluated from the inside out. Within each group, an operator (if present) binds its arguments and produces a typed result. Groups without an operator are composed according to the rules described in \code{\link{eval_operator}}.
}
\value{
A character vector of length 2: \code{c(result_type, result_translation)}.

On error (e.g. incompatible types), a character vector of length 1 containing the error message.
}
\examples{
x <- "mec3-ki-aj2-ga-ce-er"
x <- as.cuneiform(x)
x

meaning <- rbind( c("S",      "a man who relies on his own strength"),
                  c("S",      "place {earth}"),
                  c("Sx->A",  ", whose allocated resource is S"),
                  c("xS->A",   ", whose sustenance is S"),
                  c("S",      "grain"),
                  c("Sx->S",  "lamented S"))

df <- data.frame(
    type  = meaning[,1],
    translation = meaning[,2],
    expr  = split_sumerian(x)$signs)

s <- x
for(i in 1:nrow(df)){
 s <- sub(df$expr[i], paste0("#", i), s)
}
s

s_bracketed <- sumer:::add_brackets(s, df$type)
s_bracketed

apply_translation_rules(s_bracketed$string, df$type, df$translation)
}

\seealso{
\code{\link{eval_operator}} for the evaluation rules,
\code{\link{add_brackets}} for the previous pipeline step,
\code{\link{compose_skeleton_entry}} which calls this function
}

\name{init_substr_info}
\alias{init_substr_info}
\title{Initialize a Data Frame of All Substrings}
\description{
Creates a data frame containing all contiguous substrings of a token vector, including the full token sequence itself. Each row represents one substring, with its starting position, length in tokens, the concatenated expression, and empty columns for type and translation.

The rows are ordered by \code{n_tokens} descending and \code{start} ascending, so that the row number can be computed from \code{start} and \code{n_tokens} using \code{\link{substr_position}}.

This is an internal helper function.
}
\usage{
init_substr_info(token)
}
\arguments{
  \item{token}{A character vector of Sumerian tokens (e.g. cuneiform signs).}
}
\details{
For a token vector of length \eqn{N}, the function generates all \eqn{N(N+1)/2} contiguous substrings. The substrings are ordered by \code{n_tokens} descending (longest first) and within each group by \code{start} ascending. This ordering ensures that the row index of any substring can be computed with the formula

\deqn{\mathrm{row} = \frac{(N - k)(N - k + 1)}{2} + s}{row = (N - k)(N - k + 1) / 2 + s}

where \eqn{k} is the number of tokens (\code{n_tokens}) and \eqn{s} is the starting position (\code{start}).

The \code{expr} column contains the tokens concatenated without separators. The \code{type} and \code{translation} columns are initialized as empty strings, intended to be filled in later.
}
\value{
A data frame with \eqn{N(N+1)/2} rows and the following columns:
  \item{start}{Integer. The position of the first token in the substring (1-based).}
  \item{n_tokens}{Integer. The number of tokens in the substring.}
  \item{expr}{Character. The concatenated token sequence (without separators).}
  \item{type}{Character. Initialized as empty string \code{""}.}
  \item{translation}{Character. Initialized as empty string \code{""}.}
}
\seealso{
\code{\link{substr_position}} for computing the row index from \code{start} and \code{n_tokens},
\code{\link{skeleton}} for creating translation templates,
\code{\link{make_dictionary}} for creating dictionaries from filled-in templates
}
\examples{
x<-"<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra. gaba jal2. an ki a"

token <- split_sumerian(as.cuneiform(x))$signs

df <- sumer:::init_substr_info(token)
df

# Verify that substr_position recovers the row indices
N <- length(token)
all(seq_len(nrow(df)) == sumer:::substr_position(df$start, df$n_tokens, N))
}
\keyword{internal}

\name{grammatical_structure}
\alias{grammatical_structure}
\alias{print.grammatical_structure}
\title{Grammatical Structure of a Sumerian Expression}
\description{
Determines and visualizes the grammatical structure of a Sumerian expression. The function groups sub-expressions according to operator binding and composition rules and returns a bracketed string in which each bracket type indicates the grammatical role of the group:

\itemize{
  \item \code{()} -- substantive (S)
  \item \code{<>} -- verb (V)
  \item \code{[]} -- attribute (A)
  \item \code{\{\}} -- sentence (SEN)
}

The result has class \code{"grammatical_structure"} and comes with a \code{print} method that displays the bracket tree with color-coded groups in the console (requires ANSI color support).
}
\usage{
grammatical_structure(s, type, expr = NULL)

\method{print}{grammatical_structure}(x, \dots)
}
\arguments{
  \item{s}{Character string. A Sumerian expression in cuneiform characters.}
  \item{type}{Character vector of grammatical types, one per sub-expression. Each entry is either a base type (\code{"S"}, \code{"V"}, \code{"A"}) or an operator type (e.g.  \code{"Sx->V"}, \code{"xS->A"}).}
  \item{expr}{Character vector of sub-expressions (e.g. the individual signs or sign groups that make up \code{s}). The function matches each \code{expr[k]} in \code{s}, determines the grouping, and returns the result with the original expressions in place.}
  \item{x}{An object of class \code{"grammatical_structure"}.}
  \item{\dots}{Further arguments (currently unused).}
}
\details{
The grouping is performed in two stages. First, \code{\link{add_brackets}} inserts bracket groups based on operator binding strength and pairwise composition rules. Then each group is assigned a bracket type that reflects its grammatical role, as determined by the operator it contains or by the types of its elements.

The \code{print} method displays the resulting string with ANSI colors in the console. Each bracket type and its direct content (nesting level 0) are shown in a distinct color: green for \code{()}, blue for \code{[]}, red for \code{<>}, and yellowish-brown for \code{\{\}}. Bracket pairs that contain only nested sub-groups (no bare symbols at nesting level 0) are shown in light gray.
}
\value{
A character string of class \code{"grammatical_structure"} with typed brackets showing the grammatical grouping. On error, a plain character string containing the error message.
}
\examples{
# Example 1
x    <- "mec3-ki-aj2-ga-ce-er ce du"
expr <- split_sumerian(x)$signs
type <- c("S", "S", "Sx->A", "xS->A",  "S",  "Sx->S", "S", "Sx->V")

grammatical_structure(x, type, expr)

grammatical_structure(as.cuneiform(x), type, as.cuneiform(expr))

# Example 2: An example with a proper name and verb prefixes
x    <- "an-en-ki-en-gan-ig-la"
expr <- c("an-en-ki", "en", "gan", "ig", "la")
type <- c("S", "S", "xV->V", "xV->V", "Vt")
grammatical_structure(as.cuneiform(x), type, as.cuneiform(expr))

}
\seealso{
\code{\link{apply_translation_rules}} for translating (instead of visualizing) the structure,
\code{\link{add_brackets}} for the underlying grouping algorithm,
\code{\link{split_sumerian}} for obtaining the sub-expressions from a Sumerian string
}

\name{translate_line}
\alias{translate_line}
\alias{translation_context}
\title{Line-by-Line Translation of a Sumerian Text}
\description{
\code{translation_context} bundles the settings for a translation project: the project directory (with a \file{lines/} subfolder for saving line files), the dictionaries, the source text, and parameters for the interactive translator.

\code{translate_line} opens the interactive translation tool (\code{\link{translate}}) for a single line. If a saved translation exists for this line, it is loaded so that previous work can be continued. When the user clicks \dQuote{Done}, the result is saved back to the line file.
}
\usage{
translation_context(project_dir, dic = NULL, text = NULL,
                    mapping = NULL, min_freq = c(6, 4, 2),
                    sentence_prob = 1.0)

translate_line(n, context)
}
\arguments{
  \item{project_dir}{Character string. Path to the project directory. Translated lines are saved in the subdirectory \file{lines/} as \file{Line_1.txt}, \file{Line_2.txt}, etc. If \code{text} or \code{dic} are given as filenames without a path, they are assumed to reside in \code{project_dir}.}
  \item{dic}{Dictionaries to use for translation lookup. A character vector of file paths (or filenames relative to \code{project_dir}), a single data frame, or a list of data frames. The first dictionary serves as the primary source. Additional dictionaries serve as fallback references. If \code{NULL}, the built-in dictionary is used.}
  \item{text}{The full text being translated. A file path (or filename relative to \code{project_dir}) or a character vector (one line per element). Needed for n-gram analysis and context display in \code{\link{translate}}.}
  \item{mapping}{A data frame with the sign mapping table, or a file path to a semicolon-separated mapping file. If \code{NULL} (the default), the package's internal mapping is used.}
  \item{min_freq}{Minimum frequency thresholds passed to \code{\link{translate}}. Default is \code{c(6, 4, 2)}.}
  \item{sentence_prob}{Probability that a randomly chosen sign is part of a sentence with a verb, passed to \code{\link{translate}}. Default is \code{1.0}.}
  \item{n}{An integer line number. If a file \file{Line_<n>.txt} exists in the \file{lines/} subdirectory of \code{project_dir}, the saved translation is loaded for revision. Otherwise, the most frequent translation for each substring is looked up in the dictionaries.}
  \item{context}{A context object as created by \code{translation_context}.}
}
\details{
\code{translate_line} performs the following steps:

\enumerate{
  \item If \file{Line_<n>.txt} exists in the \file{lines/} subdirectory, the cuneiform text and previous translations are loaded via \code{\link{fill_substr_info}}.
  \item A project dictionary is built from saved line files using \code{\link{make_dictionary}}. Line \code{n} itself is excluded to avoid confirmation bias.
  \item The interactive translator (\code{\link{translate}}) is opened with the primary dictionary, the project dictionary, and any additional dictionaries as fallback references.
  \item If the user clicks \dQuote{Done}, the result is saved to \file{Line_<n>.txt}. If the window is closed in another way, the result is not saved.
}
}
\value{
\code{translation_context} returns a list of class \code{"translation_context"}.

\code{translate_line} is called for its side effect (opening the interactive translator and saving the result). Returns \code{NULL} invisibly.
}
\seealso{
\code{\link{translate}} for the underlying interactive translation tool,
\code{\link{fill_substr_info}} for reading back a saved translation,
\code{\link{make_dictionary}} for creating a dictionary from translated lines
}
\examples{
# Note: The folder containing the built-in project
#       is copied to a temporary directory.
#       This prevents you from altering the package files.

path  <- system.file("extdata", package = "sumer")

file.copy(
  from = file.path(path, "project"),
  to   = tempdir(),
  recursive = TRUE
)

ctx <- translation_context(
  project_dir = file.path(tempdir(), "project"),
  text        = "enki_and_the_world_order.txt",
  dic         = file.path(path, "sumer-dictionary.txt"),
  sentence_prob = 0.25
)
ctx

\dontrun{
# Translate line 29 (opens interactive translator)
translate_line(29, ctx)

# Confirm that your changes are still there:
translate_line(29, ctx)

# Continue with the next line:
translate_line(30, ctx)
}
}
\keyword{utilities}

\name{eval_operator}
\alias{eval_operator}
\title{Evaluate an Operator or Composition}
\description{
Evaluates a single operator by substituting resolved arguments into its translation string, or composes two elements without an operator using compositional rules.

This function contains the complete translation logic: extraction of specific meanings from \code{\{specific\}} annotations, article insertion (\code{"the "}), verb-prefix stripping (\code{"to "}), placeholder substitution, and composition rules.
}
\usage{
eval_operator(operator, rule_translation, args = list(), seps = "", trailing_sep = "")
}
\arguments{
  \item{operator}{Type string of the operator (e.g. \code{"Sx->V"}, \code{"SSx->V"}, \code{"xS->A"}), or \code{NULL} for composition (no operator present). For base types without arguments (e.g. \code{"S"}, \code{"V"}), pass the type string here.}
  \item{rule_translation}{Translation string with placeholders (e.g. \code{"to utilize S"}, \code{"S1 and therefore S2"}), or \code{NULL} for composition.}
  \item{args}{List of \code{c(name, translation)} character vectors:
    \describe{
      \item{name}{Placeholder name encoding the grammatical type (e.g. \code{"S"}, \code{"S1"}, \code{"V"}, \code{"SEN2"}). The type is reconstructed internally by stripping the trailing index digits.}
      \item{translation}{Resolved translation of the argument.}
    }
    Default: \code{list()} (no arguments, for base types).
    The helper function \code{build_args()} can be used to construct this list from a set of elements.
  }
  \item{seps}{Character vector of separators between elements, as extracted by \code{\link{apply_translation_rules}}. Hyphens in separators are converted to spaces for composition. Default: \code{""}.}
  \item{trailing_sep}{Text after the last element (e.g. trailing punctuation). Hyphens are converted to spaces for composition. Default: \code{""}.}
}
\details{
The function operates in two modes:

\strong{Operator mode} (\code{operator} is not \code{NULL}):
\enumerate{
  \item Determine the result type from the operator string via \code{parse_type()}.
  \item For each argument, apply transformations:
    \itemize{
      \item Strip \code{"to "} from verb-type (\code{"V"}) arguments.
      \item Add \code{"the "} before substantive (\code{"S"}) arguments when the result type is not \code{"S"} (i.e., the operator changes the type).
    }
  \item Replace each placeholder in \code{rule_translation} by whole-word match.
}

\strong{Composition mode} (\code{operator} is \code{NULL}):

Exactly two arguments are required. The composition rules are:
\describe{
  \item{S + A \verb{->} S}{Simple juxtaposition: \code{"X Y"}.}
  \item{S + S \verb{->} S}{Without comma separator: \code{"X of/with the Y"}.}
  \item{S, S \verb{->} S}{With comma separator: \code{"X, the Y"}.}
  \item{S + V \verb{->} SEN}{Subject-verb: \code{"X Y"} (with \code{"to "} stripped from verb).}
  \item{SEN + SEN \verb{->} SEN}{Sentence concatenation: \code{"X. Y"} (period added if missing).}
}

In both modes, translations of the form \code{"general \{specific\}"} are reduced to the specific part before processing.
}
\value{
A character vector of length 2: \code{c(result_type, result_translation)}.

On error (e.g. incompatible types in composition), a character vector of length 1 containing the error message.
}
\examples{
# --- Base type (operator without arguments) ---
# A simple substantive passes through unchanged:
sumer:::eval_operator("S", "temple")
# [1] "S"      "temple"

# With {specific} extraction:
sumer:::eval_operator("S", "place {earth}")
# [1] "S"     "earth"

# --- Operator with one argument: Sx->V ---
# The S argument gets "the " prepended (since result type is V, not S):
args <- list(c("S", "temple"))
sumer:::eval_operator("Sx->V", "to utilize S", args)
# [1] "V"                      "to utilize the temple"

# --- Operator with two S arguments: SSx->V ---
# Duplicate types get indexed names (S1, S2):
args <- list(c("S1", "leader"), c("S2", "temple"))
sumer:::eval_operator("SSx->V", "to equip S1 with S2", args)
# [1] "V"                                  "to equip the leader with the temple"

# --- Composition: S + A -> S ---
args <- list(c("S", "temple"), c("A", ", which is a great one"))
sumer:::eval_operator(NULL, NULL, args)
# [1] "S"            "temple, which is a great one"

# --- Composition: S + V -> SEN ---
args <- list(c("S", "men"), c("V", "to bring grain into the temple"))
sumer:::eval_operator(NULL, NULL, args)
# [1] "SEN"       "men bring grain into the temple"

# --- Full pipeline with add_brackets and apply_translation_rules ---
x <- "mec3-ki-aj2-ga-ce-er"
x <- as.cuneiform(x)
x

meaning <- rbind( c("S",      "a man who relies on his own strength"),
                  c("S",      "place {earth}"),
                  c("Sx->A",  ", whose allocated resource is S"),
                  c("xS->A",   ", whose sustenance is S"),
                  c("S",      "grain"),
                  c("Sx->S",  "lamented S"))

df <- data.frame(
    type  = meaning[,1],
    translation = meaning[,2],
    expr  = split_sumerian(x)$signs)

s <- x
for(i in 1:nrow(df)){
 s <- sub(df$expr[i], paste0("#", i), s)
}
s

s_bracketed <- sumer:::add_brackets(s, df$type)
s_bracketed

apply_translation_rules(s_bracketed$string, df$type, df$translation)
}

\seealso{
\code{\link{apply_translation_rules}} which calls this function,
\code{\link{compose_skeleton_entry}} for the full pipeline
}
\keyword{internal}

\name{fill_substr_info}
\alias{fill_substr_info}
\title{Read Back a Saved Translation into a Substring Data Frame}
\description{
Reads a saved skeleton file (or character vector) and reconstructs the substring data frame that can be passed as \code{fill} to \code{\link{translate}} or \code{\link{skeleton}}. This allows resuming work on a translation that was saved earlier with \code{writeLines}.

The counterpart to this function is \code{\link{guess_substr_info}}, which fills the substring data frame from dictionaries. In contrast, \code{fill_substr_info} fills it from a skeleton that already contains manually entered types and translations.
}
\usage{
fill_substr_info(skeleton)
}
\arguments{
  \item{skeleton}{A file path to a saved skeleton text file, or a character vector as returned by \code{\link{skeleton}} or \code{\link{translate}}.}
}
\details{
A typical workflow for translating Sumerian texts spans multiple sessions:

\enumerate{
  \item Call \code{\link{translate}} to interactively translate a line.
  \item Save the result with \code{writeLines(result, "Line_29.txt")}.
  \item In a later session, call \code{fill_substr_info("Line_29.txt")} to reload the saved translation.
  \item Pass the result as \code{fill} to \code{\link{translate}} to continue editing.
}

The function parses the skeleton format (header line \code{"Structure: ..."} followed by entry lines starting with \code{|}), extracts the type and translation from each entry, and places them at the correct positions in a substring data frame as created by \code{\link{init_substr_info}}.
}
\value{
A data frame with \eqn{N(N+1)/2} rows (where \eqn{N} is the number of cuneiform tokens) and the following columns:
\describe{
  \item{start}{Integer. The 1-based position of the first token in the substring.}
  \item{n_tokens}{Integer. The number of tokens in the substring.}
  \item{expr}{Character. The concatenated cuneiform signs of the substring.}
  \item{type}{Character. The grammatical type (e.g. \code{"S"}, \code{"V"}, \code{"Sx->V"}), or \code{""} if not yet specified.}
  \item{translation}{Character. The translation, or \code{""} if not yet specified.}
}
Rows without a corresponding skeleton entry have empty type and translation fields. The row order matches \code{\link{init_substr_info}}, so indices can be computed with \code{\link{substr_position}}.
}
\seealso{
\code{\link{translate}} for the interactive translation tool,
\code{\link{skeleton}} for creating and displaying translation templates,
\code{\link{guess_substr_info}} for filling a substring data frame from dictionaries,
\code{\link{init_substr_info}} for the underlying data frame structure
}
\examples{
skeleton_file  <- system.file("extdata", "project/lines/Line_29.txt", package = "sumer")
the_skeleton   <- readLines(skeleton_file)

#Get the cuneiform text of the line:
x <- sub("^Structure: ", "", the_skeleton[1])
x

#See the whole file:
cat(the_skeleton, sep="\n")


df_fill <- fill_substr_info(skeleton_file)


\dontrun{

#Use the result of the function to revise the translation:

dict_file <- system.file("extdata", "sumer-dictionary.txt", package = "sumer")
text_file <- system.file("extdata", "project", "enki_and_the_world_order.txt", package = "sumer")

result <- translate(x,
                text = text_file,
                dic = dict_file,
                fill = df_fill,
                min_freq = c(6, 4, 2),
                sentence_prob = 0.25)
print(result)
# Now you may save the result with writeLines.
}

}
\keyword{utilities}

\name{as.cuneiform}
\alias{as.cuneiform}
\alias{as.cuneiform.default}
\alias{as.cuneiform.character}
\alias{print.cuneiform}

\title{Convert Transliterated Sumerian Text to Cuneiform}
\description{
Converts transliterated Sumerian text to Unicode cuneiform characters. This is a generic function with a method for character vectors.
}
\usage{
as.cuneiform(x, ...)

\method{as.cuneiform}{default}(x, ...)

\method{as.cuneiform}{character}(x, mapping = NULL, ...)

\method{print}{cuneiform}(x, ...)
}
\arguments{
  \item{x}{
    For \code{as.cuneiform}: An object to be converted to cuneiform. Currently, only character vectors are supported.

    For \code{print.cuneiform}: an object of class \code{"cuneiform"}.
  }
  \item{mapping}{A data frame containing the sign mapping table with columns \code{syllables}, \code{name}, and \code{cuneiform}. If \code{NULL} (the default), the package's internal mapping file \file{etcsl_mapping.txt} is loaded. Only used by the character method.}
  \item{...}{Additional arguments passed to methods.}
}
\details{
The function processes each element of the input character vector by:
\enumerate{
  \item Calling \code{\link{info}} to look up sign information for each transliterated sign.
  \item Extracting the Unicode cuneiform symbols for each sign.
  \item Reconstructing the cuneiform text using the original separators. Hyphens and periods between signs are removed since cuneiform signs are written without separators. Hyphens before numbers are replaced by spaces so that numeric tokens remain distinguishable from adjacent signs.
}

The default method throws an error for unsupported input types.
}
\value{
  \code{as.cuneiform} returns a character vector of class \code{cuneiform} with the cuneiform representation of each input element.

  \code{print.cuneiform} displays a character vector of class \code{cuneiform}.
}
\note{
The cuneiform output requires a font that supports the Unicode Cuneiform block (U+12000 to U+12500) to display correctly.
}
\seealso{
\code{\link{info}} for retrieving detailed sign information,
\code{\link{split_sumerian}} for splitting Sumerian text into signs,
\code{\link{as.sign_name}} for converting transliterated Sumerian text to sign names
}
\examples{

# Convert transliterated text to cuneiform
as.cuneiform(c("na-an-jic li-ic ma","en tarah-an-na-ke4"))

# Load transliterated text from a file
file <- system.file("extdata", "transliterated-text.txt", package = "sumer")
x <- readLines(file)
cat(x, sep="\n")

# Convert transliterated text to cuneiform
as.cuneiform(x)

# Using a custom mapping table
path <- system.file("extdata", "etcsl_mapping.txt", package = "sumer")
my_mapping <- read.csv2(path, sep=";", na.strings="")
as.cuneiform("lugal", mapping = my_mapping)

}
\keyword{utilities}
\keyword{character}
\keyword{methods}

\name{substr_position}
\alias{substr_position}
\title{Compute Row Index of a Substring in a Substring Data Frame}
\description{
Computes the row index of a substring in the data frame created by \code{\link{init_substr_info}}, given its starting position, its length in tokens, and the total number of tokens.

This is an internal helper function.
}
\usage{
substr_position(start, n_tokens, N)
}
\arguments{
  \item{start}{Integer (or integer vector). The starting position of the substring (1-based).}
  \item{n_tokens}{Integer (or integer vector). The number of tokens in the substring.}
  \item{N}{Integer. The total number of tokens in the full token sequence.}
}
\details{
The data frame returned by \code{\link{init_substr_info}} is ordered by \code{n_tokens} descending and \code{start} ascending. This function computes the corresponding row index using the formula

\deqn{\mathrm{row} = \frac{(N - k)(N - k + 1)}{2} + s}{row = (N - k)(N - k + 1) / 2 + s}

where \eqn{k} = \code{n_tokens} and \eqn{s} = \code{start}.

The function is vectorized: if \code{start} and \code{n_tokens} are vectors of the same length, a vector of row indices is returned.
}
\value{
A numeric vector of row indices (1-based).
}
\seealso{
\code{\link{init_substr_info}} for creating the substring data frame
}
\examples{

# Create a character vector with tokens
x <- "<d-nu-dim2-mud> ki a. jal2 (e2{kur}) ra. gaba jal2. an ki a"
token <- split_sumerian(as.cuneiform(x))$signs
token

N <- length(token)

# Create a data frame with all substrings
df <- sumer:::init_substr_info(token)

# The full string (start=1, n_tokens=N) is in row 1
pos <- sumer:::substr_position(1, N, N)
pos
df$expr[pos]


# The last single token (start=N, n_tokens=1) is in the last row
pos <- sumer:::substr_position(N, 1, N)
pos
df$expr[pos]

# Vectorized call
start <- c(1, 2, 1)
n_token <- c(2, 2, 1)
pos <- sumer:::substr_position(start, n_token, N)
pos
df$expr[pos]
}
\keyword{internal}

\name{add_brackets}
\alias{add_brackets}
\title{Add Brackets to a Structure String}
\description{
Inserts curly braces into a structure string so that operators are grouped with their arguments and remaining elements are grouped by compositional rules.

The function processes \code{\{...\}} groups already present in the input recursively, then resolves operator binding and compositional grouping rules according to a priority list.
}
\usage{
add_brackets(s, type)
}
\arguments{
  \item{s}{Character string containing \code{#N} tags and optional \code{\{...\}} groups, as produced by step a of \code{\link{compose_skeleton_entry}}.}
  \item{type}{Character vector of type annotations for the children. \code{type[k]} is the type string for tag \code{#k} (e.g. \code{"S"}, \code{"V"}, \code{"Sx->V"}, \code{"xS->S"}).}
}
\details{
The function first recursively processes any \code{\{...\}} groups already present in the input. Then it applies rules from a priority list, restarting after each successful match:

\enumerate{
  \item \strong{Pipe rules:} Operators whose result type matches the first argument's base type (the type "flows through").
  \item \strong{Explicit rules:} Specific operator types bound by exact match (e.g. \code{xS->A}).
  \item \strong{General operator rules:} Any remaining operators that match their argument types.
  \item \strong{Compositional grouping:} Adjacent elements are grouped by type: S+A, S,S (comma), S+Vt, S+S, S+Vi, S+V, SEN+SEN.
}

If operators cannot find matching arguments, an error is returned.
}
\value{
A list with two elements:
\describe{
  \item{string}{The structure string with all brackets inserted.}
  \item{type}{The resulting type of the outermost group (\code{"S"}, \code{"V"}, \code{"SEN"}, or \code{"ERROR"}).}
}
}
\examples{
x <- "mec3-ki-aj2-ga-ce-er"
x <- as.cuneiform(x)
x

meaning <- rbind( c("S",      "a man who relies on his own strength"),
                  c("S",      "place {earth}"),
                  c("Sx->A",  ", whose allocated resource is S"),
                  c("xS->A",   ", whose sustenance is S"),
                  c("S",      "grain"),
                  c("Sx->S",  "lamented S"))

df <- data.frame(
    type  = meaning[,1],
    translation = meaning[,2],
    expr  = split_sumerian(x)$signs)

s <- x
for(i in 1:nrow(df)){
 s <- sub(df$expr[i], paste0("#", i), s)
}
s

(s_bracketed  <- sumer:::add_brackets(s,df$type))
apply_translation_rules(s_bracketed$string, df$type, df$translation)
}

\seealso{
\code{\link{apply_translation_rules}} for the next step (translating the bracketed string),
\code{\link{compose_skeleton_entry}} which calls this function
}
\keyword{internal}
